/*
 * This file is part of zaniWok.
 *
 * zaniWok is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.  See the file COPYING for more information.
 *
 * zaniWok is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */


#import "ClockCard.h"
#import <sys/param.h>		/* for MAXPATHLEN */
#import <assert.h>
#import <time.h>
#import "memory.h"


@implementation ClockCard


- initFromBundle:aBundle
{
  [super initFromBundle:aBundle];
  return self;
}

- insert:(MachineState *) st inComputer:(Computer *) comp inSlot:(int)slot
{
  [super insert:st inComputer:comp inSlot:slot];

 /* Trap the signature byte addresses. */
  [self addAddressHandler:IO_SELECT_ADDRESS_BEGIN(slot_number)
    :IO_SELECT_ADDRESS_END(slot_number)
    :@selector(signatureBytes::)
    :READ];

 /* Trap the clock card firmware routines. */
  [self addRoutineHandler:IO_SELECT_ADDRESS_BEGIN(slot_number) + 0x8
   :IO_SELECT_ADDRESS_BEGIN(slot_number) + 0x8
   :@selector(readTime)];
  [self addRoutineHandler:IO_SELECT_ADDRESS_BEGIN(slot_number) + 0xB
   :IO_SELECT_ADDRESS_BEGIN(slot_number) + 0xB
   :@selector(writeTime)];

  return self;
}



/* Return the signature bytes identifying us as a clock card.  See
 * ProDOS 8 Technical Reference Manual.
 */
- (int)signatureBytes:(AccessType) t :(MemoryElement *) m
{
  switch (m - &state->sysmem[IO_SELECT_ADDRESS_BEGIN(slot_number)])
  {
  case 0x00:
    return 0x08;
  case 0x02:
    return 0x28;
  case 0x04:
    return 0x58;
  case 0x06:
    return 0x70;
  }

  return 0;
}

#define TIME_BUFFER_LENGTH 16

- (BOOL)readTime
{
  struct timeval      tp;
  struct timezone     tzp;
  time_t              sec;
  struct tm          *clock;
  char                timeBuffer[TIME_BUFFER_LENGTH];
  int                 bufferIndex;

 /*
  * Copy a string to the GETLN input buffer (as described in the
  * ProDOS 8 Technical Reference Manual, p. 105). 
  */
  gettimeofday(&tp, &tzp);
  sec = tp.tv_sec;
  clock = localtime(&sec);
  sprintf(timeBuffer, "%02d,%02d,%02d,%02d,%02d\x0D",	/* FIXME - need $0D? */
	  clock->tm_mon + 1, clock->tm_wday, clock->tm_mday,
	  clock->tm_hour, clock->tm_min);

  bufferIndex = TIME_BUFFER_LENGTH - 1;
  while (bufferIndex--)
    timeBuffer[bufferIndex] |= 0x80;
  [computer copyBytesToApple:timeBuffer :0x200 :TIME_BUFFER_LENGTH - 1];

 /*
  * Prodos 8 Tech Ref Man p.104 says routines should begin with
  * CLD. 
  */
  state->d_flag = 0;

  return YES;
}

- (BOOL)writeTime
{
 /*
  * Prodos 8 Tech Ref Man p.104 says routines should begin with
  * CLD. 
  */
  state->d_flag = 0;

 /*
  * Do not actually set the time - I'm not sure what to do here.
  * Perhaps we could remember an offset from "real time"? 
  */
  return YES;
}

@end
