/*
 * This file is part of zaniWok.
 *
 * zaniWok is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.  See the file COPYING for more information.
 *
 * zaniWok is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */


#import <appkit/appkit.h>
#ifndef USE21
#import <mach/cthreads.h>
#else
#import <cthreads.h>
#endif
#import "execute.h"

@interface Computer:Responder
{
  id window;
  id wozView;
  int bufferKeys, translateNL, maxPasteLines, appleKeyBit, optionKeyBit;
  int appleKey, optionKey, shiftKey;
  int lmouse, rmouse;
  MemoryElement *realmem;
  MachineState machine_state;
  cthread_t execute_cthread;
  BOOL isPaused;
  DPSTimedEntry timedEntry;
}

/* Standard methods. */
- init;
- free;

/* Ways to manipulate the computer. */
- powerUp;
- powerOff;
- reset;
-(int) appleKeyBit;
-(int) optionKeyBit;
- setAppleKey:(int) upordown;
- setOptionKey:(int) upordown;
- setShiftKey:(int) upordown;
- changeKeyMap:sender;
- setBufferKeys:(BOOL) buffer;
- setTranslateNL:(BOOL) trans;
- setMaxPasteLines:(int) max;
-(BOOL) bufferKeys;
-(BOOL) translateNL;
-(int) maxPasteLines;

/* Fun things to do to the 6502 thread. */
- pause6502thread;
- unpause6502thread;
- kill6502thread;

/* Random informational methods. */
- window;
- wozView;
- (MachineState *)machineState;
- (BOOL)isPaused;
- cardInSlot:(int)slot;

/* Adjust the joystick's position. */
- setJoystickXY:(int)x :(int)y;

/* Memory access methods. */
- (unsigned char)readByte :(unsigned)addr;
- (unsigned short)readShort :(unsigned)addr;
- writeByte:(unsigned)addr :(unsigned char)value;
- writeShort:(unsigned)addr :(unsigned short)value;
- (unsigned char)readMemoryElement:(MemoryElement *)m;
- writeMemoryElement:(MemoryElement *)m :(unsigned char)value;
- copyBytesToApple:(const unsigned char *)src :(unsigned short)dst_addr
                  :(int)num_bytes;
- copyBytesFromApple:(unsigned short)src_addr :(unsigned char *)dst
                    :(int)num_bytes;

/* Peripheral card support. */
- insertPeripheralCard:aCard inSlot:(int) aSlot;
- loadPeripheralCard:(const char *)cardName inSlot:(int)slot;

/* Event handling sludge. */
-(BOOL) acceptsFirstResponder;
- keyDown:(NXEvent *) theEvent;
- mouseMoved:(NXEvent *)event;
- mouseDragged:(NXEvent *)event;
- rightMouseDragged:(NXEvent *)event;

- bloadFile:(const char *)file :(int)memloc;

@end
