/*
 * This file is part of zaniWok.
 *
 * zaniWok is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.  See the file COPYING for more information.
 *
 * zaniWok is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */
/*
 * Controller.m - does all initialization and handles
 *		  control panel
 * written by Chris Osborn <fozztexx%nvcc.uucp@groucho.sonoma.edu>
 */

#import "Controller.h"
#import "ZaniWokApp.h"
#import "Computer.h"
/* #import "disk.h" */
#import "IconWell.h"
#import "PeripheralCard.h"
#import <objc/objc-runtime.h>
#import <defaults/defaults.h>
#import <assert.h>
#import <appkit/appkit.h>
#import <objc/List.h>
#import <defaults/defaults.h>
#import <strings.h>
#import <sys/types.h>
#import <sys/stat.h>
#import <sys/dir.h>
#import <sys/param.h>

#define APPNAME [NXApp appName]

@implementation Controller

+ initialize
{
  static const NXDefaultsVector vec = {
				       {"CardInSlot2", "UnixShare.a2card"},
				       {"CardInSlot3", "ClockCard.a2card"},
				       {"CardInSlot6", "DiskCard.a2card"},
				       {NULL, NULL}
  };

  [super initialize];
  NXRegisterDefaults([NXApp appName], vec);
  return self;
}

- init
{
  [super init];

  emulatorList = [[List alloc] init];

  return self;
}

- free
{
  [[emulatorList freeObjects] free];
  return[super free];
}

- initPanel
{
  NXSize              space;
  int                 r, c;
  const char         *const types[] = {"a2card", NULL};


  [[myProto cell] setShowsStateBy:NX_CHANGEBACKGROUND];
  [[myProto cell] setHighlightsBy:NX_CHANGEBACKGROUND];

  optionButtons = [[Matrix allocFromZone:[self zone]] initFrame:NULL mode:NX_RADIOMODE prototype:[myProto cell] numRows:1 numCols:0];
  space.width = 64;
  space.height = 64;
  [optionButtons setCellSize:&space];
  space.width = 0;
  space.height = 0;
  [optionButtons setIntercell:&space];
  [optionButtons setAutoscroll:YES];
  [optionButtons setAction:@selector(chooseOption:)];
  [optionButtons setTarget:self];

  [optionButtons getNumRows:&r numCols:&c];
  [[[optionButtons addCol] cellAt:0 :c++] setImage:[NXImage findImageNamed:"Keyboard"]];
  [[[optionButtons addCol] cellAt:0 :c++] setImage:[NXImage findImageNamed:"PeripheralCard"]];

  [optionScroll setBackgroundGray:NX_LTGRAY];
  [optionScroll setBorderType:NX_BEZEL];
  [optionScroll setHorizScrollerRequired:YES];
  [[optionScroll setDocView:optionButtons] free];

  optionList = [[List allocFromZone:[self zone]] init];
  [optionList addObject:keyboardView];
  [optionList addObject:cardView];

  [[optionView contentView] addSubview:[optionList objectAt:0]];
  lastOpView = [optionList objectAt:0];
  builtInOptions = [optionList count];
  [optionButtons sizeToCells];

  slotWells = [[List allocFromZone:[self zone]] init];
  [slotWells addObject:[slotWell1 setTypes:types]];
  [slotWells addObject:[slotWell2 setTypes:types]];
  [slotWells addObject:[slotWell3 setTypes:types]];
  [slotWells addObject:[slotWell4 setTypes:types]];
  [slotWells addObject:[slotWell5 setTypes:types]];
  [slotWells addObject:[slotWell6 setTypes:types]];
  [slotWells addObject:[slotWell7 setTypes:types]];

  return self;
}

- addCards:computer
{
  int                 i;
  id                  card;
  int                 r, c;
  int                 t, ab, ob;
  int                 sc = [optionButtons selectedCol];


  for (i = [optionList count]; i > builtInOptions; i--)
    [optionList removeLastObject];

  [[optionButtons window] disableFlushWindow];
  [optionButtons renewRows:1 cols:builtInOptions];
  for (i = 1; i <= 7; i++)
  {
    if (card = [computer cardInSlot:i])
    {
      if ([card iconImage])
	[[slotWells objectAt:i - 1] setImage:[card iconImage]];
      else
	[[slotWells objectAt:i - 1] setImage:[NXImage findImageNamed:"PeripheralCard"]];
      [[slotWells objectAt:i - 1] setFilename:[[card bundle] directory]];

      if ([card optionView])
      {
	[optionButtons addCol];
	[optionButtons getNumRows:&r numCols:&c];
	[[optionButtons cellAt:0 :c - 1] setImage:[card iconImage]];
	[optionList addObject:[card optionView]];
	[optionButtons sizeToCells];
	[optionButtons display];
      }
    }
  }

  ab = [computer appleKeyBit];
  ob = [computer optionKeyBit];

  t = 0;
  if (ab == NX_NEXTLCMDKEYMASK && ob == NX_NEXTLALTKEYMASK)
    t = 1;
  else
  if (ab == NX_NEXTLALTKEYMASK && ob == NX_NEXTRALTKEYMASK)
    t = 2;
  else
  if (ab == NX_NEXTRCMDKEYMASK && ob == NX_NEXTRALTKEYMASK)
    t = 3;

  [keyMap selectCellAt:t :0];

  [optionButtons selectCellAt:0 :sc];
  [self chooseOption:optionButtons];
  [[[optionButtons window] reenableFlushWindow] flushWindow];

  [keyBufButton setState:[computer bufferKeys]];
  [pasteTrans setState:[computer translateNL]];
  [maxPaste setIntValue:[computer maxPasteLines] at:0];

  return self;
}

- chooseOption:sender
{
  id                  newView;


  if ((newView = [optionList objectAt:[sender selectedCol]]) && newView != lastOpView)
  {
    [[optionView contentView] replaceSubview:lastOpView with:newView];
    lastOpView = newView;
  }

  [optionView display];

  return self;
}

- changeKeyBuffer:sender
{
  return[[self mainEmulator] setBufferKeys:[sender state]];
}

- changeKeyMap:sender
{
  return[[self mainEmulator] changeKeyMap:sender];
}

- changeTrans:sender
{
  return[[self mainEmulator] setTranslateNL:[sender state]];
}

- changeMaxChars:sender
{
  return[[self mainEmulator] setMaxPasteLines:[sender intValueAt:0]];
}

- loadConfig:(const char *)filename
{
  NXTypedStream      *stream;
  id                  new;


  if (stream = NXOpenTypedStreamForFile(filename, NX_READONLY))
  {
    new = NXReadObject(stream);
    NXCloseTypedStream(stream);
    [emulatorList addObject:new];
    [self addCards:new];
    [[new window] setTitleAsFilename:filename];
    return[new powerUp];
  }
  else
    return nil;
}

- newEmulator:sender
{
  id                  new;
  char                buf[MAXPATHLEN + 1];


  sprintf(buf, "%s/Library/Apple2/Default.a2config", NXHomeDirectory());
  if (![self loadConfig:buf])
  {				/* Can't load default, make new */
    new = [[Computer alloc] init];
    if (new == nil)
      return nil;
    [emulatorList addObject:new];

    [self addCards:new];
    return[new powerUp];
  }

  return self;
}

- openEmulator:sender
{
  const char         *const types[] = {"a2config", NULL};
  id                  openPanel;


  openPanel = [OpenPanel new];
  if ([openPanel runModalForTypes:types])
    return[self loadConfig:[openPanel filename]];

  return nil;
}

- saveMainEmulatorAs:(const char *)filename
{
  id                  maincomp = [self mainEmulator];
  NXTypedStream      *stream;


  stream = NXOpenTypedStreamForFile(filename, NX_WRITEONLY);
  NXWriteRootObject(stream, maincomp);
  NXCloseTypedStream(stream);
  [maincomp setTitleAsFilename:filename];

  return self;
}

- saveEmulator:sender
{
  char                buf[MAXPATHLEN + 1];
  char                buf2[MAXPATHLEN + 1];
  char               *p;
  id                  maincomp = [self mainEmulator];


  if (maincomp)
  {
    strcpy(buf,[[maincomp window] title]);
    p = index(buf, 208);
    strcpy(buf2, p + 3);
    strcat(buf2, "/");
    strcat(buf2, buf);
    p = index(buf2, 208);
    *(p - 2) = 0;
    [self saveMainEmulatorAs:buf2];
  }

  return self;
}

- saveEmulatorAs:sender
{
  const char         *type = "a2config";
  id                  maincomp = [self mainEmulator];
  id                  savePanel;


  if (maincomp)
  {
    savePanel = [SavePanel new];
    [savePanel setRequiredFileType:type];
    if ([savePanel runModal])
      [self saveMainEmulatorAs:[savePanel filename]];
  }

  return self;
}

- removeEmulator:(id)emulator
{
  [emulatorList removeObject:emulator];
  return self;
}

/* Returns the Computer object in the main window, or nil if there isn't one.*/
- mainEmulator
{
  id                  m = [NXApp mainWindow];
  int                 i;

  for (i = [emulatorList count] - 1; i >= 0; i--)
    if ([[emulatorList objectAt:i] window] == m)
      return[emulatorList objectAt:i];
  return nil;
}

- resetEmulator:sender
{
  [[self mainEmulator] reset];
  return self;
}

/* IconWell Delegate Methods */

- iconWell:sender willUseImage:anImage forFile:(const char *)aFile
{
  id                  aCard;
  id                  bundle;
  id                  image = anImage;


  bundle = [[NXBundle alloc] initForDirectory:aFile];
  aCard = [[[bundle principalClass] alloc] initFromBundle:bundle];
  if ([aCard iconImage])
    image = [[aCard iconImage] copyFromZone:[self zone]];
  [aCard free];

  return image;
}

- iconWell:sender didAcceptFile:(const char *)aFile
{
  id                  card, bundle;
  int                 i;


  for (i = 0; i < [slotWells count]; i++)
    if ([slotWells objectAt:i] == sender)
      break;

  if (i < 7)
  {
    bundle = [[NXBundle alloc] initForDirectory:aFile];
    card = [[[bundle principalClass] alloc] initFromBundle:bundle];
    [[self mainEmulator] insertPeripheralCard:card inSlot:i + 1];
    [self addCards:[self mainEmulator]];
  }

  return self;
}

- iconWell:sender didRemoveFile:(const char *)aFile
{
  int                 i;


  for (i = 0; i < [slotWells count]; i++)
    if ([slotWells objectAt:i] == sender)
      break;

  if (i < 7)
  {
    [[[[self mainEmulator] cardInSlot:i + 1] removeFromSlot] free];
    [self addCards:[self mainEmulator]];
  }

  return self;
}

/* Application Delegate Methods */

- appDidInit:sender
{
  const char         *types[3];
  int                 i = 0;
  char                path[MAXPATHLEN], path2[MAXPATHLEN];
  id                  bundle;
  DIR                *dir;
  struct direct      *entry;


  types[i++] = NXAsciiPboardType;
  [NXApp registerServicesMenuSendTypes:types andReturnTypes:NULL];

  [self initPanel];

#ifndef USE21
  objc_setMultithreaded(YES);
#endif

 /* Load up all available a2cards */

  sprintf(path, "/LocalLibrary/Apple2/PeripheralCards");
  if (dir = opendir(path))
  {
    while (entry = readdir(dir))
      if (!strcmp(&entry->d_name[entry->d_namlen - 7], ".a2card"))
      {
	sprintf(path2, "%s/%s", path, entry->d_name);
	bundle = [[NXBundle alloc] initForDirectory:path2];
	[bundle principalClass];/* Ask for it just to make sure it loads */
      }
    closedir(dir);
  }

  sprintf(path, "%s/Library/Apple2/PeripheralCards", NXHomeDirectory());
  if (dir = opendir(path))
  {
    while (entry = readdir(dir))
      if (!strcmp(&entry->d_name[entry->d_namlen - 7], ".a2card"))
      {
	sprintf(path2, "%s/%s", path, entry->d_name);
	bundle = [[NXBundle alloc] initForDirectory:path2];
	[bundle principalClass];/* Ask for it just to make sure it loads */
      }
    closedir(dir);
  }

  return self;
}

- (BOOL)appAcceptsAnotherFile:sender
{
  return YES;
}

- (int)app:sender openFile:(const char *)filename type:(const char *)aType
{
  if (!strcmp(aType, "a2config"))
  {
    if ([self loadConfig:filename])
      return YES;
    return NO;
  }

  return NO;
}

- appWillTerminate:sender
{
 /* Stop all emulators so they save whatever state they need to. */
  [emulatorList makeObjectsPerform:@selector(powerOff)];
  return self;
}

- app:sender powerOffIn:(int)ms andSave:(int)flag
{
 /* Stop all emulators so they save whatever state they need to. */
  [emulatorList makeObjectsPerform:@selector(powerOff)];
  return self;
}

/* Window Delegate Methods */

- windowDidBecomeMain:sender
{
  return self;
}

@end
