/*
 * This file is part of zaniWok.
 *
 * zaniWok is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.  See the file COPYING for more information.
 *
 * zaniWok is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */


/*
 * TODO:
 * 1) DOS, Prodos support
 * 2) Evaluate performance effect of track cache; I suspect NXStream's "cache"
 *    may be good enough that there is no point bothering with a cache.
 * 3) Implement extensions for an inspector, etc.
 */


#define USE_MEMORY_STREAM	/* The alternative doesn't seem to work. I guess NeXT meant it when they said only use NXOpenFile for non-files... */
#define APPNAME [NXApp appName]


#import "DiskCard.h"
#import <sys/param.h>		/* for MAXPATHLEN */
#import <assert.h>
#import <sys/dir.h>
#import "memory.h"

@interface DiskCard(Private)
- _setupWithBundle:aBundle;
@end


@implementation DiskCard

- _setupDiskCardWithBundle:aBundle;
{
  char                path[MAXPATHLEN + 1];
  int                 i;


  drive_mutex = mutex_alloc();

 /* Set up default variables. */
  drive = 0;

 /* Start out with disk drive hardware in known state. */
  drive_state[0].qtrack = 0;
  drive_state[0].write_mode = NO;
  drive_state[0].on = NO;
  drive_state[0].cached_qtrack_number = NO_QTRACK;
  drive_state[0].cached_qtrack_data = NULL;
  for (i = 0; i < 4; i++)
    drive_state[0].magnet[i] = NO;
  drive_state[1] = drive_state[0];

 /* Start out with the disk structs in a clean state. */
  memset((void *)&disk[0], 0, sizeof(disk));
  disk[0].type = DISK_NO_DISK;	/* No disk in drive 1.; */
  disk[1].type = DISK_NO_DISK;	/* No disk in drive 2, either.; */

  [self loadDiskDir];

  return self;
}

- initFromBundle:aBundle
{
  [super initFromBundle:aBundle];
  return[self _setupDiskCardWithBundle:aBundle];
}

- snorkle
{
  int                 i;
  char                path[MAXPATHLEN + 1];
  char                dname1[MAXPATHLEN + 1], dname2[MAXPATHLEN + 1];


  [super awake];

  if ([bundle getPath:path forResource:"DiskCard" ofType:"nib"])
  {
    strcpy(dname1,[diskList1 title]);
    strcpy(dname2,[diskList2 title]);
    [NXApp loadNibFile:path owner:self];
    [self insertAlphabetically:dname1 into:[diskList1 target]];
    [diskList1 setTitle:dname1];
    [self insertAlphabetically:dname2 into:[diskList2 target]];
    [diskList2 setTitle:dname2];
  }

  drive_mutex = mutex_alloc();

  drive = 0;

  drive_state[0].qtrack = 0;
  drive_state[0].write_mode = NO;
  drive_state[0].on = NO;
  drive_state[0].cached_qtrack_number = NO_QTRACK;
  drive_state[0].cached_qtrack_data = NULL;
  for (i = 0; i < 4; i++)
    drive_state[0].magnet[i] = NO;
  drive_state[1] = drive_state[0];

 /* Start out with the disk structs in a clean state. */
  memset((void *)&disk[0], 0, sizeof disk);
  disk[0].type = DISK_NO_DISK;	/* No disk in drive 1.; */
  disk[1].type = DISK_NO_DISK;	/* No disk in drive 2, either.; */

  [self loadDiskDir];

  [self insertDisk:[diskList1 title] inDrive:0];
  [self insertDisk:[diskList2 title] inDrive:1];

  return self;
}


- free
{
  [[self ejectDisk:0] ejectDisk:1];	/* Eject both disks. */
  mutex_lock(drive_mutex);
  mutex_unlock(drive_mutex);
  mutex_free(drive_mutex);

  return[super free];
}


- setDiskList1:anObject
{
  id                  matrix = [[anObject target] itemList];


  [matrix setTarget:self];
  [matrix setAction:@selector(changeDisk:)];
  [matrix setTag:2];
  diskList1 = anObject;

  return self;
}

- setDiskList2:anObject
{
  id                  matrix = [[anObject target] itemList];


  [matrix setTarget:self];
  [matrix setAction:@selector(changeDisk:)];
  [matrix setTag:3];
  diskList2 = anObject;

  return self;
}

- insertAlphabetically:(const char *)str into:aPopUpList
{
  int                 i;
  int                 r, c;
  id                  matrix;


  matrix = [aPopUpList itemList];
  [matrix getNumRows:&r numCols:&c];

  for (i = 0; i < r; i++)
  {
    if (strcasecmp(str,[[matrix cellAt:i :0] title]) <= 0)
      break;
  }

  [aPopUpList insertItem:str at:i];

  return self;
}

- loadDiskDir
{
  id                  alist1, alist2;
  int                 i, j;
  DIR                *dir;
  struct direct      *entry;
  char                path[MAXPATHLEN + 1];


  alist1 = [diskList1 target];
  alist2 = [diskList2 target];
  i = [alist1 count];
  for (j = 0; j < i; j++)
    [alist1 removeItemAt:j];
  i = [alist2 count];
  for (j = 0; j < i; j++)
    [alist2 removeItemAt:j];

  if (dir = opendir("/LocalLibrary/Apple2/Disks"))
  {
    while (entry = readdir(dir))
      if (strcmp(entry->d_name, ".") && strcmp(entry->d_name, "..") && !strcmp(&entry->d_name[entry->d_namlen - 7], ".a2disk"))
      {
	sprintf(path, "/LocalLibrary/Apple2/Disks/%s", entry->d_name);
	[self insertAlphabetically:path into:alist1];
	[self insertAlphabetically:path into:alist2];
      }
    closedir(dir);
  }

  sprintf(path, "%s/Library/Apple2/Disks", NXHomeDirectory());
  if (dir = opendir(path))
  {
    while (entry = readdir(dir))
      if (strcmp(entry->d_name, ".") && strcmp(entry->d_name, "..") && !strcmp(&entry->d_name[entry->d_namlen - 7], ".a2disk"))
      {
	sprintf(path, "%s/Library/Apple2/Disks/%s", NXHomeDirectory(), entry->d_name);
	[self insertAlphabetically:path into:alist1];
	[self insertAlphabetically:path into:alist2];
      }
    closedir(dir);
  }

  [self insertAlphabetically:"No Disk" into:alist1];
  [self insertAlphabetically:"No Disk" into:alist2];

  return self;
}

static long
get4(NXStream *s)
{
#ifdef __BIG_ENDIAN__
  if (sizeof(long) == 4)
  {
    long                n;

    NXRead(s, &n, sizeof(long));
    return n;
  }
#endif
  return ((NXGetc(s) << 24) | (NXGetc(s) << 16)
	  | (NXGetc(s) << 8) | NXGetc(s));
}

static short
get2(NXStream *s)
{
#ifdef __BIG_ENDIAN__
  if (sizeof(short) == 2)
  {
    short               n;

    NXRead(s, &n, sizeof(short));
    return n;
  }
#endif
  return ((NXGetc(s) << 8) | NXGetc(s));
}


- insert:(MachineState *) st inComputer:aComputer inSlot:(int)slot
{
  char                path[MAXPATHLEN + 1];
  NXStream           *s;


  [super insert:st inComputer:aComputer inSlot:slot];
  mutex_lock(drive_mutex);

 /* Temporary hack - we shouldn't need the real PROMs. */
  if ([bundle getPath:path forResource:"DISK" ofType:"PROM"]
      && (s = NXMapFile(path, NX_READONLY)) != NULL)
  {
    MemoryElement      *copyElement;
    int                 copyIndex;

  /* Read in the PROM. */
    NXRead(s, prom_image, 256);
    NXCloseMemory(s, NX_FREEBUFFER);

  /* Temporary hack - set status byte appropriately. */
    prom_image[0xFE] = 0x9F;	/* See Prodos 8 Tech Ref Man p.114 */
    prom_image[0xFF] = 0x80;	/* Address of virtual driver. */

  /* Copy it to Apple2 memory. */
    copyElement = &state->sysmem[ROM_FOR_SLOT_END(slot_number)];
    copyIndex = 0x100;
    while (copyIndex--)
      [computer writeMemoryElement:copyElement--:prom_image[copyIndex]];
  }
  else
  {
    mutex_unlock(drive_mutex);
    NXRunAlertPanel(NULL, "Fatal error: unable to load file "
		    "\"DISK.PROM\".", NULL, NULL, NULL);
    return nil;
  }

 /* Trap references to the stepper motor. */
  [self addAddressHandler:SWITCHES_FOR_SLOT(slot_number)
    :SWITCHES_FOR_SLOT(slot_number) + 0x7
   :@selector(stepperMotorReferenced::)
    :BOTH];

 /* Trap drive on/drive off soft switches. */
  [self addAddressHandler:0xC088 + (slot * 0x10)
    :0xC089 + (slot * 0x10)
    :@selector(setMotorOn::)
    :BOTH];

 /* Trap drive select soft switches. */
  [self addAddressHandler:0xC08A + (slot * 0x10)
    :0xC08B + (slot * 0x10)
    :@selector(selectDrive::)
    :BOTH];

 /* Trap drive select soft switches. */
  [self addAddressHandler:0xC08C + (slot * 0x10)
    :0xC08F + (slot * 0x10)
    :@selector(handleDataLatch::)
    :BOTH];

  [self addAddressHandler:ROM_FOR_SLOT(slot_number)
    :ROM_FOR_SLOT_END(slot_number)
    :@selector(undoWrite::)
    :WRITE];

  [self addRoutineHandler:ROM_FOR_SLOT(slot_number)
   + prom_image[0xFF]
    :ROM_FOR_SLOT(slot_number)
   + prom_image[0xFF]
    :@selector(deviceDriverCalled)];

  [self addRoutineHandler:ROM_FOR_SLOT(slot_number)
    :ROM_FOR_SLOT(slot_number)
    :@selector(bootUp)];
  [self addRoutineHandler:ROM_FOR_SLOT(slot_number) + 0x5C
   :ROM_FOR_SLOT(slot_number) + 0x5C
   :@selector(cardReadSector)];

  mutex_unlock(drive_mutex);

  return self;
}

/* Emulates write protection by immediately undoing the effect of any write.
 * Who cares how slow this is?
 */
- (int)undoWrite:(AccessType) t :(MemoryElement *) m
{
  int                 val = prom_image[(m - &state->sysmem[ROM_FOR_SLOT(slot_number)]) & 0xFF];

  fprintf(stderr, "AAAAAAAAAAABBBBBBBBBBBBBB\n");
  m->mem.byte = val;		/* Can't use [computer writeMemoryElement ...] here. */
  return val;			/* otherwise, you'll get an infinite loop of traps  */
}

- removeFromSlot
{
  int                 i;

 /*
  * Make sure that any changes to either disk get saved, just to
  * be safe. 
  */
  for (i = 0; i < 2; i++)
  {
    Disk               *d = &disk[i];

    if (d->type != DISK_NO_DISK)
    {
      NXFlush(d->stream);
#ifdef USE_MEMORY_STREAM
      if (d->dirty)
	NXSaveToFile(d->stream, d->diskpath);
#endif
    }
  }

  mutex_lock(drive_mutex);
  {
    [super removeFromSlot];
  }
  mutex_unlock(drive_mutex);

  return self;
}


- changeDisk:sender
{
  id                  buttonObject;
  id                  myOP;
  const char         *const types[] = {"a2disk", NULL};


  if ([sender tag] > 1)
  {
    [self insertDisk:(char *)[[sender selectedCell] title] inDrive:[sender tag] % 2];
  }
  else
  {
    buttonObject = [sender tag] ? diskList2 : diskList1;

    myOP = [OpenPanel new];

    [myOP setTitle:"Choose disk"];
    if ([myOP runModalForTypes:types])
    {
      [self insertAlphabetically:[myOP filename] into:[buttonObject target]];
      [buttonObject setTitle:[myOP filename]];
      [self insertDisk:(char *)[buttonObject title] inDrive:[sender tag]];
    }
  }

  return self;
}

/* This inserts a new disk into the drive, ejecting the current disk if there
 * is already one in the drive.  Returns NULL if successful; otherwise it
 * returns a string describing the problem.
 * driveNumber should be either 0 or 1.
 */
- (const char *)insertDisk:(const char *)path inDrive:(int)driveNumber
{
  DiskType            old_type;
  unsigned long       file_size, start_of_data, pos;
  char                buf[MAXPATHLEN + 1];
  Disk               *d;
  DriveState         *s;
  extern char        *realpath(const char *, char *);

  mutex_lock(drive_mutex);

  old_type = disk[driveNumber].type;
  d = &disk[driveNumber];
  s = &drive_state[driveNumber];

  if (d->type != DISK_NO_DISK)
  {
    mutex_unlock(drive_mutex);
    [self ejectDisk:driveNumber];
  /*
   * To assure that we have the mutex and d and old_type are
   * valid, just call ourselves again. 
   */
    return[self insertDisk:path inDrive:driveNumber];
  }

 /* Get a stream for the disk. */
  d->write_protected = (access(path, W_OK) == -1);
#ifdef USE_MEMORY_STREAM
  d->stream = NXMapFile(path, (d->write_protected
			       ? NX_READONLY : NX_READWRITE));
#else
  d->fd = open(path, d->write_protected ? O_RDONLY : O_RDWR, 0777);
  if (d->fd >= 0)
    d->stream = NXOpenFile(d->fd, (d->write_protected
				   ? NX_READONLY : NX_READWRITE));
  else
    d->stream = NULL;
#endif

  if (d->stream == NULL)
  {
    mutex_unlock(drive_mutex);
    return "Unable to open disk file.";
  }

 /* Initialize various disk attributes. */
  d->auxdata = NULL;
  d->dirty = NO;

 /* Clear drive cache. */
  s->cached_qtrack_number = NO_QTRACK;
  NX_FREE(s->cached_qtrack_data);
  s->cached_qtrack_data = NULL;

 /* Figure out the size of the file. */
  NXSeek(d->stream, 0, NX_FROMEND);
  file_size = NXTell(d->stream);

 /* Check the header word. */
  NXSeek(d->stream, 0, NX_FROMSTART);
  if (get4(d->stream) == DISK_MAGIC_NUMBER)
  {
  /* Read in the header. */
    start_of_data = get4(d->stream);
    if (start_of_data > file_size)
    {
      mutex_unlock(drive_mutex);
      return "Corrupted disk file; invalid start_of_data field.";
    }
    d->type = NXGetc(d->stream);

    switch (d->type)
    {
    case DISK_PRODOS:
      d->u.prodos.num_blocks = (file_size - start_of_data) / 512;
      d->u.prodos.first_block_offset = start_of_data;
      if (old_type != DISK_PRODOS
	  || d->u.prodos.current_block >= d->u.prodos.num_blocks)
	d->u.prodos.current_block = 0;
      break;
    case DISK_DOS33:
      d->u.dos33.num_tracks = ((file_size - start_of_data)
			       / BYTES_PER_TRACK);
      if (d->u.dos33.num_tracks != 35 && d->u.dos33.num_tracks != 40)
      {
	d->type = DISK_NO_DISK;
	mutex_unlock(drive_mutex);
	return "Bad number of tracks specified on DOS disk.";
      }
      d->u.dos33.first_track_offset = start_of_data;
      d->u.dos33.current_sector = 0;

      if (drive_state[driveNumber].qtrack / 4 >= d->u.dos33.num_tracks)
	drive_state[driveNumber].qtrack = (d->u.dos33.num_tracks - 1) * 4;
      break;

    case DISK_RAW_BYTES:
      pos = start_of_data;
      d->u.raw_bytes.num_tracks = 35;

    /* Default to all tracks being empty (zero out all lengths). */
      memset(d->u.raw_bytes.track, 0, sizeof d->u.raw_bytes.track);

      while (1)
      {
	int                 qtr, length;

      /* Move on to the next track. */
	NXSeek(d->stream, pos, NX_FROMSTART);
	if (NXAtEOS(d->stream))
	  break;

      /* Get parameters for this track. */
	qtr = NXGetc(d->stream);
	length = get2(d->stream);

      /* Verify them. */
	if (qtr < 0 || qtr >= NUM_QUARTER_TRACKS)
	{
	  d->type = DISK_NO_DISK;
	  mutex_unlock(drive_mutex);
	  return "Bad disk file; invalid track number specified.";
	}
	if (d->u.raw_bytes.track[qtr].track_length != 0)
	{
	  d->type = DISK_NO_DISK;
	  mutex_unlock(drive_mutex);
	  return "Bad disk file; a track is listed more than once.";
	}

      /* Save parameters for this track for later use. */
	d->u.raw_bytes.track[qtr].track_length = length;
	d->u.raw_bytes.track[qtr].track_offset = pos + 3;

      /* See if this disk has 40 tracks. */
	if (qtr / 4 >= 35)
	  d->u.raw_bytes.num_tracks = 40;

      /* Move on to next track. */
	pos += 3 + length;
      }

    /*
     * See if the current quartertrack number too large for this
     * disk. 
     */
      if (drive_state[driveNumber].qtrack / 4 >= d->u.raw_bytes.num_tracks)
	drive_state[driveNumber].qtrack = (d->u.raw_bytes.num_tracks - 1) * 4;

      break;

    default:
      d->type = DISK_NO_DISK;
      mutex_unlock(drive_mutex);
      return "Unknown disk type specified in disk file.";
    }
  }
  else				/* No magic cookie, so it's just a straight Prodos disk image. */
  {
    d->type = DISK_PRODOS;
    d->u.prodos.num_blocks = file_size / 512;
    d->u.prodos.first_block_offset = 0;
    if (old_type != DISK_PRODOS
	|| d->u.prodos.current_block >= d->u.prodos.num_blocks)
      d->u.prodos.current_block = 0;
  }

 /*
  * Set up the display name of the disk.  To do this we will
  * strip any leading component of the path and strip off the
  * ".a2disk" suffix. "/tmp/foobar.a2disk" becomes "foobar", etc. 
  */
  if (strrchr(path, '/'))
    strcpy(buf, strrchr(path, '/') + 1);
  else
    strcpy(buf, path);
  if (strlen(buf) >= 7 && !strcmp(buf + strlen(buf) - 7, ".a2disk"))
    buf[strlen(buf) - 7] = '\0';
  buf[sizeof d->diskname - 1] = '\0';	/* Make sure it's short enough. */
  strcpy(d->diskname, buf);

 /*
  * Save the real path of the disk file so we can save it later
  * if it's been modified. 
  */
  if (realpath(path, buf) != NULL)
    strcpy(d->diskpath, buf);
  else
    strcpy(d->diskpath, "/dev/null");	/* What else can ya do? */

  mutex_unlock(drive_mutex);

  return NULL;
}

/* This ejects the disk from a drive, making sure that any changes made to
 * the disk actually get flushed to the file.  If there is no disk in that
 * drive, the request is ignored.  Returns self.
 */
- ejectDisk:(int)driveNumber
{
  Disk               *d;
  DriveState         *s;

  mutex_lock(drive_mutex);

  d = &disk[driveNumber];
  s = &drive_state[driveNumber];

 /* Clear drive cache. */
  s->cached_qtrack_number = NO_QTRACK;
  NX_FREE(s->cached_qtrack_data);
  s->cached_qtrack_data = NULL;

  if (d->type != DISK_NO_DISK)
  {
  /* Close the disk file and flush any changes. */

#ifdef USE_MEMORY_STREAM
/* This sucks!  I have to save the whole @#%*! file if I change one byte?
 * Someone clue me in on this.  Perhaps I should open the file via a file
 * descriptor and then use NXOpenFile()?
 */
    if (d->dirty)
      NXSaveToFile(d->stream, d->diskpath);
    NXCloseMemory(d->stream, NX_FREEBUFFER);
#else
    NXClose(d->stream);
    close(d->fd);
    d->fd = -1;
#endif

  /* No stream left. */
    d->stream = NULL;

  /* No disk in the drive any more... */
    d->type = DISK_NO_DISK;

  /* Update our display on the screen. */
  }

  mutex_unlock(drive_mutex);

  return self;
}


/* Here are the actual Apple ][ interface routines. */

/* Higher level Prodos stuff. */
- (BOOL)bootUp
{
  Disk               *d = &disk[0];


  drive = 0;

  [computer writeByte:0x2B :slot_number << 4];
  [computer writeByte:0x27 :8];
  [computer writeByte:0x41 :0];	/* Track 0 */
  [computer writeByte:0x3D :0];	/* Sector 0 */
  [computer writeShort:0x26 :0x800];	/* To 0x800 */

  return[self cardReadSector];
}


/* Read a sector from both DOS 3.3 and ProDOS disks */
- readDisk:(Disk *) d sector:(int)sector data:(unsigned char *)data
{
  int                 track;
  static const int    physicalSectorMap[] = {
					     0x0, 0x7, 0xE, 0x6, 0xD, 0x5, 0xC, 0x4,
					     0xB, 0x3, 0xA, 0x2, 0x9, 0x1, 0x8, 0xF,
  };
  static const int    prodosSectorMap[] = {
					   0x0, 0xe, 0xd, 0xc, 0xb, 0xa, 0x9, 0x8,
					   0x7, 0x6, 0x5, 0x4, 0x3, 0x2, 0x1, 0xf,
  };


  switch (d->type)
  {
  case DISK_DOS33:
    track = sector / 16;
    sector = physicalSectorMap[sector % 16];

    NXSeek(d->stream, d->u.dos33.first_track_offset + track * BYTES_PER_TRACK + sector * BYTES_PER_SECTOR, NX_FROMSTART);
    if (NXRead(d->stream, data, 256) != 256)
      return nil;

    break;

  case DISK_PRODOS:
    track = sector / 16;
    sector = prodosSectorMap[physicalSectorMap[sector % 16]];
    if (sector + track * SECTORS_PER_TRACK >= d->u.prodos.num_blocks * 2)
      return nil;

    NXSeek(d->stream, d->u.prodos.first_block_offset + track * BYTES_PER_TRACK + sector * BYTES_PER_SECTOR, NX_FROMSTART);
    if (NXRead(d->stream, data, 256) != 256)
      return nil;
    break;

  default:
    return nil;
  }

  return self;
}

- writeDisk:(Disk *) d sector:(int)sector data:(unsigned char *)data
{
  int                 track;
  static const int    physicalSectorMap[] = {
					     0x0, 0x7, 0xE, 0x6, 0xD, 0x5, 0xC, 0x4,
					     0xB, 0x3, 0xA, 0x2, 0x9, 0x1, 0x8, 0xF,
  };
  static const int    prodosSectorMap[] = {
					   0x0, 0xe, 0xd, 0xc, 0xb, 0xa, 0x9, 0x8,
					   0x7, 0x6, 0x5, 0x4, 0x3, 0x2, 0x1, 0xf,
  };


  switch (d->type)
  {
  case DISK_DOS33:
    track = sector / 16;
    sector = physicalSectorMap[sector % 16];

    if (d->write_protected)
      return nil;

    NXSeek(d->stream, d->u.dos33.first_track_offset + track * BYTES_PER_TRACK + sector * BYTES_PER_SECTOR, NX_FROMSTART);
    if (NXWrite(d->stream, data, 256) != 256)
      return nil;
    d->dirty = YES;
    break;

  case DISK_PRODOS:
    if (d->write_protected)
      return nil;

    track = sector / 16;
    sector = prodosSectorMap[physicalSectorMap[sector % 16]];
    if (sector + track * SECTORS_PER_TRACK >= d->u.prodos.num_blocks * 2)
      return nil;

    NXSeek(d->stream, d->u.prodos.first_block_offset + track * BYTES_PER_TRACK + sector * BYTES_PER_SECTOR, NX_FROMSTART);
    if (NXWrite(d->stream, data, 256) != 256)
      return nil;
    d->dirty = YES;
    break;

  default:
    return nil;
  }

  return self;
}

/* This is only an approximate implementation of this routine.  The goal
 * is to allow DOS 3.3 disks to boot without having to have the 6502 code
 * hit the disk drive soft switches directly.
 */
- (BOOL)cardReadSector
{
  int                 track, sector, addr;
  Disk               *d = &disk[drive];
  unsigned char       buf[256];


  if (d->type != DISK_DOS33 && d->type != DISK_PRODOS)
  {				/* Whoops, don't trap this...CLC and return */
    state->c_flag = 0;
    state->cycles += 2;
    state->pc++;
    return NO;
  }

  do
  {
  /* Get parms. */
    track = [computer readByte:0x41];
    if (track >= d->u.dos33.num_tracks)
      track = d->u.dos33.num_tracks - 1;
    sector = [computer readByte:0x3D];
    addr = [computer readShort:0x26];

/*    fprintf(stderr, "Reading %i %i\n", sector, track); */
    if (addr == 0x800 && sector == 0x0 && track == 0x0)
    {				/* FIXME */
      [self readDisk:d sector:track * SECTORS_PER_TRACK + (sector & 0xF) + 2 data:buf];
      [computer copyBytesToApple:buf :addr + 256:256];
    }
    [self readDisk:d sector:track * SECTORS_PER_TRACK + (sector & 0xF) data :buf];
    [computer copyBytesToApple:buf :addr :256];

    drive_state[drive].qtrack = track * 4;

    addr += 256,[computer writeByte:0x27 :addr >> 8];
    sector++,[computer writeByte:0x3D :sector];
  } while (sector < [computer readByte:0x800]);

/*  fprintf(stderr, "Jumping to 0x801\n"); */
  state->x = slot_number << 4;
  state->cycles += 1000;	/* Who knows?  More than this though... */
  state->pc = 0x801;

 /*
  * FIXME - this is a dubious thing to do!!!  Check for RWTS
  * first... 
  */
  [self addRoutineHandler:0xBD00 :0xBD00 :@selector(rwts)];

  return NO;
}

- (BOOL)deviceDriverCalled
{
  Disk               *d = &disk[[computer readByte:0x43] >> 7];
  unsigned char       buf[512];
  int                 block, sector;
  static const int    logicalSectorMap[] = {
					    0x0, 0x2, 0x4, 0x6, 0x8, 0xa, 0xc, 0xe,
					    0x1, 0x3, 0x5, 0x7, 0x9, 0xb, 0xd, 0xf,
  };


  switch ([computer readByte:0x42])
  {
  case PRODOS_STATUS:
    state->c_flag = 0;
    state->a = PRODOS_ERR_NO_ERROR;
    state->x = d->u.prodos.num_blocks % 256;
    state->y = d->u.prodos.num_blocks / 256;
    break;

  case PRODOS_READ:
    block = [computer readShort:0x46] * 2;
    sector = block % 16;
    block /= 16;
    if (![self readDisk:d sector:block * 16 + logicalSectorMap[sector] data:buf] ||
	![self readDisk:d sector:block * 16 + logicalSectorMap[sector + 1] data:&buf[BYTES_PER_SECTOR]])
    {
      state->c_flag = 1;
      state->a = PRODOS_ERR_IO_ERROR;
      break;
    }

    [computer copyBytesToApple:buf :[computer readShort:0x44] :512];
    state->c_flag = 0;
    state->a = PRODOS_ERR_NO_ERROR;
    break;

  case PRODOS_WRITE:
    if (d->write_protected)
    {
      state->c_flag = 1;
      state->a = PRODOS_ERR_WRITE_PROTECTED;
      break;
    }
    block = [computer readShort:0x46] * 2;
    sector = block % 16;
    block /= 16;

    [computer copyBytesFromApple:[computer readShort:0x44] :buf :512];

    if (![self writeDisk:d sector:block * 16 + logicalSectorMap[sector] data:buf] ||
	![self writeDisk:d sector:block * 16 + logicalSectorMap[sector + 1] data:&buf[BYTES_PER_SECTOR]])
    {
      state->c_flag = 1;
      state->a = PRODOS_ERR_IO_ERROR;
      break;
    }

    state->c_flag = 0;
    state->a = PRODOS_ERR_NO_ERROR;
    break;

  case PRODOS_FORMAT:
    if (d->write_protected)
    {
      state->c_flag = 1;
      state->a = PRODOS_ERR_WRITE_PROTECTED;
    }
    else
    {
      fprintf(stderr, "ProDOS FORMAT not yet supported!\n");
      state->c_flag = 1;
      state->a = PRODOS_ERR_IO_ERROR;
    }
    break;

  default:
    state->c_flag = 1;
    state->a = PRODOS_ERR_BAD_CALL;
  }

  return YES;
}

- (BOOL)rwts
{
  int                 iob_addr, track, sector, addr, volume_number = 254;
  Disk               *d;
  enum DosRWTSCall    cmd;
  unsigned char       iobuf[0x11], buf[256];
  static const int    logicalSectorMap[] = {
					    0x0, 0xd, 0xb, 0x9, 0x7, 0x5, 0x3, 0x1,
					    0xe, 0xc, 0xa, 0x8, 0x6, 0x4, 0x2, 0xf,
  };


 /* Write out temp stuff. */
  [computer writeByte:0x48 :state->y];
  [computer writeByte:0x49 :state->a];
  [computer writeByte:0x6F8 :2];
  [computer writeByte:0x4F8 :4];

 /* Get parameters from I/O control block. */
  iob_addr = state->y | (state->a << 8);
  [computer copyBytesFromApple:iob_addr :iobuf :0x11];
  track = iobuf[4];
  sector = iobuf[5];
  addr = iobuf[8] | (iobuf[9] << 8);
  cmd = iobuf[12];
  drive = (iobuf[2] == 2);

  d = &disk[drive];

 /* See if we are trying to modify a write-protected disk. */
  if ((cmd == DOS_WRITE || cmd == DOS_FORMAT) && d->write_protected)
  {
    state->c_flag = 1;
    [computer writeByte:iob_addr + 0xD:DOS_ERR_WRITE_PROTECTED];
    return YES;
  }

 /* Check for generic I/O error. */
  if (iobuf[0] != 1
      || track >= d->u.dos33.num_tracks
      || sector > 15
      || (cmd != DOS_SEEK && cmd != DOS_READ && cmd != DOS_WRITE
	  && cmd != DOS_FORMAT))
  {
    state->c_flag = 1;
    [computer writeByte:iob_addr + 0xD:DOS_ERR_IO_ERROR];
    return YES;
  }

  switch (cmd)
  {
  case DOS_SEEK:
    drive_state[drive].qtrack = track * 4;
  /* FIXME - what are stepper motor states here? */
    break;
  case DOS_READ:
    if (![self readDisk:d sector:track * SECTORS_PER_TRACK + logicalSectorMap[sector] data:buf])
    {
      state->c_flag = 1;
      [computer writeByte:iob_addr + 0xD:DOS_ERR_IO_ERROR];
      return YES;
    }
    [computer copyBytesToApple:buf :addr :256];
    break;
  case DOS_WRITE:
    [computer copyBytesFromApple:addr :buf :256];
    if (![self writeDisk:d sector:track * SECTORS_PER_TRACK + logicalSectorMap[sector] data:buf])
    {
      state->c_flag = 1;
      [computer writeByte:iob_addr + 0xD:DOS_ERR_IO_ERROR];
      return YES;
    }
    break;
  case DOS_FORMAT:
    memset(buf, 0, sizeof buf);
    NXSeek(d->stream, d->u.dos33.first_track_offset, NX_FROMSTART);
    for (track = 0; track < d->u.dos33.num_tracks; track++)
      for (sector = 0; sector < SECTORS_PER_TRACK; sector++)
	NXWrite(d->stream, buf, 256);
    NXFlush(d->stream);
    d->dirty = YES;
    break;
  }

 /* Write out common output stuff. */
  if (cmd != DOS_SEEK)
    [computer writeByte:iob_addr + 0xE:volume_number];
  [computer writeByte:iob_addr + 0xD:DOS_ERR_NO_ERROR];
  [computer writeByte:iob_addr + 0xF:slot_number * 16];
  [computer writeByte:iob_addr + 0x10:drive + 1];
  state->c_flag = 0;
  state->a = 0;			/* FIXME - is this necessary? */

  return YES;
}


- (int)stepperMotorReferenced:(AccessType) type :(MemoryElement *) m
{
  int                 phase = ((m - state->sysmem) >> 1) & 0x3;
  int                 on = (m - state->sysmem) & 1;
  static const int    table[] = {0, -1, 1};
  DriveState         *s = &drive_state[drive];
  int                 i;


  mutex_lock(drive_mutex);

 /* Set the magnet to its new state. */
  s->magnet[phase] = on;

 /*
  * See where the drive head should move to.  Here we assume the
  * drive head moves instantaneously...this should be fixed!
  * Perhaps when we do quarter tracks... :-) 
  */
  for (i = 0; i < sizeof table / sizeof table[0]; i++)
    if (s->magnet[((s->qtrack / 2) + 4 + table[i]) % 4])
    {
      s->qtrack = (s->qtrack & ~1) + table[i] * 2;

    /* Make sure we are on a valid track number. */
      if (s->qtrack < 0)
	s->qtrack = 0;
      else
      {
	int                 num_qtracks;

      /* Figure out the maximum # of qtracks for this disk. */
	if (disk[drive].type == DISK_DOS33)
	  num_qtracks = disk[drive].u.dos33.num_tracks * 4;
	else
	if (disk[drive].type == DISK_RAW_BYTES)
	  num_qtracks = disk[drive].u.raw_bytes.num_tracks * 4;
	else
	  num_qtracks = 35 * 4;	/* Default value. */

      /*
       * If we've moved the drive head too far, pin it to max
       * value. 
       */
	if (s->qtrack >= num_qtracks)
	  s->qtrack = num_qtracks - 2;
      }

      break;
    }

  mutex_unlock(drive_mutex);

  return 0;
}


- (int)setMotorOn:(AccessType) type :(MemoryElement *) m
{
  mutex_lock(drive_mutex);

  drive_state[drive].on = (m - state->sysmem) & 1;

  mutex_unlock(drive_mutex);
  return 0;
}


- (int)selectDrive:(AccessType) type :(MemoryElement *) m
{
  mutex_lock(drive_mutex);
  drive = (m - state->sysmem) & 1;
  mutex_unlock(drive_mutex);
  return 0;
}


- (int)handleDataLatch:(AccessType) type :(MemoryElement *) m
{
  Disk               *d;
  DriveState         *s;


  mutex_lock(drive_mutex);

  d = &disk[drive];
  s = &drive_state[drive];

  switch (m - state->sysmem - (slot_number * 0x10))
  {

  case 0xC08C:			/* Strobe Data Latch for I/O. */
  /* FIXME!  Only supported type for now. */
    if (d->type == DISK_NO_DISK || d->type != DISK_RAW_BYTES)
    {
      mutex_unlock(drive_mutex);
      return 0x00;		/* FIXME - should return randomness */
    }

    if (d->type == DISK_RAW_BYTES)
    {
      int                 track_length = d->u.raw_bytes.track[s->qtrack].track_length;

    /* FIXME - rotate disk depending on cycles elapsed here! */


    /* Get index for next byte. */
      if (track_length == 0)
      {
	mutex_unlock(drive_mutex);
	return 0xFF;
      }
      d->u.raw_bytes.head_position = ((d->u.raw_bytes.head_position + 1)
				      % track_length);

    /* Perform the actual I/O. */
      if (s->write_mode)
      {
	if (!d->write_protected)
	{
	/* Seek to current byte. */
	  NXSeek(d->stream, (d->u.raw_bytes.track[s->qtrack].track_offset
			     + d->u.raw_bytes.head_position),
		 NX_FROMSTART);

	/* Update cache, if appropriate. */
	  if (s->cached_qtrack_number == s->qtrack)
	    s->cached_qtrack_data[d->u.raw_bytes.head_position]
	      = s->write_register;

	/* "write-through" to the real file. */
	  NXPutc(d->stream, s->write_register);

	/* Remember that the disk has been changed. */
	  d->dirty = YES;
	}
	mutex_unlock(drive_mutex);
	return 0;
      }
      else
      {
	int                 v;

      /* Cache new track if necessary. */
	if (s->cached_qtrack_number != s->qtrack)
	{
	  int                 tlen = d->u.raw_bytes.track[s->qtrack].track_length;

	  NX_FREE(s->cached_qtrack_data);
	  NX_MALLOC(s->cached_qtrack_data, unsigned char, tlen);
	  NXSeek(d->stream, d->u.raw_bytes.track[s->qtrack].track_offset,
		 NX_FROMSTART);
	  NXRead(d->stream, s->cached_qtrack_data, tlen);
	  s->cached_qtrack_number = s->qtrack;
	}

	v = s->cached_qtrack_data[d->u.raw_bytes.head_position];
	mutex_unlock(drive_mutex);
	return v;
      }
    }
    break;

  case 0xC08D:			/* Load Data Latch. */
    drive_state[drive].write_register = m->mem.byte;
    break;

  case 0xC08E:			/* Prepare latch for input. */
    drive_state[drive].write_mode = NO;
    if (disk[drive].type != DISK_NO_DISK && disk[drive].write_protected)
    {
      mutex_unlock(drive_mutex);
      return 0x80;
    }
    break;

  case 0xC08F:			/* Prepare latch for output. */
    drive_state[drive].write_mode = YES;
    break;

  default:
  /* Oops!  We were called when we shouldn't have been! */
    abort();
  }

  mutex_unlock(drive_mutex);
  return 0;
}

@end
