
#import "IconWell.h"
#import <appkit/appkit.h>

@implementation IconWell

- initFrame:(const NXRect *)frameRect
{
  const char         *fileType[] = {NXFilenamePboardType};


  [super initFrame:frameRect];

  [self registerForDraggedTypes:fileType count:1];

  filename = NULL;
  types = NULL;
  delegate = nil;

  textField = [[TextField allocFromZone:[self zone]] initFrame:&bounds];
  [textField setAlignment:NX_CENTERED];
  [textField setSelectable:NO];
  [textField setBezeled:NO];
  [textField setBordered:NO];
  [textField setBackgroundTransparent:YES];
  [textField sizeToFit];

  iconWellFlags.multiFile = NO;
  iconWellFlags.canAccept = YES;
  iconWellFlags.acceptWhenFilled = YES;
  iconWellFlags.removeOnDrag = YES;
  iconWellFlags.bezeled = YES;

  return self;
}

- calcPosition:(NXRect *)aRect forImage:anImage withName:(const char *)aName
{
  aRect->origin.x = aRect->origin.y = aRect->size.width = aRect->size.height = 0.0;
  [anImage getSize:&aRect->size];
  aRect->origin.x = bounds.origin.x + (bounds.size.width - aRect->size.width) / 2;
  aRect->origin.y = bounds.origin.y + (bounds.size.height - aRect->size.height) / 2;

  return self;
}

- drawResult:image
{
  NXRect              aRect;


  aRect.origin.x = aRect.origin.y = aRect.size.width = aRect.size.height = 0.0;
  [newImage getSize:&aRect.size];
  PSsetalpha(NX_DKGRAY);
  PSsetgray(NX_WHITE);
  PScompositerect(aRect.origin.x, aRect.origin.y, aRect.size.width, aRect.size.height, NX_COPY);
  [newImage composite:NX_SOUT toPoint:&aRect.origin];

  return self;
}

- drawLight
{
  id                  result;
  NXRect              rect1, rect2;


  [self calcPosition:&rect1 forImage:newImage withName:NULL];
  [self calcPosition:&rect2 forImage:image withName:filename];
  [(result = [[NXImage allocFromZone:[self zone]] initSize:&rect1.size]) useDrawMethod :@selector(drawResult:) inObject :self];

  NXUnionRect(&rect1, &rect2);

  PSsetalpha(NX_WHITE);
  PSsetgray(NX_LTGRAY);
  PScompositerect(rect2.origin.x, rect2.origin.y, rect2.size.width, rect2.size.height, NX_COPY);
  [result composite:NX_SOVER toPoint:&rect1.origin];
  [window flushWindow];

  [result free];

  return self;
}

- drawNormal
{
  NXRect              rect1, rect2;


  [self calcPosition:&rect1 forImage:newImage withName:NULL];
  [self calcPosition:&rect2 forImage:image withName:filename];

  NXUnionRect(&rect2, &rect1);

  PSsetalpha(NX_WHITE);
  PSsetgray(NX_LTGRAY);
  PScompositerect(rect1.origin.x, rect1.origin.y, rect1.size.width, rect1.size.height, NX_COPY);
  [image composite:NX_SOVER toPoint:&rect2.origin];
  [window flushWindow];

  return self;
}

- drawSelf:(const NXRect *)rects :(int)rectCount
{
  if (iconWellFlags.bezeled)
    NXDrawGrayBezel(&bounds, NULL);

  [self drawNormal];

  return self;
}

- (NXDragOperation) draggingEntered:sender
{
  int                 i, j;
  char               *data;
  int                 len;
  id                  pb;
  const char         *fileType[] = {NXFilenamePboardType};
  NXDragOperation     op = NX_DragOperationPrivate;


  if (!iconWellFlags.canAccept || (!iconWellFlags.acceptWhenFilled && image))
    return NX_DragOperationNone;

  pb = [sender draggingPasteboard];
  if (![pb findAvailableTypeFrom:fileType num:1])
    return NX_DragOperationNone;

  [pb readType:NXFilenamePboardType data:&data length:&len];

  if (!iconWellFlags.multiFile)
  {				/* Scan for tab */
    for (i = 0; i < len; i++)
      if (data[i] == '\t')
      {
	op = NX_DragOperationNone;
	break;
      }
  }

 /* Make sure all files are of correct type */
  if (types && types[0])
    for (i = 0; i < len; i++)
      if (!data[i] || data[i] == '\t')
      {
	for (j = 0; types[j]; j++)
	  if (!strncmp(&data[i - strlen(types[j])], types[j], strlen(types[j])) && data[i - strlen(types[j]) - 1] == '.')
	    break;
	if (!types[j])
	{
	  op = NX_DragOperationNone;
	  break;
	}
      }

  if (op)
  {
    newImage = [[sender draggedImage] copyFromZone:[self zone]];

    if (delegate && [delegate respondsTo:@selector(iconWell:willUseImage:forFile:)])
      newImage = [delegate iconWell:self willUseImage:newImage forFile:data];

    [self lockFocus];
    [self drawLight];
    [self unlockFocus];
  }

  [pb deallocatePasteboardData:data length:len];

  return op;
}

- draggingExited:sender
{
  [self lockFocus];
  [self drawNormal];
  [self unlockFocus];
  [newImage free];
  newImage = nil;

  return self;
}

- (BOOL)prepareForDragOperation:sender
{
  NXRect              aRect;
  id                  oldImage = image;


  image = newImage;
  [window disableFlushWindow];

  [self lockFocus];
  [self drawNormal];
  [self unlockFocus];

  [self calcPosition:&aRect forImage:newImage withName:NULL];
  [self convertPoint:&aRect.origin toView:nil];
  [window convertBaseToScreen:&aRect.origin];
  [sender slideDraggedImageTo:&aRect.origin];

  [window reenableFlushWindow];

  image = oldImage;

  return YES;
}

- (BOOL)performDragOperation:sender
{
  char               *data;
  int                 len;
  id                  pb;
  const char         *fileType[] = {NXFilenamePboardType};


  pb = [sender draggingPasteboard];
  if (![pb findAvailableTypeFrom:fileType num:1])
    return NO;
  [pb readType:NXFilenamePboardType data:&data length:&len];

  if (filename)
    free(filename);
  filename = NXCopyStringBuffer(data);

  if (image)
    [image free];
  image = newImage;		/* I already made a copy */
  newImage = nil;

  [pb deallocatePasteboardData:data length:len];

  if (delegate && [delegate respondsTo:@selector(iconWell:didAcceptFile:)])
    [delegate iconWell:self didAcceptFile:filename];

  return YES;
}

- (NXDragOperation) draggingSourceOperationMaskForLocal:(BOOL)isLocal
{
  return NX_DragOperationAll;
}

- (BOOL)acceptsFirstMouse
{
  return YES;
}

- (BOOL)shouldDelayWindowOrderingForEvent:(NXEvent *)anEvent
{
  return YES;
}

- mouseDown:(NXEvent *)theEvent
{
  NXRect              aRect;
  Pasteboard         *pboard = [Pasteboard newName:NXDragPboard];
  NXEvent             anEvent = *theEvent;
  const char         *fileType[] = {NXFilenamePboardType};
  int                 oldMask;
  BOOL                oldAccept = iconWellFlags.acceptWhenFilled;
  id                  oldImage;


  oldMask = [window addToEventMask:NX_MOUSEDRAGGEDMASK];

  if (!image)
    return self;

  do
  {
    theEvent = [NXApp getNextEvent:NX_MOUSEUPMASK | NX_MOUSEDRAGGEDMASK];
    if (theEvent->type = NX_MOUSEDRAGGED)
    {
      oldImage = image;

      [self calcPosition:&aRect forImage:oldImage withName:filename];
      [oldImage getSize:&aRect.size];
      aRect.origin.x = bounds.origin.x + (bounds.size.width - aRect.size.width) / 2;
      aRect.origin.y = bounds.origin.y + (bounds.size.height - aRect.size.height) / 2;

      if (!iconWellFlags.removeOnDrag)
	iconWellFlags.acceptWhenFilled = NO;
      else
	newImage = image = nil;

      [pboard declareTypes:fileType num:1 owner:self];
      if (filename)
	[pboard writeType:NXFilenamePboardType data:filename length:strlen(filename) + 1];
      else
	[pboard writeType:NXFilenamePboardType data:"" length:1];

      [self dragImage:oldImage at:&aRect.origin offset:NULL event:&anEvent pasteboard:pboard source:self slideBack:!iconWellFlags.removeOnDrag];

      if (!image)
      {
	if (delegate && [delegate respondsTo:@selector(iconWell:didRemoveFile:)])
	  [delegate iconWell:self didRemoveFile:filename];
	if (filename)
	  free(filename);
	if (oldImage)
	  [oldImage free];
      }

      break;
    }
  } while (theEvent->type != NX_MOUSEUP);

  [window setEventMask:oldMask];
  iconWellFlags.acceptWhenFilled = oldAccept;

  return self;
}

- setTypes:(const char *const *)filetypes
{
  int                 i;


  if (types)
  {
    for (i = 0; types[i]; i++)
      free(types[i]);
    free(types);
  }

  for (i = 0; filetypes[i]; i++);
  types = malloc((i + 1) * sizeof(const char *));
  for (i = 0; filetypes[i]; i++)
    types[i] = NXCopyStringBuffer(filetypes[i]);
  types[i] = NULL;

  return self;
}

- setAllowMultipleFiles:(BOOL)allow
{
  iconWellFlags.multiFile = allow;
  return self;
}

- setCanAccept:(BOOL)can
{
  iconWellFlags.canAccept = can;
  return self;
}

- setAcceptWhenFilled:(BOOL)accept
{
  iconWellFlags.acceptWhenFilled = accept;
  return self;
}

- setRemoveOnDrag:(BOOL)remove
{
  iconWellFlags.removeOnDrag = remove;
  return self;
}

- setFile:(const char *)aFile
{
  if (filename)
    free(filename);

  filename = NXCopyStringBuffer(aFile);
  image = [[Application workspace] getIconForFile:filename];
  [self display];

  return self;
}

- setImage:anImage
{
  if (image)
    [image free];

  image = [anImage copyFromZone:[self zone]];
  [self display];

  return self;
}

- setFilename:(const char *)aName
{
  if (filename)
    free(filename);
  filename = NXCopyStringBuffer(aName);
  return self;
}

@end
