/*
 * This file is part of zaniWok.
 *
 * zaniWok is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.  See the file COPYING for more information.
 *
 * zaniWok is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include "opcodes.h"
#include "memory.h"
#include "execute.h"
#include "translate.h"
#include "compile.h"
#include <stdio.h>

/* Bits in the value returned by compile_instruction. */
#define NEEDS_C      CBIT
#define NEEDS_N      NBIT
#define NEEDS_V      VBIT
#define NEEDS_Z      ZBIT
#define NEEDS_ALL    (NEEDS_C | NEEDS_N | NEEDS_V | NEEDS_Z)


/* Returns some combination of the above flags indicating which cc bits
 * must be valid on entry to the opcode it just compiled.
 */

int
compile_instruction(unsigned short pc, unsigned short parent_pc)
{


/* WRITE ME!
 * A solid performance enhancement for the emulator has not yet been
 * implemented.  The "interpreter" can handle opcodes that generate
 * fewer than all CC bits, but the "compiler" does not currently use
 * this feature; instead, it assumes that all instructions require all
 * CC bits to be valid.  I am not sure what sort of performance will
 * be gained by implementing this.
 *
 * A warning to anyone who tries to do this: beware of self-modifying
 * code!  I have a plan to deal with this.
 */

  return 0;
}


int
decompile_instruction(MachineState * state, MemoryElement * m)
{
  int                 i;

  for (i = 2; i >= 0; m--, i--)
    if (m->inst.opcode != OPC_TRAPTHIS)
    {
      m->inst.opcode = OPC_NOT_COMPILED;
      if (m->mem.write_func == HANDLE_MODIFIED_CODE_FUNCNUM)
	m->mem.write_special = 0;
    }

  return 0;
}




/* I just wrote this to help me figure out Wizardry's copy protection,
 * but I'll leave it here for future generations...
 */

static char        *
disassemble_memory_element(const MemoryElement * m, int pc, char *result)
{
  const TranslateInfo *t = &translate_info[m->mem.byte];
  unsigned int        c = m[1].mem.byte;
  unsigned int        s = m[1].mem.byte | (m[2].mem.byte << 8);

  switch ((AddressingMode) t->addressing_mode)
  {
  case AM_IMPLICIT:
    strcpy(result, t->opcode_text);
    break;
  case AM_IMM:
    sprintf(result, "%s\t#$%02X", t->opcode_text, c);
    break;
  case AM_ZP:
    sprintf(result, "%s\t$%02X", t->opcode_text, c);
    break;
  case AM_ZP_X:
    sprintf(result, "%s\t$%02X,X", t->opcode_text, c);
    break;
  case AM_ZP_Y:
    sprintf(result, "%s\t$%02X,Y", t->opcode_text, c);
    break;
  case AM_IND:
    sprintf(result, "%s\t($%02X)", t->opcode_text, c);
    break;
  case AM_IND_X:
    sprintf(result, "%s\t($%02X,X)", t->opcode_text, c);
    break;
  case AM_IND_Y:
    sprintf(result, "%s\t($%02X),Y", t->opcode_text, c);
    break;
  case AM_ABS:
    sprintf(result, "%s\t$%04X", t->opcode_text, s);
    break;
  case AM_ABS_X:
    sprintf(result, "%s\t$%04X,X", t->opcode_text, s);
    break;
  case AM_ABS_Y:
    sprintf(result, "%s\t$%04X,Y", t->opcode_text, s);
    break;
  case AM_ABSIND:
    sprintf(result, "%s\t($%04X)", t->opcode_text, s);
    break;
  case AM_ABSIND_X:
    sprintf(result, "%s\t($%04X,X)", t->opcode_text, s);
    break;
  case AM_RELATIVE:
    sprintf(result, "%s\t$%04X", t->opcode_text,
	    (unsigned)(pc + 2 + (signed char)c) & 0xFFFF);
    break;
  }

  return result;
}


static void
disassemble_many_opcodes(FILE * fp, int pc, MemoryElement * m, int how_many)
{
  char                buf[256];
  int                 n;

  for (n = 0; how_many > 0; n++, how_many--)
  {
    const TranslateInfo *t = &translate_info[m->mem.byte];

    printf("%04X:\t%s\t", (unsigned)pc,
	   disassemble_memory_element(m, pc, buf));

    putc('\n', fp);

    pc += t->operand_bytes + 1;
    m += t->operand_bytes + 1;
  }
}


/* Compiles an opcode.  If state is NULL any read traps are ignored. */

void
handle_uncompiled_opcode(MachineState * state)
{
  unsigned char       b1, b2, b3;
  MemoryElement      *m1, *m2, *m3;
  const TranslateInfo *t;
  int                 pc = state->pc;
  static const unsigned char page_segment[] = {
					       0, 0, 1, 1, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1,	/* 0x0000 - 0x0FFF */
					       1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,	/* 0x1000 - 0x1FFF */
					       3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,	/* 0x2000 - 0x2FFF */
					       3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,	/* 0x3000 - 0x3FFF */
					       1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,	/* 0x4000 - 0x4FFF */
					       1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,	/* 0x5000 - 0x5FFF */
					       1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,	/* 0x6000 - 0x6FFF */
					       1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,	/* 0x7000 - 0x7FFF */
					       1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,	/* 0x8000 - 0x8FFF */
					       1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,	/* 0x9000 - 0x9FFF */
					       1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,	/* 0xA000 - 0xAFFF */
					       1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,	/* 0xB000 - 0xBFFF */
					       7, 8, 9, 10, 11, 12, 13, 14, 6, 6, 6, 6, 6, 6, 6, 6,	/* 0xC000 - 0xCFFF */
					       4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,	/* 0xD000 - 0xDFFF */
					       5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,	/* 0xE000 - 0xEFFF */
					       5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,	/* 0xF000 - 0xFFFF */
					       0
  };

 /*
  * Fetch the MemoryElement containing the opcode and the value
  * therein. 
  */
  m1 = get_memory_element(state, pc, READ);
  b1 = read_memory_element(state, m1);

  t = &translate_info[b1];
  if (!m1->mem.write_special)
  {
    m1->mem.write_special = 1;
    m1->mem.write_func = HANDLE_MODIFIED_CODE_FUNCNUM;
  }

  m1->inst.opcode = t->base_opcode | ALLBITS;

  if (t->operand_bytes == 1)
  {
    m2 = get_memory_element(state, pc + 1, READ);
    b2 = read_memory_element(state, m2);

    if (t->is_relative_branch)
    {
      m1->inst.u.s = (pc + 2 + ((signed char)b2)) & 65535;
    }
    else
      m1->inst.u.s = b2;	/* FIXME - could span banks! */
    if (!m2->mem.write_special)
    {
      m2->mem.write_special = 1;
      m2->mem.write_func = HANDLE_MODIFIED_CODE_FUNCNUM;
    }
  }
  else
  if (t->operand_bytes == 2)
  {
    m2 = get_memory_element(state, pc + 1, READ);
    b2 = read_memory_element(state, m2);
    m3 = get_memory_element(state, pc + 2, READ);
    b3 = read_memory_element(state, m3);

    m1->inst.u.s = b2 | (b3 << 8);
    if (!m2->mem.write_special)
    {
      m2->mem.write_special = 1;
      m2->mem.write_func = HANDLE_MODIFIED_CODE_FUNCNUM;
    }
    if (!m3->mem.write_special)
    {
      m3->mem.write_special = 1;
      m3->mem.write_func = HANDLE_MODIFIED_CODE_FUNCNUM;
    }

  /*
   * If this instruction is a "read" instruction, and there is no
   * way it could possibly touch a $C000 address, change the
   * opcode to the "safe" variant that doesn't even check. 
   */
    if ((t->abs_mode_and_is_safe && (m1->inst.u.s & 0xC000) != 0xC000)
	|| (t->abs_xy_mode_and_is_safe
	    && (m1->inst.u.s <= 0xBF00 || m1->inst.u.s >= 0xD000)))
      m1->inst.opcode += BLOCKSIZE;
  }

 /*
  * Figure out if this instruction's child could be in
  * non-contiguous memory 
  */
  if (((t->is_normal_instr || t->is_relative_branch)
       && ((page_segment[(pc + 1 + t->operand_bytes) >> 8])
	   != page_segment[pc >> 8]))
      || ((t->is_relative_branch || (m1->inst.opcode & ~ALLBITS) == OPC_JMP_ABS
	   || (m1->inst.opcode & ~ALLBITS) == OPC_JSR)
	  && page_segment[pc >> 8] != page_segment[m1->inst.u.s >> 8]))
    m1->inst.opcode |= NEWBANK;
}


int
handle_modified_opcode(AccessType rw, MemoryElement * m, MachineState * state)
{
  return decompile_instruction(state, m);
}
