/*
 * This file is part of zaniWok.
 *
 * zaniWok is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.  See the file COPYING for more information.
 *
 * zaniWok is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

/*
 * TODO:
 *
 * 1) Try a.l instead of A, x.l instead of X, etc. as arg to _COMP.
 * 2) Write DEX_BNE_N, INY_BNE_ALL, etc. for faster looping.
 * 3) Check BCD behavior; odd cases are almost certainly wrong, but I don't
 *    have any documentation.  Need to play with a real machine...
 */

#define DEBUG 1

#include "execute.h"
#include "opcodes.h"
#include "bcd.h"
#include "compile.h"
#include "memory.h"
#include <stddef.h>
#include <stdio.h>
#include <assert.h>
#include "execute_private.h"


void
exec(MachineState * state)
{
  register Instruction i;
  register Reg        pc;
  register Reg        a, x, y;
  Reg                 sp;
  MemoryElement     **read_base, **write_base;
  void               *special_return_addr;
  int                 special_return_value;
  struct MemInfo     *special_arg;
  int                 n;

 /*
  * This next struct is because we are really hurting for
  * registers... 
  */
  register struct
  {
    unsigned char       v_flag, n_flag, nz_flag, c_flag;
  }                   cc_bits;
  unsigned char       b_flag, d_flag, i_flag, unused;
  unsigned long       cycles;
  register void      *pc_base;
  register const void **jump_table_reg;


#include "jmptable.h"


  jump_table_reg = &jump_table[0];


/* NOTE: #defining ASM_COMMENT to actually generate an asm statement will
 * force gcc into making less-than-optimal code to accomodate it!
 */
#define ASM_COMMENT(comm)	/* asm ("\t; " comm) */


#define HANDLER(x) \
handle_ ## x: \
{ \
    unsigned char temp_uchar; \
    ASM_COMMENT (#x)
#define END } continue

 /* Set up the machine in the appropriate initial state. */
  a.l = x.l = y.l = pc.l = sp.l = 0;	/* Clear out high bytes of reg's */
  LOAD_MACHINE_STATE();
  RELOAD_PC_BASE();

#if 0
  printf("pc_base = %p, read_base = %p, write_base = %p, pc = 0x%X, "
	 "read_base[pc.l >> 8] = %p, pc.l & ~0xFFL = 0x%X\n",
	 pc_base, read_base, write_base, (unsigned int)pc.l,
	 read_base[pc.l >> 8], pc.l & ~0xFFL);
#endif

  while (1)
  {
#ifdef DEBUG
    if (state->debug)
      printf("pc = 0x%04X, a = %02X, x = %02X, y = %02X, "
	     "sp = %02X, cnvzbdi=%d%d%d%d%d%d%d\n",
	     pc.l, A, X, Y, SP, !!CFLAG, !!NFLAG, !!VFLAG,
	     !NZFLAG, !!BFLAG, !!DFLAG, !!IFLAG);
#endif

  /* Dispatch this instruction. */
    i = *(Instruction *) (pc_base + (pc.l * sizeof(MemoryElement)));
    goto *jump_table_reg[i.opcode];


  /*
   * These handle the cases where an instruction may be going to
   * another instruction whose corresponding memory element does
   * not contiguously follow the current one because of the way
   * banks can be switched in and out.  Don't worry about it, it
   * works. 
   */
newbank_pc_plus1:
    PC += 1, RELOAD_PC_BASE();
    PC -= 1;
    goto *jump_table_reg[i.opcode & ~NEWBANK];

newbank_pc_plus2:
    PC += 2, RELOAD_PC_BASE();
    PC -= 2;
    goto *jump_table_reg[i.opcode & ~NEWBANK];

newbank_pc_plus3:
    PC += 3, RELOAD_PC_BASE();
    PC -= 3;
    goto *jump_table_reg[i.opcode & ~NEWBANK];


  /* Oops.  Save machine state and punt. */
    HANDLER(OPC_INVALID);
    SAVE_MACHINE_STATE();
    return;
    END;

    HANDLER(OPC_NOT_COMPILED);
    state->pc = PC;
    handle_uncompiled_opcode(state);
    END;

    HANDLER(OPC_TRAPTHIS);
    {
      BOOL                do_rts;
      MemoryElement      *old_elt = ((MemoryElement *)
				     (pc_base - offsetof(MemoryElement, inst)
				      + (pc.l * sizeof(MemoryElement))));
      MemoryElement      *new_elt;
      unsigned short      trap_num = USHORTARG;

    /*
     * This used to be extremely complicated and have many
     * different types of handlers, some providing less
     * information (and thereby being faster) than others.  I
     * have decided that the confusion and complexity added by
     * having many different types of handlers greatly outweighs
     * the minor cost of unnecessary saves/loads of 6502
     * registers and CC flags. 
     */
      SAVE_MACHINE_STATE();
      do_rts = state->routine_func_list[trap_num] (state);
      LOAD_MACHINE_STATE();
      if (do_rts)
	goto handle_OPC_RTS;

    /* PC or memory map may have changed. */
      RELOAD_PC_BASE();

    /*
     * See if we are executing the same instruction we trapped on
     * before, the trap is still enabled and it's exactly the
     * same trap. 
     */
      new_elt = (MemoryElement *) (pc_base - offsetof(MemoryElement, inst)
				   + (pc.l * sizeof(MemoryElement)));
      if (old_elt == new_elt && new_elt->inst.opcode == OPC_TRAPTHIS
	  && new_elt->inst.u.s == trap_num)
      {
	MemoryElement       temp;

      /* Save the old MemoryElement. */
	temp = *new_elt;

      /* Compile the instruction at this address. */
	new_elt->inst.opcode = OPC_NOT_COMPILED;
	new_elt->mem.write_special = 0;
	state->pc = PC;
	handle_uncompiled_opcode(state);

      /* Get the opcode we compiled to. */
	i = new_elt->inst;

      /* Restore the original opcode at this address. */
	*new_elt = temp;

      /* Execute the new instruction. */
	goto *jump_table_reg[i.opcode];
      }
    }
    END;

    HANDLER(OPC_CARDTRAP);
    {
      int                 do_rts;

      SAVE_MACHINE_STATE();

/* Because of a gcc 2.2.2 bug I am forced to #ifdef out the "correct" code.
 * The problem is that jmptable.h will not be correctly parsed in objective-c
 * mode.  Without being in objc mode I can't send the following message.
 * I reported the bug to GNU 8/3/92.
 * A workaround follows...
 *
 * A friend fixed this 8/16/92 and submitted it to GNU.  Since I have a working
 * version, I'll put in the cleaner code.
 */
#if NX_CURRENT_COMPILER_RELEASE > 310
      do_rts = (int)[state->peripheral_card[USHORTARG & 7] perform
		     :state->card_selector_list[USHORTARG >> 3]];
#else				/* FIXME!  Well, fix gcc. */
      do_rts = (int)objc_msgSend(state->peripheral_card[USHORTARG & 7],
				 state->card_selector_list[USHORTARG >> 3]);
#endif
      LOAD_MACHINE_STATE();
      if (do_rts)
	goto handle_OPC_RTS;
      RELOAD_PC_BASE();
    }
    END;

#define _SUMBASE(op, mode, cyc, size) \
  if (DFLAG) \
    { \
      int tmp = (bcd_to_binary[a.l] op bcd_to_binary[read (mode)] \
		 + CFLAG - ((1 op (-1)) / 2)); \
      A = (binary_to_bcd + 100)[tmp];\
    }\
  else A = (A op read (mode) + CFLAG - ((1 op (-1)) / 2))

#define _SUM(op, mode, cyc, size) \
  _SUMBASE (op, mode, cyc, size); \
  ADDCYC (cyc); \
  PC += size

#define _SUM_N(op, mode, cyc, size) \
  _SUMBASE (op, mode, cyc, size); \
  SET_N (A); \
  ADDCYC (cyc); \
  PC += size

#define _SUM_Z(op, mode, cyc, size) \
  _SUMBASE (op, mode, cyc, size); \
  SET_Z (A); \
  ADDCYC (cyc); \
  PC += size

#define _SUM_V(op, mode, cyc, size) \
    {\
      unsigned char tmp; \
      temp_uchar = read (mode); \
      tmp = (temp_uchar ^ A ^ 0x80) & 0x80;  /* tmp = 0x80 iff signs same */ \
      _SUMBASE (op, TEMPUCHAR, cyc, size); \
      if (!DFLAG) \
        VFLAG = tmp & (temp_uchar ^ A); /* 1 iff signs == but sum sign != */\
      ADDCYC (cyc); \
      PC += size; \
    }

#define _SUMBASE_C(op, mode, cyc, size) \
  if (DFLAG) \
    { \
      int tmp = (bcd_to_binary[a.l] op bcd_to_binary[read (mode)] \
		 + CFLAG - ((1 op (-1)) / 2)); \
      if (#op[0] == '-') CFLAG = (tmp >= 0); \
      else CFLAG = (tmp >= 100); \
      A = (binary_to_bcd + 100)[tmp];\
    }\
  else\
    {\
      unsigned short tmp;\
      A = tmp = A op read (mode) + CFLAG - ((1 op (-1)) / 2);\
      CFLAG = ((1 op -1) / 2) ^ (tmp >> 8);\
      if (#op[0] == '-') \
	CFLAG &= 0x1; \
    }

#define _SUM_C(op, mode, cyc, size) \
  _SUMBASE_C (op, mode, cyc, size); \
  ADDCYC (cyc); \
  PC += size

#define _SUM_ALL(op, mode, cyc, size) \
    {\
      unsigned char tmp; \
      temp_uchar = read (mode); \
      tmp = (temp_uchar ^ A ^ 0x80) & 0x80; /* tmp = 0x80 iff signs same */ \
      _SUMBASE_C (op, TEMPUCHAR, cyc, size); \
      if (!DFLAG) \
        VFLAG = tmp & (temp_uchar ^ A);/* 1 iff signs == but sum sign diff */\
      SET_N (A); \
      SET_Z (A); \
      ADDCYC (cyc); \
      PC += size; \
    }



#define _SUMBASESAFE(op, mode, cyc, size) \
  if (DFLAG) \
    { \
      int tmp = (bcd_to_binary[a.l] op bcd_to_binary[readsafe (mode)] \
		 + CFLAG - ((1 op (-1)) / 2)); \
      if (#op[0] == '-') CFLAG = (tmp >= 0); \
      else CFLAG = (tmp >= 100); \
      A = (binary_to_bcd + 100)[tmp];\
    }\
  else A = (A op readsafe (mode) + CFLAG - ((1 op (-1)) / 2))

#define _SUMSAFE(op, mode, cyc, size) \
  _SUMBASESAFE(op, mode, cyc, size); \
  ADDCYC (cyc); \
  PC += size

#define _SUMSAFE_N(op, mode, cyc, size) \
  _SUMBASESAFE (op, mode, cyc, size); \
  SET_N (A); \
  ADDCYC (cyc); \
  PC += size

#define _SUMSAFE_Z(op, mode, cyc, size) \
  _SUMBASESAFE (op, mode, cyc, size); \
  SET_Z (A); \
  ADDCYC (cyc); \
  PC += size

#define _SUMSAFE_V(op, mode, cyc, size) \
    {\
      unsigned char tmp; \
      temp_uchar = readsafe (mode); \
      tmp = (temp_uchar ^ A ^ 0x80) & 0x80;  /* tmp = 0x80 iff signs same */ \
      _SUMBASESAFE (op, TEMPUCHAR, cyc, size); \
      if (!DFLAG) \
        VFLAG = tmp & (temp_uchar ^ A);/* 1 iff signs == but sum sign diff */\
      ADDCYC (cyc); \
      PC += size; \
    }

#define _SUMBASESAFE_C(op, mode, cyc, size) \
  if (DFLAG) \
    { \
      int tmp = (bcd_to_binary[a.l] op bcd_to_binary[readsafe(mode)]\
		 + CFLAG - ((1 op (-1)) / 2)); \
      if (#op[0] == '-') CFLAG = (tmp >= 0); \
      CFLAG = (tmp >= 100); \
      A = (binary_to_bcd + 100)[tmp];\
    }\
  else\
    {\
      unsigned short tmp;\
      A = tmp = A op readsafe (mode) + CFLAG - ((1 op (-1)) / 2);\
      CFLAG = ((1 op -1) / 2) ^ (tmp >> 8);\
      if (#op[0] == '-') /* gcc does the right thing, don't worry */ \
	CFLAG &= 0x1; \
    }

#define _SUMSAFE_C(op, mode, cyc, size) \
  _SUMBASESAFE_C (op, mode, cyc, size); \
  ADDCYC (cyc); \
  PC += size

#define _SUMSAFE_ALL(op, mode, cyc, size) \
    {\
      unsigned char tmp; \
      temp_uchar = readsafe (mode); \
      tmp = (temp_uchar ^ A ^ 0x80) & 0x80; /* tmp = 0x80 iff signs same */ \
      _SUMBASESAFE_C (op, TEMPUCHAR, cyc, size); \
      if (!DFLAG) \
        VFLAG = tmp & (temp_uchar ^ A);/* 1 iff signs == but sum sign diff */\
      SET_N (A); \
      SET_Z (A); \
      ADDCYC (cyc); \
      PC += size; \
    }


  /* ADC */
#ifdef USE_CC_VARIANTS
#define _ADC(mode, cyc, size)   _SUM     (+, mode, cyc, size)
#define _ADC_C(mode, cyc, size) _SUM_C   (+, mode, cyc, size)
#define _ADC_N(mode, cyc, size) _SUM_N   (+, mode, cyc, size)
#define _ADC_Z(mode, cyc, size) _SUM_Z   (+, mode, cyc, size)
#define _ADC_V(mode, cyc, size) _SUM_V   (+, mode, cyc, size)
#endif
#define _ADC_ALL(mode,cyc,size) _SUM_ALL (+, mode, cyc, size)

#ifdef USE_CC_VARIANTS
#define _ADCSAFE(mode, cyc, size)   _SUMSAFE     (+, mode, cyc, size)
#define _ADCSAFE_C(mode, cyc, size) _SUMSAFE_C   (+, mode, cyc, size)
#define _ADCSAFE_N(mode, cyc, size) _SUMSAFE_N   (+, mode, cyc, size)
#define _ADCSAFE_Z(mode, cyc, size) _SUMSAFE_Z   (+, mode, cyc, size)
#define _ADCSAFE_V(mode, cyc, size) _SUMSAFE_V   (+, mode, cyc, size)
#endif
#define _ADCSAFE_ALL(mode,cyc,size) _SUMSAFE_ALL (+, mode, cyc, size)

  /* ADC #immediate */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_ADC_IMM);
    _ADC(IMM, 2, 2);
    END;
    HANDLER(OPC_ADC_IMM_C);
    _ADC_C(IMM, 2, 2);
    END;
    HANDLER(OPC_ADC_IMM_N);
    _ADC_N(IMM, 2, 2);
    END;
    HANDLER(OPC_ADC_IMM_V);
    _ADC_V(IMM, 2, 2);
    END;
    HANDLER(OPC_ADC_IMM_Z);
    _ADC_Z(IMM, 2, 2);
    END;
#endif
    HANDLER(OPC_ADC_IMM_ALL);
    _ADC_ALL(IMM, 2, 2);
    END;

  /* ADC ZP */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_ADC_ZP);
    _ADC(ZP, 3, 2);
    END;
    HANDLER(OPC_ADC_ZP_C);
    _ADC_C(ZP, 3, 2);
    END;
    HANDLER(OPC_ADC_ZP_N);
    _ADC_N(ZP, 3, 2);
    END;
    HANDLER(OPC_ADC_ZP_V);
    _ADC_V(ZP, 3, 2);
    END;
    HANDLER(OPC_ADC_ZP_Z);
    _ADC_Z(ZP, 3, 2);
    END;
#endif
    HANDLER(OPC_ADC_ZP_ALL);
    _ADC_ALL(ZP, 3, 2);
    END;

  /* ADC ZP,x */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_ADC_ZP_X);
    _ADC(ZPX, 4, 2);
    END;
    HANDLER(OPC_ADC_ZP_X_C);
    _ADC_C(ZPX, 4, 2);
    END;
    HANDLER(OPC_ADC_ZP_X_N);
    _ADC_N(ZPX, 4, 2);
    END;
    HANDLER(OPC_ADC_ZP_X_V);
    _ADC_V(ZPX, 4, 2);
    END;
    HANDLER(OPC_ADC_ZP_X_Z);
    _ADC_Z(ZPX, 4, 2);
    END;
#endif
    HANDLER(OPC_ADC_ZP_X_ALL);
    _ADC_ALL(ZPX, 4, 2);
    END;

  /* ADC ABS */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_ADC_ABS);
    _ADC(ABS, 4, 3);
    END;
    HANDLER(OPC_ADC_ABS_C);
    _ADC_C(ABS, 4, 3);
    END;
    HANDLER(OPC_ADC_ABS_N);
    _ADC_N(ABS, 4, 3);
    END;
    HANDLER(OPC_ADC_ABS_V);
    _ADC_V(ABS, 4, 3);
    END;
    HANDLER(OPC_ADC_ABS_Z);
    _ADC_Z(ABS, 4, 3);
    END;
#endif
    HANDLER(OPC_ADC_ABS_ALL);
    _ADC_ALL(ABS, 4, 3);
    END;

  /* ADC ABS,x */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_ADC_ABS_X);
    _ADC(ABSX, 4, 3);
    END;
    HANDLER(OPC_ADC_ABS_X_C);
    _ADC_C(ABSX, 4, 3);
    END;
    HANDLER(OPC_ADC_ABS_X_N);
    _ADC_N(ABSX, 4, 3);
    END;
    HANDLER(OPC_ADC_ABS_X_V);
    _ADC_V(ABSX, 4, 3);
    END;
    HANDLER(OPC_ADC_ABS_X_Z);
    _ADC_Z(ABSX, 4, 3);
    END;
#endif
    HANDLER(OPC_ADC_ABS_X_ALL);
    _ADC_ALL(ABSX, 4, 3);
    END;

  /* ADC ABS,Y */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_ADC_ABS_Y);
    _ADC(ABSY, 4, 3);
    END;
    HANDLER(OPC_ADC_ABS_Y_C);
    _ADC_C(ABSY, 4, 3);
    END;
    HANDLER(OPC_ADC_ABS_Y_N);
    _ADC_N(ABSY, 4, 3);
    END;
    HANDLER(OPC_ADC_ABS_Y_V);
    _ADC_V(ABSY, 4, 3);
    END;
    HANDLER(OPC_ADC_ABS_Y_Z);
    _ADC_Z(ABSY, 4, 3);
    END;
#endif
    HANDLER(OPC_ADC_ABS_Y_ALL);
    _ADC_ALL(ABSY, 4, 3);
    END;

  /* ADCSAFE ABS */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_ADCSAFE_ABS);
    _ADCSAFE(ABS, 4, 3);
    END;
    HANDLER(OPC_ADCSAFE_ABS_C);
    _ADCSAFE_C(ABS, 4, 3);
    END;
    HANDLER(OPC_ADCSAFE_ABS_N);
    _ADCSAFE_N(ABS, 4, 3);
    END;
    HANDLER(OPC_ADCSAFE_ABS_V);
    _ADCSAFE_V(ABS, 4, 3);
    END;
    HANDLER(OPC_ADCSAFE_ABS_Z);
    _ADCSAFE_Z(ABS, 4, 3);
    END;
#endif
    HANDLER(OPC_ADCSAFE_ABS_ALL);
    _ADCSAFE_ALL(ABS, 4, 3);
    END;

  /* ADCSAFE ABS,x */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_ADCSAFE_ABS_X);
    _ADCSAFE(ABSX, 4, 3);
    END;
    HANDLER(OPC_ADCSAFE_ABS_X_C);
    _ADCSAFE_C(ABSX, 4, 3);
    END;
    HANDLER(OPC_ADCSAFE_ABS_X_N);
    _ADCSAFE_N(ABSX, 4, 3);
    END;
    HANDLER(OPC_ADCSAFE_ABS_X_V);
    _ADCSAFE_V(ABSX, 4, 3);
    END;
    HANDLER(OPC_ADCSAFE_ABS_X_Z);
    _ADCSAFE_Z(ABSX, 4, 3);
    END;
#endif
    HANDLER(OPC_ADCSAFE_ABS_X_ALL);
    _ADCSAFE_ALL(ABSX, 4, 3);
    END;

  /* ADCSAFE ABS,Y */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_ADCSAFE_ABS_Y);
    _ADCSAFE(ABSY, 4, 3);
    END;
    HANDLER(OPC_ADCSAFE_ABS_Y_C);
    _ADCSAFE_C(ABSY, 4, 3);
    END;
    HANDLER(OPC_ADCSAFE_ABS_Y_N);
    _ADCSAFE_N(ABSY, 4, 3);
    END;
    HANDLER(OPC_ADCSAFE_ABS_Y_V);
    _ADCSAFE_V(ABSY, 4, 3);
    END;
    HANDLER(OPC_ADCSAFE_ABS_Y_Z);
    _ADCSAFE_Z(ABSY, 4, 3);
    END;
#endif
    HANDLER(OPC_ADCSAFE_ABS_Y_ALL);
    _ADCSAFE_ALL(ABSY, 4, 3);
    END;

  /* ADC (zp)   [65c02] */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_ADC_IND);
    _ADC(IND, 5, 2);
    END;
    HANDLER(OPC_ADC_IND_C);
    _ADC_C(IND, 5, 2);
    END;
    HANDLER(OPC_ADC_IND_N);
    _ADC_N(IND, 5, 2);
    END;
    HANDLER(OPC_ADC_IND_V);
    _ADC_V(IND, 5, 2);
    END;
    HANDLER(OPC_ADC_IND_Z);
    _ADC_Z(IND, 5, 2);
    END;
#endif
    HANDLER(OPC_ADC_IND_ALL);
    _ADC_ALL(IND, 5, 2);
    END;

  /* ADC (zp,x) */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_ADC_IND_X);
    _ADC(INDX, 6, 2);
    END;
    HANDLER(OPC_ADC_IND_X_C);
    _ADC_C(INDX, 6, 2);
    END;
    HANDLER(OPC_ADC_IND_X_N);
    _ADC_N(INDX, 6, 2);
    END;
    HANDLER(OPC_ADC_IND_X_V);
    _ADC_V(INDX, 6, 2);
    END;
    HANDLER(OPC_ADC_IND_X_Z);
    _ADC_Z(INDX, 6, 2);
    END;
#endif
    HANDLER(OPC_ADC_IND_X_ALL);
    _ADC_ALL(INDX, 6, 2);
    END;

  /* ADC (zp),y */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_ADC_IND_Y);
    _ADC(INDY, 5, 2);
    END;
    HANDLER(OPC_ADC_IND_Y_C);
    _ADC_C(INDY, 5, 2);
    END;
    HANDLER(OPC_ADC_IND_Y_N);
    _ADC_N(INDY, 5, 2);
    END;
    HANDLER(OPC_ADC_IND_Y_V);
    _ADC_V(INDY, 5, 2);
    END;
    HANDLER(OPC_ADC_IND_Y_Z);
    _ADC_Z(INDY, 5, 2);
    END;
#endif
    HANDLER(OPC_ADC_IND_Y_ALL);
    _ADC_ALL(INDY, 5, 2);
    END;


  /* SBC */

#define _SBC(mode, cyc, size)     _SUM     (-, mode, cyc, size)
#define _SBC_C(mode, cyc, size)   _SUM_C   (-, mode, cyc, size)
#define _SBC_N(mode, cyc, size)   _SUM_N   (-, mode, cyc, size)
#define _SBC_V(mode, cyc, size)   _SUM_V   (-, mode, cyc, size)
#define _SBC_Z(mode, cyc, size)   _SUM_Z   (-, mode, cyc, size)
#define _SBC_ALL(mode, cyc, size) _SUM_ALL (-, mode, cyc, size)
#define _SBCSAFE(mode, cyc, size)     _SUMSAFE     (-, mode, cyc, size)
#define _SBCSAFE_C(mode, cyc, size)   _SUMSAFE_C   (-, mode, cyc, size)
#define _SBCSAFE_N(mode, cyc, size)   _SUMSAFE_N   (-, mode, cyc, size)
#define _SBCSAFE_V(mode, cyc, size)   _SUMSAFE_V   (-, mode, cyc, size)
#define _SBCSAFE_Z(mode, cyc, size)   _SUMSAFE_Z   (-, mode, cyc, size)
#define _SBCSAFE_ALL(mode, cyc, size) _SUMSAFE_ALL (-, mode, cyc, size)


  /* SBC #immediate */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_SBC_IMM);
    _SBC(IMM, 2, 2);
    END;
    HANDLER(OPC_SBC_IMM_C);
    _SBC_C(IMM, 2, 2);
    END;
    HANDLER(OPC_SBC_IMM_N);
    _SBC_N(IMM, 2, 2);
    END;
    HANDLER(OPC_SBC_IMM_V);
    _SBC_V(IMM, 2, 2);
    END;
    HANDLER(OPC_SBC_IMM_Z);
    _SBC_Z(IMM, 2, 2);
    END;
#endif
    HANDLER(OPC_SBC_IMM_ALL);
    _SBC_ALL(IMM, 2, 2);
    END;

  /* SBC ZP */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_SBC_ZP);
    _SBC(ZP, 3, 2);
    END;
    HANDLER(OPC_SBC_ZP_C);
    _SBC_C(ZP, 3, 2);
    END;
    HANDLER(OPC_SBC_ZP_N);
    _SBC_N(ZP, 3, 2);
    END;
    HANDLER(OPC_SBC_ZP_V);
    _SBC_V(ZP, 3, 2);
    END;
    HANDLER(OPC_SBC_ZP_Z);
    _SBC_Z(ZP, 3, 2);
    END;
#endif
    HANDLER(OPC_SBC_ZP_ALL);
    _SBC_ALL(ZP, 3, 2);
    END;

  /* SBC ZP,x */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_SBC_ZP_X);
    _SBC(ZPX, 4, 2);
    END;
    HANDLER(OPC_SBC_ZP_X_C);
    _SBC_C(ZPX, 4, 2);
    END;
    HANDLER(OPC_SBC_ZP_X_N);
    _SBC_N(ZPX, 4, 2);
    END;
    HANDLER(OPC_SBC_ZP_X_V);
    _SBC_V(ZPX, 4, 2);
    END;
    HANDLER(OPC_SBC_ZP_X_Z);
    _SBC_Z(ZPX, 4, 2);
    END;
#endif
    HANDLER(OPC_SBC_ZP_X_ALL);
    _SBC_ALL(ZPX, 4, 2);
    END;

  /* SBC ABS */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_SBC_ABS);
    _SBC(ABS, 4, 3);
    END;
    HANDLER(OPC_SBC_ABS_C);
    _SBC_C(ABS, 4, 3);
    END;
    HANDLER(OPC_SBC_ABS_N);
    _SBC_N(ABS, 4, 3);
    END;
    HANDLER(OPC_SBC_ABS_V);
    _SBC_V(ABS, 4, 3);
    END;
    HANDLER(OPC_SBC_ABS_Z);
    _SBC_Z(ABS, 4, 3);
    END;
#endif
    HANDLER(OPC_SBC_ABS_ALL);
    _SBC_ALL(ABS, 4, 3);
    END;

  /* SBC ABS,x */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_SBC_ABS_X);
    _SBC(ABSX, 4, 3);
    END;
    HANDLER(OPC_SBC_ABS_X_C);
    _SBC_C(ABSX, 4, 3);
    END;
    HANDLER(OPC_SBC_ABS_X_N);
    _SBC_N(ABSX, 4, 3);
    END;
    HANDLER(OPC_SBC_ABS_X_V);
    _SBC_V(ABSX, 4, 3);
    END;
    HANDLER(OPC_SBC_ABS_X_Z);
    _SBC_Z(ABSX, 4, 3);
    END;
#endif
    HANDLER(OPC_SBC_ABS_X_ALL);
    _SBC_ALL(ABSX, 4, 3);
    END;

  /* SBC ABS,Y */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_SBC_ABS_Y);
    _SBC(ABSY, 4, 3);
    END;
    HANDLER(OPC_SBC_ABS_Y_C);
    _SBC_C(ABSY, 4, 3);
    END;
    HANDLER(OPC_SBC_ABS_Y_N);
    _SBC_N(ABSY, 4, 3);
    END;
    HANDLER(OPC_SBC_ABS_Y_V);
    _SBC_V(ABSY, 4, 3);
    END;
    HANDLER(OPC_SBC_ABS_Y_Z);
    _SBC_Z(ABSY, 4, 3);
    END;
#endif
    HANDLER(OPC_SBC_ABS_Y_ALL);
    _SBC_ALL(ABSY, 4, 3);
    END;

  /* SBCSAFE ABS */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_SBCSAFE_ABS);
    _SBCSAFE(ABS, 4, 3);
    END;
    HANDLER(OPC_SBCSAFE_ABS_C);
    _SBCSAFE_C(ABS, 4, 3);
    END;
    HANDLER(OPC_SBCSAFE_ABS_N);
    _SBCSAFE_N(ABS, 4, 3);
    END;
    HANDLER(OPC_SBCSAFE_ABS_V);
    _SBCSAFE_V(ABS, 4, 3);
    END;
    HANDLER(OPC_SBCSAFE_ABS_Z);
    _SBCSAFE_Z(ABS, 4, 3);
    END;
#endif
    HANDLER(OPC_SBCSAFE_ABS_ALL);
    _SBCSAFE_ALL(ABS, 4, 3);
    END;

  /* SBCSAFE ABS,x */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_SBCSAFE_ABS_X);
    _SBCSAFE(ABSX, 4, 3);
    END;
    HANDLER(OPC_SBCSAFE_ABS_X_C);
    _SBCSAFE_C(ABSX, 4, 3);
    END;
    HANDLER(OPC_SBCSAFE_ABS_X_N);
    _SBCSAFE_N(ABSX, 4, 3);
    END;
    HANDLER(OPC_SBCSAFE_ABS_X_V);
    _SBCSAFE_V(ABSX, 4, 3);
    END;
    HANDLER(OPC_SBCSAFE_ABS_X_Z);
    _SBCSAFE_Z(ABSX, 4, 3);
    END;
#endif
    HANDLER(OPC_SBCSAFE_ABS_X_ALL);
    _SBCSAFE_ALL(ABSX, 4, 3);
    END;

  /* SBCSAFE ABS,Y */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_SBCSAFE_ABS_Y);
    _SBCSAFE(ABSY, 4, 3);
    END;
    HANDLER(OPC_SBCSAFE_ABS_Y_C);
    _SBCSAFE_C(ABSY, 4, 3);
    END;
    HANDLER(OPC_SBCSAFE_ABS_Y_N);
    _SBCSAFE_N(ABSY, 4, 3);
    END;
    HANDLER(OPC_SBCSAFE_ABS_Y_V);
    _SBCSAFE_V(ABSY, 4, 3);
    END;
    HANDLER(OPC_SBCSAFE_ABS_Y_Z);
    _SBCSAFE_Z(ABSY, 4, 3);
    END;
#endif
    HANDLER(OPC_SBCSAFE_ABS_Y_ALL);
    _SBCSAFE_ALL(ABSY, 4, 3);
    END;

  /* SBC (zp)  [65c02] */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_SBC_IND);
    _SBC(IND, 5, 2);
    END;
    HANDLER(OPC_SBC_IND_C);
    _SBC_C(IND, 5, 2);
    END;
    HANDLER(OPC_SBC_IND_N);
    _SBC_N(IND, 5, 2);
    END;
    HANDLER(OPC_SBC_IND_V);
    _SBC_V(IND, 5, 2);
    END;
    HANDLER(OPC_SBC_IND_Z);
    _SBC_Z(IND, 5, 2);
    END;
#endif
    HANDLER(OPC_SBC_IND_ALL);
    _SBC_ALL(IND, 5, 2);
    END;

  /* SBC (zp,x) */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_SBC_IND_X);
    _SBC(INDX, 6, 2);
    END;
    HANDLER(OPC_SBC_IND_X_C);
    _SBC_C(INDX, 6, 2);
    END;
    HANDLER(OPC_SBC_IND_X_N);
    _SBC_N(INDX, 6, 2);
    END;
    HANDLER(OPC_SBC_IND_X_V);
    _SBC_V(INDX, 6, 2);
    END;
    HANDLER(OPC_SBC_IND_X_Z);
    _SBC_Z(INDX, 6, 2);
    END;
#endif
    HANDLER(OPC_SBC_IND_X_ALL);
    _SBC_ALL(INDX, 6, 2);
    END;

  /* SBC (zp),y */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_SBC_IND_Y);
    _SBC(INDY, 5, 2);
    END;
    HANDLER(OPC_SBC_IND_Y_C);
    _SBC_C(INDY, 5, 2);
    END;
    HANDLER(OPC_SBC_IND_Y_N);
    _SBC_N(INDY, 5, 2);
    END;
    HANDLER(OPC_SBC_IND_Y_V);
    _SBC_V(INDY, 5, 2);
    END;
    HANDLER(OPC_SBC_IND_Y_Z);
    _SBC_Z(INDY, 5, 2);
    END;
#endif
    HANDLER(OPC_SBC_IND_Y_ALL);
    _SBC_ALL(INDY, 5, 2);
    END;


#define _BITOP(op,mode,cyc,size)   A op read (mode); ADDCYC(cyc); PC += size
#define _BITOP_N(op,mode,cyc,size) A op read (mode); SET_N (A); \
      ADDCYC (cyc); PC += size
#define _BITOP_Z(op,mode,cyc,size) A op read (mode); SET_Z (A); \
	ADDCYC (cyc); PC += size
#define _BITOP_ALL(op, mode,cyc,size) A op read (mode); SET_N (A); SET_Z (A); \
	  ADDCYC (cyc); PC += size


  /* AND */
#ifdef USE_CC_VARIANTS
#define _AND(mode,cyc,size)     _BITOP     (&=, mode, cyc, size)
#define _AND_N(mode,cyc,size)   _BITOP_N   (&=, mode, cyc, size)
#define _AND_Z(mode,cyc,size)   _BITOP_Z   (&=, mode, cyc, size)
#endif
#define _AND_ALL(mode,cyc,size) _BITOP_ALL (&=, mode, cyc, size)


  /* AND #immediate */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_AND_IMM);
    _AND(IMM, 2, 2);
    END;
    HANDLER(OPC_AND_IMM_N);
    _AND_N(IMM, 2, 2);
    END;
    HANDLER(OPC_AND_IMM_Z);
    _AND_Z(IMM, 2, 2);
    END;
#endif
    HANDLER(OPC_AND_IMM_ALL);
    _AND_ALL(IMM, 2, 2);
    END;

  /* AND ZP */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_AND_ZP);
    _AND(ZP, 3, 2);
    END;
    HANDLER(OPC_AND_ZP_N);
    _AND_N(ZP, 3, 2);
    END;
    HANDLER(OPC_AND_ZP_Z);
    _AND_Z(ZP, 3, 2);
    END;
#endif
    HANDLER(OPC_AND_ZP_ALL);
    _AND_ALL(ZP, 3, 2);
    END;

  /* AND ZP,x */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_AND_ZP_X);
    _AND(ZPX, 4, 2);
    END;
    HANDLER(OPC_AND_ZP_X_N);
    _AND_N(ZPX, 4, 2);
    END;
    HANDLER(OPC_AND_ZP_X_Z);
    _AND_Z(ZPX, 4, 2);
    END;
#endif
    HANDLER(OPC_AND_ZP_X_ALL);
    _AND_ALL(ZPX, 4, 2);
    END;

  /* AND ABS */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_AND_ABS);
    _AND(ABS, 4, 3);
    END;
    HANDLER(OPC_AND_ABS_N);
    _AND_N(ABS, 4, 3);
    END;
    HANDLER(OPC_AND_ABS_Z);
    _AND_Z(ABS, 4, 3);
    END;
#endif
    HANDLER(OPC_AND_ABS_ALL);
    _AND_ALL(ABS, 4, 3);
    END;

  /* AND ABS,x */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_AND_ABS_X);
    _AND(ABSX, 4, 3);
    END;
    HANDLER(OPC_AND_ABS_X_N);
    _AND_N(ABSX, 4, 3);
    END;
    HANDLER(OPC_AND_ABS_X_Z);
    _AND_Z(ABSX, 4, 3);
    END;
#endif
    HANDLER(OPC_AND_ABS_X_ALL);
    _AND_ALL(ABSX, 4, 3);
    END;

  /* AND ABS,Y */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_AND_ABS_Y);
    _AND(ABSY, 4, 3);
    END;
    HANDLER(OPC_AND_ABS_Y_N);
    _AND_N(ABSY, 4, 3);
    END;
    HANDLER(OPC_AND_ABS_Y_Z);
    _AND_Z(ABSY, 4, 3);
    END;
#endif
    HANDLER(OPC_AND_ABS_Y_ALL);
    _AND_ALL(ABSY, 4, 3);
    END;

  /* AND (zp)  [65c02] */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_AND_IND);
    _AND(IND, 5, 2);
    END;
    HANDLER(OPC_AND_IND_N);
    _AND_N(IND, 5, 2);
    END;
    HANDLER(OPC_AND_IND_Z);
    _AND_Z(IND, 5, 2);
    END;
#endif
    HANDLER(OPC_AND_IND_ALL);
    _AND_ALL(IND, 5, 2);
    END;

  /* AND (zp,x) */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_AND_IND_X);
    _AND(INDX, 6, 2);
    END;
    HANDLER(OPC_AND_IND_X_N);
    _AND_N(INDX, 6, 2);
    END;
    HANDLER(OPC_AND_IND_X_Z);
    _AND_Z(INDX, 6, 2);
    END;
#endif
    HANDLER(OPC_AND_IND_X_ALL);
    _AND_ALL(INDX, 6, 2);
    END;

  /* AND (zp),y */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_AND_IND_Y);
    _AND(INDY, 5, 2);
    END;
    HANDLER(OPC_AND_IND_Y_N);
    _AND_N(INDY, 5, 2);
    END;
    HANDLER(OPC_AND_IND_Y_Z);
    _AND_Z(INDY, 5, 2);
    END;
#endif
    HANDLER(OPC_AND_IND_Y_ALL);
    _AND_ALL(INDY, 5, 2);
    END;


  /* OR */
#define _OR(mode,cyc,size)     _BITOP     (|=, mode, cyc, size)
#define _OR_N(mode,cyc,size)   _BITOP_N   (|=, mode, cyc, size)
#define _OR_Z(mode,cyc,size)   _BITOP_Z   (|=, mode, cyc, size)
#define _OR_ALL(mode,cyc,size) _BITOP_ALL (|=, mode, cyc, size)


  /* OR #immediate */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_OR_IMM);
    _OR(IMM, 2, 2);
    END;
    HANDLER(OPC_OR_IMM_N);
    _OR_N(IMM, 2, 2);
    END;
    HANDLER(OPC_OR_IMM_Z);
    _OR_Z(IMM, 2, 2);
    END;
#endif
    HANDLER(OPC_OR_IMM_ALL);
    _OR_ALL(IMM, 2, 2);
    END;

  /* OR ZP */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_OR_ZP);
    _OR(ZP, 3, 2);
    END;
    HANDLER(OPC_OR_ZP_N);
    _OR_N(ZP, 3, 2);
    END;
    HANDLER(OPC_OR_ZP_Z);
    _OR_Z(ZP, 3, 2);
    END;
#endif
    HANDLER(OPC_OR_ZP_ALL);
    _OR_ALL(ZP, 3, 2);
    END;

  /* OR ZP,x */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_OR_ZP_X);
    _OR(ZPX, 4, 2);
    END;
    HANDLER(OPC_OR_ZP_X_N);
    _OR_N(ZPX, 4, 2);
    END;
    HANDLER(OPC_OR_ZP_X_Z);
    _OR_Z(ZPX, 4, 2);
    END;
#endif
    HANDLER(OPC_OR_ZP_X_ALL);
    _OR_ALL(ZPX, 4, 2);
    END;

  /* OR ABS */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_OR_ABS);
    _OR(ABS, 4, 3);
    END;
    HANDLER(OPC_OR_ABS_N);
    _OR_N(ABS, 4, 3);
    END;
    HANDLER(OPC_OR_ABS_Z);
    _OR_Z(ABS, 4, 3);
    END;
#endif
    HANDLER(OPC_OR_ABS_ALL);
    _OR_ALL(ABS, 4, 3);
    END;

  /* OR ABS,x */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_OR_ABS_X);
    _OR(ABSX, 4, 3);
    END;
    HANDLER(OPC_OR_ABS_X_N);
    _OR_N(ABSX, 4, 3);
    END;
    HANDLER(OPC_OR_ABS_X_Z);
    _OR_Z(ABSX, 4, 3);
    END;
#endif
    HANDLER(OPC_OR_ABS_X_ALL);
    _OR_ALL(ABSX, 4, 3);
    END;

  /* OR ABS,Y */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_OR_ABS_Y);
    _OR(ABSY, 4, 3);
    END;
    HANDLER(OPC_OR_ABS_Y_N);
    _OR_N(ABSY, 4, 3);
    END;
    HANDLER(OPC_OR_ABS_Y_Z);
    _OR_Z(ABSY, 4, 3);
    END;
#endif
    HANDLER(OPC_OR_ABS_Y_ALL);
    _OR_ALL(ABSY, 4, 3);
    END;

  /* OR (zp,x) */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_OR_IND_X);
    _OR(INDX, 6, 2);
    END;
    HANDLER(OPC_OR_IND_X_N);
    _OR_N(INDX, 6, 2);
    END;
    HANDLER(OPC_OR_IND_X_Z);
    _OR_Z(INDX, 6, 2);
    END;
#endif
    HANDLER(OPC_OR_IND_X_ALL);
    _OR_ALL(INDX, 6, 2);
    END;

  /* OR (zp)  [65c02] */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_OR_IND);
    _OR(IND, 5, 2);
    END;
    HANDLER(OPC_OR_IND_N);
    _OR_N(IND, 5, 2);
    END;
    HANDLER(OPC_OR_IND_Z);
    _OR_Z(IND, 5, 2);
    END;
#endif
    HANDLER(OPC_OR_IND_ALL);
    _OR_ALL(IND, 5, 2);
    END;

  /* OR (zp),y */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_OR_IND_Y);
    _OR(INDY, 5, 2);
    END;
    HANDLER(OPC_OR_IND_Y_N);
    _OR_N(INDY, 5, 2);
    END;
    HANDLER(OPC_OR_IND_Y_Z);
    _OR_Z(INDY, 5, 2);
    END;
#endif
    HANDLER(OPC_OR_IND_Y_ALL);
    _OR_ALL(INDY, 5, 2);
    END;


  /* EOR */
#define _EOR(mode,cyc,size)     _BITOP     (^=, mode, cyc, size)
#define _EOR_N(mode,cyc,size)   _BITOP_N   (^=, mode, cyc, size)
#define _EOR_Z(mode,cyc,size)   _BITOP_Z   (^=, mode, cyc, size)
#define _EOR_ALL(mode,cyc,size) _BITOP_ALL (^=, mode, cyc, size)


  /* EOR #immediate */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_EOR_IMM);
    _EOR(IMM, 2, 2);
    END;
    HANDLER(OPC_EOR_IMM_N);
    _EOR_N(IMM, 2, 2);
    END;
    HANDLER(OPC_EOR_IMM_Z);
    _EOR_Z(IMM, 2, 2);
    END;
#endif
    HANDLER(OPC_EOR_IMM_ALL);
    _EOR_ALL(IMM, 2, 2);
    END;

  /* EOR ZP */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_EOR_ZP);
    _EOR(ZP, 3, 2);
    END;
    HANDLER(OPC_EOR_ZP_N);
    _EOR_N(ZP, 3, 2);
    END;
    HANDLER(OPC_EOR_ZP_Z);
    _EOR_Z(ZP, 3, 2);
    END;
#endif
    HANDLER(OPC_EOR_ZP_ALL);
    _EOR_ALL(ZP, 3, 2);
    END;

  /* EOR ZP,x */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_EOR_ZP_X);
    _EOR(ZPX, 4, 2);
    END;
    HANDLER(OPC_EOR_ZP_X_N);
    _EOR_N(ZPX, 4, 2);
    END;
    HANDLER(OPC_EOR_ZP_X_Z);
    _EOR_Z(ZPX, 4, 2);
    END;
#endif
    HANDLER(OPC_EOR_ZP_X_ALL);
    _EOR_ALL(ZPX, 4, 2);
    END;

  /* EOR ABS */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_EOR_ABS);
    _EOR(ABS, 4, 3);
    END;
    HANDLER(OPC_EOR_ABS_N);
    _EOR_N(ABS, 4, 3);
    END;
    HANDLER(OPC_EOR_ABS_Z);
    _EOR_Z(ABS, 4, 3);
    END;
#endif
    HANDLER(OPC_EOR_ABS_ALL);
    _EOR_ALL(ABS, 4, 3);
    END;

  /* EOR ABS,x */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_EOR_ABS_X);
    _EOR(ABSX, 4, 3);
    END;
    HANDLER(OPC_EOR_ABS_X_N);
    _EOR_N(ABSX, 4, 3);
    END;
    HANDLER(OPC_EOR_ABS_X_Z);
    _EOR_Z(ABSX, 4, 3);
    END;
#endif
    HANDLER(OPC_EOR_ABS_X_ALL);
    _EOR_ALL(ABSX, 4, 3);
    END;

  /* EOR ABS,Y */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_EOR_ABS_Y);
    _EOR(ABSY, 4, 3);
    END;
    HANDLER(OPC_EOR_ABS_Y_N);
    _EOR_N(ABSY, 4, 3);
    END;
    HANDLER(OPC_EOR_ABS_Y_Z);
    _EOR_Z(ABSY, 4, 3);
    END;
#endif
    HANDLER(OPC_EOR_ABS_Y_ALL);
    _EOR_ALL(ABSY, 4, 3);
    END;

  /* EOR (zp)  [65c02] */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_EOR_IND);
    _EOR(IND, 5, 2);
    END;
    HANDLER(OPC_EOR_IND_N);
    _EOR_N(IND, 5, 2);
    END;
    HANDLER(OPC_EOR_IND_Z);
    _EOR_Z(IND, 5, 2);
    END;
#endif
    HANDLER(OPC_EOR_IND_ALL);
    _EOR_ALL(IND, 5, 2);
    END;

  /* EOR (zp,x) */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_EOR_IND_X);
    _EOR(INDX, 6, 2);
    END;
    HANDLER(OPC_EOR_IND_X_N);
    _EOR_N(INDX, 6, 2);
    END;
    HANDLER(OPC_EOR_IND_X_Z);
    _EOR_Z(INDX, 6, 2);
    END;
#endif
    HANDLER(OPC_EOR_IND_X_ALL);
    _EOR_ALL(INDX, 6, 2);
    END;

  /* EOR (zp),y */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_EOR_IND_Y);
    _EOR(INDY, 5, 2);
    END;
    HANDLER(OPC_EOR_IND_Y_N);
    _EOR_N(INDY, 5, 2);
    END;
    HANDLER(OPC_EOR_IND_Y_Z);
    _EOR_Z(INDY, 5, 2);
    END;
#endif
    HANDLER(OPC_EOR_IND_Y_ALL);
    _EOR_ALL(INDY, 5, 2);
    END;


#define _ASL(mode,cyc,size) write (mode, read (mode) << 1); \
	      ADDCYC(cyc); PC += size
#define _ASL_C(mode,cyc,size) temp_uchar = read (mode); \
	      CFLAG = temp_uchar >> 7; write (mode, temp_uchar << 1); \
	      ADDCYC(cyc); PC += size
#define _ASL_N(mode,cyc,size) temp_uchar = read (mode) << 1; \
	      write (mode, temp_uchar); SET_N (temp_uchar); \
	      ADDCYC (cyc); PC += size
#define _ASL_Z(mode,cyc,size) temp_uchar = read (mode) << 1; \
	      write (mode, temp_uchar); SET_Z (temp_uchar); \
	      ADDCYC (cyc); PC += size
#define _ASL_ALL(mode,cyc,size) temp_uchar = read (mode); \
	      CFLAG = temp_uchar >> 7; temp_uchar <<= 1; \
	      write (mode, temp_uchar); SET_N(temp_uchar); SET_Z(temp_uchar); \
	      ADDCYC(cyc); PC += size

#ifdef USE_CC_VARIANTS
    HANDLER(OPC_ASL_ACC);
    _ASL(AREG, 2, 1);
    END;
    HANDLER(OPC_ASL_ACC_C);
    _ASL_C(AREG, 2, 1);
    END;
    HANDLER(OPC_ASL_ACC_N);
    _ASL_N(AREG, 2, 1);
    END;
    HANDLER(OPC_ASL_ACC_Z);
    _ASL_Z(AREG, 2, 1);
    END;
#endif
    HANDLER(OPC_ASL_ACC_ALL);
    _ASL_ALL(AREG, 2, 1);
    END;


#ifdef USE_CC_VARIANTS
    HANDLER(OPC_ASL_ZP);
    _ASL(ZP, 5, 2);
    END;
    HANDLER(OPC_ASL_ZP_C);
    _ASL_C(ZP, 5, 2);
    END;
    HANDLER(OPC_ASL_ZP_N);
    _ASL_N(ZP, 5, 2);
    END;
    HANDLER(OPC_ASL_ZP_Z);
    _ASL_Z(ZP, 5, 2);
    END;
#endif
    HANDLER(OPC_ASL_ZP_ALL);
    _ASL_ALL(ZP, 5, 2);
    END;


#ifdef USE_CC_VARIANTS
    HANDLER(OPC_ASL_ZP_X);
    _ASL(ZPX, 6, 2);
    END;
    HANDLER(OPC_ASL_ZP_X_C);
    _ASL_C(ZPX, 6, 2);
    END;
    HANDLER(OPC_ASL_ZP_X_N);
    _ASL_N(ZPX, 6, 2);
    END;
    HANDLER(OPC_ASL_ZP_X_Z);
    _ASL_Z(ZPX, 6, 2);
    END;
#endif
    HANDLER(OPC_ASL_ZP_X_ALL);
    _ASL_ALL(ZPX, 6, 2);
    END;


#ifdef USE_CC_VARIANTS
    HANDLER(OPC_ASL_ABS);
    _ASL(ABS, 6, 3);
    END;
    HANDLER(OPC_ASL_ABS_C);
    _ASL_C(ABS, 6, 3);
    END;
    HANDLER(OPC_ASL_ABS_N);
    _ASL_N(ABS, 6, 3);
    END;
    HANDLER(OPC_ASL_ABS_Z);
    _ASL_Z(ABS, 6, 3);
    END;
#endif
    HANDLER(OPC_ASL_ABS_ALL);
    _ASL_ALL(ABS, 6, 3);
    END;


#ifdef USE_CC_VARIANTS
    HANDLER(OPC_ASL_ABS_X);
    _ASL(ABSX, 7, 3);
    END;
    HANDLER(OPC_ASL_ABS_X_C);
    _ASL_C(ABSX, 7, 3);
    END;
    HANDLER(OPC_ASL_ABS_X_N);
    _ASL_N(ABSX, 7, 3);
    END;
    HANDLER(OPC_ASL_ABS_X_Z);
    _ASL_Z(ABSX, 7, 3);
    END;
#endif
    HANDLER(OPC_ASL_ABS_X_ALL);
    _ASL_ALL(ABSX, 7, 3);
    END;


  /* LSR */
#define _LSR(mode,cyc,size) write (mode, read (mode) >> 1); \
	      ADDCYC(cyc); PC += size
#define _LSR_C(mode,cyc,size) temp_uchar = read (mode); \
	      CFLAG = temp_uchar & 0x1; write (mode, temp_uchar >> 1); \
	      ADDCYC(cyc); PC += size
#define _LSR_N(mode,cyc,size) temp_uchar = read (mode) >> 1; \
	      write (mode, temp_uchar); SET_N (0); \
	      ADDCYC (cyc); PC += size
#define _LSR_Z(mode,cyc,size) temp_uchar = read (mode) >> 1; \
	      write (mode, temp_uchar); SET_Z (temp_uchar); \
	      ADDCYC (cyc); PC += size
#define _LSR_ALL(mode,cyc,size) temp_uchar = read (mode); \
	      CFLAG = temp_uchar & 0x1; temp_uchar >>= 1; \
	      write (mode, temp_uchar); SET_N(0); SET_Z(temp_uchar); \
	      ADDCYC(cyc); PC += size


  /* LSR A */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_LSR_ACC);
    _LSR(AREG, 2, 1);
    END;
    HANDLER(OPC_LSR_ACC_C);
    _LSR_C(AREG, 2, 1);
    END;
    HANDLER(OPC_LSR_ACC_N);
    _LSR_N(AREG, 2, 1);
    END;
    HANDLER(OPC_LSR_ACC_Z);
    _LSR_Z(AREG, 2, 1);
    END;
#endif
    HANDLER(OPC_LSR_ACC_ALL);
    _LSR_ALL(AREG, 2, 1);
    END;

  /* LSR zp */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_LSR_ZP);
    _LSR(ZP, 5, 2);
    END;
    HANDLER(OPC_LSR_ZP_C);
    _LSR_C(ZP, 5, 2);
    END;
    HANDLER(OPC_LSR_ZP_N);
    _LSR_N(ZP, 5, 2);
    END;
    HANDLER(OPC_LSR_ZP_Z);
    _LSR_Z(ZP, 5, 2);
    END;
#endif
    HANDLER(OPC_LSR_ZP_ALL);
    _LSR_ALL(ZP, 5, 2);
    END;

  /* LSR zp,x */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_LSR_ZP_X);
    _LSR(ZPX, 6, 2);
    END;
    HANDLER(OPC_LSR_ZP_X_C);
    _LSR_C(ZPX, 6, 2);
    END;
    HANDLER(OPC_LSR_ZP_X_N);
    _LSR_N(ZPX, 6, 2);
    END;
    HANDLER(OPC_LSR_ZP_X_Z);
    _LSR_Z(ZPX, 6, 2);
    END;
#endif
    HANDLER(OPC_LSR_ZP_X_ALL);
    _LSR_ALL(ZPX, 6, 2);
    END;

  /* LSR ABS */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_LSR_ABS);
    _LSR(ABS, 6, 3);
    END;
    HANDLER(OPC_LSR_ABS_C);
    _LSR_C(ABS, 6, 3);
    END;
    HANDLER(OPC_LSR_ABS_N);
    _LSR_N(ABS, 6, 3);
    END;
    HANDLER(OPC_LSR_ABS_Z);
    _LSR_Z(ABS, 6, 3);
    END;
#endif
    HANDLER(OPC_LSR_ABS_ALL);
    _LSR_ALL(ABS, 6, 3);
    END;

  /* LSR ABS,x */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_LSR_ABS_X);
    _LSR(ABSX, 7, 3);
    END;
    HANDLER(OPC_LSR_ABS_X_C);
    _LSR_C(ABSX, 7, 3);
    END;
    HANDLER(OPC_LSR_ABS_X_N);
    _LSR_N(ABSX, 7, 3);
    END;
    HANDLER(OPC_LSR_ABS_X_Z);
    _LSR_Z(ABSX, 7, 3);
    END;
#endif
    HANDLER(OPC_LSR_ABS_X_ALL);
    _LSR_ALL(ABSX, 7, 3);
    END;


#define _ROL(mode,cyc,size) temp_uchar = (read (mode) << 1) | CFLAG; \
	write (mode, temp_uchar); ADDCYC (cyc); PC += size
#define _ROL_C(mode,cyc,size) temp_uchar = read (mode); \
	write (mode, (temp_uchar << 1) | CFLAG); CFLAG = temp_uchar >> 7; \
	ADDCYC (cyc); PC += size
#define _ROL_N(mode,cyc,size) temp_uchar = (read (mode) << 1) | CFLAG; \
	write (mode, temp_uchar); SET_N (temp_uchar); ADDCYC (cyc); PC += size
#define _ROL_Z(mode,cyc,size) temp_uchar = (read (mode) << 1) | CFLAG; \
	write (mode, temp_uchar); SET_Z (temp_uchar); ADDCYC (cyc); PC += size
#define _ROL_ALL(mode,cyc,size) temp_uchar = read (mode); \
	write (mode, (temp_uchar << 1) | CFLAG); CFLAG = temp_uchar >> 7; \
	temp_uchar <<= 1; SET_N (temp_uchar); SET_Z (temp_uchar); \
	ADDCYC (cyc); PC += size

  /* ROL A */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_ROL_ACC);
    _ROL(AREG, 2, 1);
    END;
    HANDLER(OPC_ROL_ACC_C);
    _ROL_C(AREG, 2, 1);
    END;
    HANDLER(OPC_ROL_ACC_N);
    _ROL_N(AREG, 2, 1);
    END;
    HANDLER(OPC_ROL_ACC_Z);
    _ROL_Z(AREG, 2, 1);
    END;
#endif
    HANDLER(OPC_ROL_ACC_ALL);
    _ROL_ALL(AREG, 2, 1);
    END;

  /* ROL zp */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_ROL_ZP);
    _ROL(ZP, 5, 2);
    END;
    HANDLER(OPC_ROL_ZP_C);
    _ROL_C(ZP, 5, 2);
    END;
    HANDLER(OPC_ROL_ZP_N);
    _ROL_N(ZP, 5, 2);
    END;
    HANDLER(OPC_ROL_ZP_Z);
    _ROL_Z(ZP, 5, 2);
    END;
#endif
    HANDLER(OPC_ROL_ZP_ALL);
    _ROL_ALL(ZP, 5, 2);
    END;

  /* ROL zp,x */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_ROL_ZP_X);
    _ROL(ZPX, 6, 2);
    END;
    HANDLER(OPC_ROL_ZP_X_C);
    _ROL_C(ZPX, 6, 2);
    END;
    HANDLER(OPC_ROL_ZP_X_N);
    _ROL_N(ZPX, 6, 2);
    END;
    HANDLER(OPC_ROL_ZP_X_Z);
    _ROL_Z(ZPX, 6, 2);
    END;
#endif
    HANDLER(OPC_ROL_ZP_X_ALL);
    _ROL_ALL(ZPX, 6, 2);
    END;

  /* ROL ABS */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_ROL_ABS);
    _ROL(ABS, 6, 3);
    END;
    HANDLER(OPC_ROL_ABS_C);
    _ROL_C(ABS, 6, 3);
    END;
    HANDLER(OPC_ROL_ABS_N);
    _ROL_N(ABS, 6, 3);
    END;
    HANDLER(OPC_ROL_ABS_Z);
    _ROL_Z(ABS, 6, 3);
    END;
#endif
    HANDLER(OPC_ROL_ABS_ALL);
    _ROL_ALL(ABS, 6, 3);
    END;

  /* ROL ABS,x */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_ROL_ABS_X);
    _ROL(ABSX, 7, 3);
    END;
    HANDLER(OPC_ROL_ABS_X_C);
    _ROL_C(ABSX, 7, 3);
    END;
    HANDLER(OPC_ROL_ABS_X_N);
    _ROL_N(ABSX, 7, 3);
    END;
    HANDLER(OPC_ROL_ABS_X_Z);
    _ROL_Z(ABSX, 7, 3);
    END;
#endif
    HANDLER(OPC_ROL_ABS_X_ALL);
    _ROL_ALL(ABSX, 7, 3);
    END;


#define _ROR(mode,cyc,size) temp_uchar = (read (mode) >> 1) | (CFLAG << 7); \
	write (mode, temp_uchar); ADDCYC (cyc); PC += size
#define _ROR_C(mode,cyc,size) temp_uchar = read (mode); \
	write (mode, (temp_uchar >> 1) | (CFLAG << 7)); \
	CFLAG = temp_uchar & 1; ADDCYC (cyc); PC += size
#define _ROR_N(mode,cyc,size) temp_uchar = (read (mode) >> 1) | (CFLAG << 7);\
	write (mode, temp_uchar); NFLAG = CFLAG; ADDCYC (cyc); PC += size
#define _ROR_Z(mode,cyc,size) temp_uchar = (read (mode) >> 1) | (CFLAG << 7);\
	write (mode, temp_uchar); SET_Z (temp_uchar); ADDCYC (cyc); PC += size
#define _ROR_ALL(mode,cyc,size) \
  { \
    unsigned char tmp; \
    temp_uchar = read (mode); \
    NFLAG = CFLAG; \
    tmp = (temp_uchar >> 1) | (CFLAG << 7); \
    write (mode, tmp); \
    SET_Z (tmp); \
    CFLAG = temp_uchar & 1; \
    ADDCYC (cyc); \
    PC += size; \
  }


  /* ROR A */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_ROR_ACC);
    _ROR(AREG, 2, 1);
    END;
    HANDLER(OPC_ROR_ACC_C);
    _ROR_C(AREG, 2, 1);
    END;
    HANDLER(OPC_ROR_ACC_N);
    _ROR_N(AREG, 2, 1);
    END;
    HANDLER(OPC_ROR_ACC_Z);
    _ROR_Z(AREG, 2, 1);
    END;
#endif
    HANDLER(OPC_ROR_ACC_ALL);
    _ROR_ALL(AREG, 2, 1);
    END;

  /* ROR zp */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_ROR_ZP);
    _ROR(ZP, 5, 2);
    END;
    HANDLER(OPC_ROR_ZP_C);
    _ROR_C(ZP, 5, 2);
    END;
    HANDLER(OPC_ROR_ZP_N);
    _ROR_N(ZP, 5, 2);
    END;
    HANDLER(OPC_ROR_ZP_Z);
    _ROR_Z(ZP, 5, 2);
    END;
#endif
    HANDLER(OPC_ROR_ZP_ALL);
    _ROR_ALL(ZP, 5, 2);
    END;

  /* ROR zp,x */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_ROR_ZP_X);
    _ROR(ZPX, 6, 2);
    END;
    HANDLER(OPC_ROR_ZP_X_C);
    _ROR_C(ZPX, 6, 2);
    END;
    HANDLER(OPC_ROR_ZP_X_N);
    _ROR_N(ZPX, 6, 2);
    END;
    HANDLER(OPC_ROR_ZP_X_Z);
    _ROR_Z(ZPX, 6, 2);
    END;
#endif
    HANDLER(OPC_ROR_ZP_X_ALL);
    _ROR_ALL(ZPX, 6, 2);
    END;

  /* ROR ABS */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_ROR_ABS);
    _ROR(ABS, 6, 3);
    END;
    HANDLER(OPC_ROR_ABS_C);
    _ROR_C(ABS, 6, 3);
    END;
    HANDLER(OPC_ROR_ABS_N);
    _ROR_N(ABS, 6, 3);
    END;
    HANDLER(OPC_ROR_ABS_Z);
    _ROR_Z(ABS, 6, 3);
    END;
#endif
    HANDLER(OPC_ROR_ABS_ALL);
    _ROR_ALL(ABS, 6, 3);
    END;

  /* ROR ABS,x */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_ROR_ABS_X);
    _ROR(ABSX, 7, 3);
    END;
    HANDLER(OPC_ROR_ABS_X_C);
    _ROR_C(ABSX, 7, 3);
    END;
    HANDLER(OPC_ROR_ABS_X_N);
    _ROR_N(ABSX, 7, 3);
    END;
    HANDLER(OPC_ROR_ABS_X_Z);
    _ROR_Z(ABSX, 7, 3);
    END;
#endif
    HANDLER(OPC_ROR_ABS_X_ALL);
    _ROR_ALL(ABSX, 7, 3);
    END;


  /* Bxx */
    HANDLER(OPC_BCC);
    if (!CFLAG)
    {
      ADDCYC(3);
      PC = USHORTARG;
      CHECK_FOR_INTERRUPT();
    }
    else
    {
      ADDCYC(2);
      PC += 2;
    }
    END;
    HANDLER(OPC_BCC_NEWBANK);
    if (!CFLAG)
    {
      ADDCYC(3);
      PC = USHORTARG;
      CHECK_FOR_INTERRUPT();
    }
    else
    {
      ADDCYC(2);
      PC += 2;
    }
    RELOAD_PC_BASE();
    END;
    HANDLER(OPC_BCS);
    if (CFLAG)
    {
      ADDCYC(3);
      PC = USHORTARG;
      CHECK_FOR_INTERRUPT();
    }
    else
    {
      ADDCYC(2);
      PC += 2;
    }
    END;
    HANDLER(OPC_BCS_NEWBANK);
    if (CFLAG)
    {
      ADDCYC(3);
      PC = USHORTARG;
      CHECK_FOR_INTERRUPT();
    }
    else
    {
      ADDCYC(2);
      PC += 2;
    }
    RELOAD_PC_BASE();
    END;
    HANDLER(OPC_BMI);
    if (NFLAG)
    {
      ADDCYC(3);
      PC = USHORTARG;
      CHECK_FOR_INTERRUPT();
    }
    else
    {
      ADDCYC(2);
      PC += 2;
    }
    END;
    HANDLER(OPC_BMI_NEWBANK);
    if (NFLAG)
    {
      ADDCYC(3);
      PC = USHORTARG;
      CHECK_FOR_INTERRUPT();
    }
    else
    {
      ADDCYC(2);
      PC += 2;
    }
    RELOAD_PC_BASE();
    END;
    HANDLER(OPC_BEQ);
    if (!NZFLAG)
    {
      ADDCYC(3);
      PC = USHORTARG;
      CHECK_FOR_INTERRUPT();
    }
    else
    {
      ADDCYC(2);
      PC += 2;
    }
    END;
    HANDLER(OPC_BEQ_NEWBANK);
    if (!NZFLAG)
    {
      ADDCYC(3);
      PC = USHORTARG;
      CHECK_FOR_INTERRUPT();
    }
    else
    {
      ADDCYC(2);
      PC += 2;
    }
    RELOAD_PC_BASE();
    END;
    HANDLER(OPC_BNE);
    if (NZFLAG)
    {
      ADDCYC(3);
      PC = USHORTARG;
      CHECK_FOR_INTERRUPT();
    }
    else
    {
      ADDCYC(2);
      PC += 2;
    }
    END;
    HANDLER(OPC_BNE_NEWBANK);
    if (NZFLAG)
    {
      ADDCYC(3);
      PC = USHORTARG;
      CHECK_FOR_INTERRUPT();
    }
    else
    {
      ADDCYC(2);
      PC += 2;
    }
    RELOAD_PC_BASE();
    END;
    HANDLER(OPC_BPL);
    if (!NFLAG)
    {
      ADDCYC(3);
      PC = USHORTARG;
      CHECK_FOR_INTERRUPT();
    }
    else
    {
      ADDCYC(2);
      PC += 2;
    }
    END;
    HANDLER(OPC_BPL_NEWBANK);
    if (!NFLAG)
    {
      ADDCYC(3);
      PC = USHORTARG;
      CHECK_FOR_INTERRUPT();
    }
    else
    {
      ADDCYC(2);
      PC += 2;
    }
    RELOAD_PC_BASE();
    END;
    HANDLER(OPC_BRA);		/* [65c02] */
    ADDCYC(3);
    PC = USHORTARG;
    CHECK_FOR_INTERRUPT();
    END;
    HANDLER(OPC_BRA_NEWBANK);	/* [65c02] */
    ADDCYC(3);
    PC = USHORTARG;
    RELOAD_PC_BASE();
    CHECK_FOR_INTERRUPT();
    END;
    HANDLER(OPC_BVC);
    if (!VFLAG)
    {
      ADDCYC(3);
      PC = USHORTARG;
      CHECK_FOR_INTERRUPT();
    }
    else
    {
      ADDCYC(2);
      PC += 2;
    }
    END;
    HANDLER(OPC_BVC_NEWBANK);
    if (!VFLAG)
    {
      ADDCYC(3);
      PC = USHORTARG;
      CHECK_FOR_INTERRUPT();
    }
    else
    {
      ADDCYC(2);
      PC += 2;
    }
    RELOAD_PC_BASE();
    END;
    HANDLER(OPC_BVS);
    if (VFLAG)
    {
      ADDCYC(3);
      PC = USHORTARG;
      CHECK_FOR_INTERRUPT();
    }
    else
    {
      ADDCYC(2);
      PC += 2;
    }
    END;
    HANDLER(OPC_BVS_NEWBANK);
    if (VFLAG)
    {
      ADDCYC(3);
      PC = USHORTARG;
      CHECK_FOR_INTERRUPT();
    }
    else
    {
      ADDCYC(2);
      PC += 2;
    }
    RELOAD_PC_BASE();
    END;


  /* BIT */
#define _BIT(mode,cyc,size) readcheck (mode); ADDCYC(cyc); PC += size
#define _BIT_N(mode,cyc,size) SET_N (read (mode));  ADDCYC(cyc); PC += size
#define _BIT_Z(mode,cyc,size) SET_Z (A & read (mode));  ADDCYC(cyc); PC += size
#define _BIT_V(mode,cyc,size) VFLAG = (0x40 & read (mode)); \
      ADDCYC(cyc); PC += size
#define _BIT_ALL(mode,cyc,size) temp_uchar = read (mode); \
	SET_N (temp_uchar); VFLAG = temp_uchar & 0x40; \
	SET_Z (temp_uchar & A); ADDCYC(cyc); PC += size

  /* BIT #imm  [65c02] */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_BIT_IMM);
    _BIT(IMM, 2, 2);
    END;
    HANDLER(OPC_BIT_IMM_N);
    _BIT_N(IMM, 2, 2);
    END;
    HANDLER(OPC_BIT_IMM_Z);
    _BIT_Z(IMM, 2, 2);
    END;
    HANDLER(OPC_BIT_IMM_V);
    _BIT_V(IMM, 2, 2);
    END;
#endif
    HANDLER(OPC_BIT_IMM_ALL);
    _BIT_ALL(IMM, 2, 2);
    END;

  /* BIT zp */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_BIT_ZP);
    _BIT(ZP, 3, 2);
    END;
    HANDLER(OPC_BIT_ZP_N);
    _BIT_N(ZP, 3, 2);
    END;
    HANDLER(OPC_BIT_ZP_Z);
    _BIT_Z(ZP, 3, 2);
    END;
    HANDLER(OPC_BIT_ZP_V);
    _BIT_V(ZP, 3, 2);
    END;
#endif
    HANDLER(OPC_BIT_ZP_ALL);
    _BIT_ALL(ZP, 3, 2);
    END;

  /* BIT zp,x  [65c02] */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_BIT_ZP_X);
    _BIT(ZPX, 4, 2);
    END;
    HANDLER(OPC_BIT_ZP_X_N);
    _BIT_N(ZPX, 4, 2);
    END;
    HANDLER(OPC_BIT_ZP_X_Z);
    _BIT_Z(ZPX, 4, 2);
    END;
    HANDLER(OPC_BIT_ZP_X_V);
    _BIT_V(ZPX, 4, 2);
    END;
#endif
    HANDLER(OPC_BIT_ZP_X_ALL);
    _BIT_ALL(ZPX, 4, 2);
    END;

  /* BIT ABS */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_BIT_ABS);
    _BIT(ABS, 4, 3);
    END;
    HANDLER(OPC_BIT_ABS_N);
    _BIT_N(ABS, 4, 3);
    END;
    HANDLER(OPC_BIT_ABS_Z);
    _BIT_Z(ABS, 4, 3);
    END;
    HANDLER(OPC_BIT_ABS_V);
    _BIT_V(ABS, 4, 3);
    END;
#endif
    HANDLER(OPC_BIT_ABS_ALL);
    _BIT_ALL(ABS, 4, 3);
    END;

  /* BIT abs,x  [65c02] */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_BIT_ABS_X);
    _BIT(ABSX, 4, 3);
    END;
    HANDLER(OPC_BIT_ABS_X_N);
    _BIT_N(ABSX, 4, 3);
    END;
    HANDLER(OPC_BIT_ABS_X_Z);
    _BIT_Z(ABSX, 4, 3);
    END;
    HANDLER(OPC_BIT_ABS_X_V);
    _BIT_V(ABSX, 4, 3);
    END;
#endif
    HANDLER(OPC_BIT_ABS_X_ALL);
    _BIT_ALL(ABSX, 4, 3);
    END;


  /* BRK */
    HANDLER(OPC_BRK);
    PC += 2;
    PUSH(PC >> 8);
    PUSH(PC);
    PUSH(((NFLAG != 0) << N_BIT) | ((NZFLAG == 0) << Z_BIT)
	 | (BFLAG << B_BIT) | (CFLAG << C_BIT)
	 | (IFLAG << I_BIT) | (DFLAG << D_BIT)
	 | ((VFLAG != 0) << V_BIT) | (1 << X_BIT));
    BFLAG = 1;
    IFLAG = 1;			/* FIXME - is this right?  6502 A.L.S. p.65 says so. */
    DFLAG = 0;			/* FIXME - Apple IIgs asm.lang.prog. p.407 says it's OK. */

    PC = READADDR(0xFFFE) | (READADDR(0xFFFF) << 8);
    ADDCYC(7);
    RELOAD_PC_BASE();
    CHECK_FOR_INTERRUPT();
    END;


  /* CLcc */
    HANDLER(OPC_CLC);
    CFLAG = 0;
    ADDCYC(2);
    PC++;
    END;
    HANDLER(OPC_CLD);
    DFLAG = 0;
    ADDCYC(2);
    PC++;
    END;
    HANDLER(OPC_CLI);
    IFLAG = 0;
    ADDCYC(2);
    PC++;
    END;
    HANDLER(OPC_CLV);
    VFLAG = 0;
    ADDCYC(2);
    PC++;
    END;


#define _COMP(reg,mode,cyc,size) readcheck (mode); ADDCYC (cyc); PC += size
#define _COMP_C(reg,mode,cyc,size) CFLAG = (((reg - read (mode)) >> 8) & 1)^1;\
        ADDCYC (cyc); PC += size
#define _COMP_N(reg,mode,cyc,size) SET_N (reg - read (mode)); \
	ADDCYC (cyc); PC += size
#define _COMP_Z(reg,mode,cyc,size) SET_Z (reg - read (mode)); \
	ADDCYC (cyc); PC += size
#define _COMP_ALL(reg,mode,cyc,size) \
      { \
	unsigned short temp_ushort = (reg - read (mode)); \
	CFLAG = ((temp_ushort >> 8) & 1) ^ 1; \
	SET_N ((unsigned char) temp_ushort); \
	SET_Z ((unsigned char) temp_ushort); \
	ADDCYC (cyc); \
	PC += size; \
      }

#define _COMPSAFE(reg,mode,cyc,size) ADDCYC (cyc); PC += size
#define _COMPSAFE_C(reg,mode,cyc,size) \
        CFLAG = (((reg - readsafe (mode)) >> 8) & 1) ^ 1; \
        ADDCYC (cyc); PC += size
#define _COMPSAFE_N(reg,mode,cyc,size) SET_N (reg - readsafe (mode)); \
	ADDCYC (cyc); PC += size
#define _COMPSAFE_Z(reg,mode,cyc,size) SET_Z (reg - readsafe (mode)); \
	ADDCYC (cyc); PC += size
#define _COMPSAFE_ALL(reg,mode,cyc,size) \
      { \
	unsigned short temp_ushort = (reg - readsafe (mode)); \
	CFLAG = ((temp_ushort >> 8) & 1) ^ 1; \
	SET_N ((unsigned char) temp_ushort); \
	SET_Z ((unsigned char) temp_ushort); \
	ADDCYC (cyc); \
	PC += size; \
      }


  /* CMP #immediate */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_CMP_IMM);
    _COMP(A, IMM, 2, 2);
    END;
    HANDLER(OPC_CMP_IMM_C);
    _COMP_C(A, IMM, 2, 2);
    END;
    HANDLER(OPC_CMP_IMM_N);
    _COMP_N(A, IMM, 2, 2);
    END;
    HANDLER(OPC_CMP_IMM_Z);
    _COMP_Z(A, IMM, 2, 2);
    END;
#endif
    HANDLER(OPC_CMP_IMM_ALL);
    _COMP_ALL(A, IMM, 2, 2);
    END;

  /* CMP ZP */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_CMP_ZP);
    _COMP(A, ZP, 3, 2);
    END;
    HANDLER(OPC_CMP_ZP_C);
    _COMP_C(A, ZP, 3, 2);
    END;
    HANDLER(OPC_CMP_ZP_N);
    _COMP_N(A, ZP, 3, 2);
    END;
    HANDLER(OPC_CMP_ZP_Z);
    _COMP_Z(A, ZP, 3, 2);
    END;
#endif
    HANDLER(OPC_CMP_ZP_ALL);
    _COMP_ALL(A, ZP, 3, 2);
    END;

  /* CMP ZP,x */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_CMP_ZP_X);
    _COMP(A, ZPX, 4, 2);
    END;
    HANDLER(OPC_CMP_ZP_X_C);
    _COMP_C(A, ZPX, 4, 2);
    END;
    HANDLER(OPC_CMP_ZP_X_N);
    _COMP_N(A, ZPX, 4, 2);
    END;
    HANDLER(OPC_CMP_ZP_X_Z);
    _COMP_Z(A, ZPX, 4, 2);
    END;
#endif
    HANDLER(OPC_CMP_ZP_X_ALL);
    _COMP_ALL(A, ZPX, 4, 2);
    END;

  /* CMP ABS */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_CMP_ABS);
    _COMP(A, ABS, 4, 3);
    END;
    HANDLER(OPC_CMP_ABS_C);
    _COMP_C(A, ABS, 4, 3);
    END;
    HANDLER(OPC_CMP_ABS_N);
    _COMP_N(A, ABS, 4, 3);
    END;
    HANDLER(OPC_CMP_ABS_Z);
    _COMP_Z(A, ABS, 4, 3);
    END;
#endif
    HANDLER(OPC_CMP_ABS_ALL);
    _COMP_ALL(A, ABS, 4, 3);
    END;

  /* CMP ABS,x */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_CMP_ABS_X);
    _COMP(A, ABSX, 4, 3);
    END;
    HANDLER(OPC_CMP_ABS_X_C);
    _COMP_C(A, ABSX, 4, 3);
    END;
    HANDLER(OPC_CMP_ABS_X_N);
    _COMP_N(A, ABSX, 4, 3);
    END;
    HANDLER(OPC_CMP_ABS_X_Z);
    _COMP_Z(A, ABSX, 4, 3);
    END;
#endif
    HANDLER(OPC_CMP_ABS_X_ALL);
    _COMP_ALL(A, ABSX, 4, 3);
    END;

  /* CMP ABS,Y */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_CMP_ABS_Y);
    _COMP(A, ABSY, 4, 3);
    END;
    HANDLER(OPC_CMP_ABS_Y_C);
    _COMP_C(A, ABSY, 4, 3);
    END;
    HANDLER(OPC_CMP_ABS_Y_N);
    _COMP_N(A, ABSY, 4, 3);
    END;
    HANDLER(OPC_CMP_ABS_Y_Z);
    _COMP_Z(A, ABSY, 4, 3);
    END;
#endif
    HANDLER(OPC_CMP_ABS_Y_ALL);
    _COMP_ALL(A, ABSY, 4, 3);
    END;

  /* CMPSAFE ABS */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_CMPSAFE_ABS);
    _COMPSAFE(A, ABS, 4, 3);
    END;
    HANDLER(OPC_CMPSAFE_ABS_C);
    _COMPSAFE_C(A, ABS, 4, 3);
    END;
    HANDLER(OPC_CMPSAFE_ABS_N);
    _COMPSAFE_N(A, ABS, 4, 3);
    END;
    HANDLER(OPC_CMPSAFE_ABS_Z);
    _COMPSAFE_Z(A, ABS, 4, 3);
    END;
#endif
    HANDLER(OPC_CMPSAFE_ABS_ALL);
    _COMPSAFE_ALL(A, ABS, 4, 3);
    END;

  /* CMPSAFE ABS,x */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_CMPSAFE_ABS_X);
    _COMPSAFE(A, ABSX, 4, 3);
    END;
    HANDLER(OPC_CMPSAFE_ABS_X_C);
    _COMPSAFE_C(A, ABSX, 4, 3);
    END;
    HANDLER(OPC_CMPSAFE_ABS_X_N);
    _COMPSAFE_N(A, ABSX, 4, 3);
    END;
    HANDLER(OPC_CMPSAFE_ABS_X_Z);
    _COMPSAFE_Z(A, ABSX, 4, 3);
    END;
#endif
    HANDLER(OPC_CMPSAFE_ABS_X_ALL);
    _COMPSAFE_ALL(A, ABSX, 4, 3);
    END;

  /* CMPSAFE ABS,Y */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_CMPSAFE_ABS_Y);
    _COMPSAFE(A, ABSY, 4, 3);
    END;
    HANDLER(OPC_CMPSAFE_ABS_Y_C);
    _COMPSAFE_C(A, ABSY, 4, 3);
    END;
    HANDLER(OPC_CMPSAFE_ABS_Y_N);
    _COMPSAFE_N(A, ABSY, 4, 3);
    END;
    HANDLER(OPC_CMPSAFE_ABS_Y_Z);
    _COMPSAFE_Z(A, ABSY, 4, 3);
    END;
#endif
    HANDLER(OPC_CMPSAFE_ABS_Y_ALL);
    _COMPSAFE_ALL(A, ABSY, 4, 3);
    END;

  /* CMP (zp)  [65c02] */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_CMP_IND);
    _COMP(A, IND, 5, 2);
    END;
    HANDLER(OPC_CMP_IND_C);
    _COMP_C(A, IND, 5, 2);
    END;
    HANDLER(OPC_CMP_IND_N);
    _COMP_N(A, IND, 5, 2);
    END;
    HANDLER(OPC_CMP_IND_Z);
    _COMP_Z(A, IND, 5, 2);
    END;
#endif
    HANDLER(OPC_CMP_IND_ALL);
    _COMP_ALL(A, IND, 5, 2);
    END;

  /* CMP (zp,x) */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_CMP_IND_X);
    _COMP(A, INDX, 6, 2);
    END;
    HANDLER(OPC_CMP_IND_X_C);
    _COMP_C(A, INDX, 6, 2);
    END;
    HANDLER(OPC_CMP_IND_X_N);
    _COMP_N(A, INDX, 6, 2);
    END;
    HANDLER(OPC_CMP_IND_X_Z);
    _COMP_Z(A, INDX, 6, 2);
    END;
#endif
    HANDLER(OPC_CMP_IND_X_ALL);
    _COMP_ALL(A, INDX, 6, 2);
    END;

  /* CMP (zp),y */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_CMP_IND_Y);
    _COMP(A, INDY, 5, 2);
    END;
    HANDLER(OPC_CMP_IND_Y_C);
    _COMP_C(A, INDY, 5, 2);
    END;
    HANDLER(OPC_CMP_IND_Y_N);
    _COMP_N(A, INDY, 5, 2);
    END;
    HANDLER(OPC_CMP_IND_Y_Z);
    _COMP_Z(A, INDY, 5, 2);
    END;
#endif
    HANDLER(OPC_CMP_IND_Y_ALL);
    _COMP_ALL(A, INDY, 5, 2);
    END;


  /* CPX #immediate */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_CPX_IMM);
    _COMP(X, IMM, 2, 2);
    END;
    HANDLER(OPC_CPX_IMM_C);
    _COMP_C(X, IMM, 2, 2);
    END;
    HANDLER(OPC_CPX_IMM_N);
    _COMP_N(X, IMM, 2, 2);
    END;
    HANDLER(OPC_CPX_IMM_Z);
    _COMP_Z(X, IMM, 2, 2);
    END;
#endif
    HANDLER(OPC_CPX_IMM_ALL);
    _COMP_ALL(X, IMM, 2, 2);
    END;

  /* CPX ZP */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_CPX_ZP);
    _COMP(X, ZP, 3, 2);
    END;
    HANDLER(OPC_CPX_ZP_C);
    _COMP_C(X, ZP, 3, 2);
    END;
    HANDLER(OPC_CPX_ZP_N);
    _COMP_N(X, ZP, 3, 2);
    END;
    HANDLER(OPC_CPX_ZP_Z);
    _COMP_Z(X, ZP, 3, 2);
    END;
#endif
    HANDLER(OPC_CPX_ZP_ALL);
    _COMP_ALL(X, ZP, 3, 2);
    END;

  /* CPX ABS */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_CPX_ABS);
    _COMP(X, ABS, 4, 3);
    END;
    HANDLER(OPC_CPX_ABS_C);
    _COMP_C(X, ABS, 4, 3);
    END;
    HANDLER(OPC_CPX_ABS_N);
    _COMP_N(X, ABS, 4, 3);
    END;
    HANDLER(OPC_CPX_ABS_Z);
    _COMP_Z(X, ABS, 4, 3);
    END;
#endif
    HANDLER(OPC_CPX_ABS_ALL);
    _COMP_ALL(X, ABS, 4, 3);
    END;

  /* CPXSAFE ABS */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_CPXSAFE_ABS);
    _COMPSAFE(X, ABS, 4, 3);
    END;
    HANDLER(OPC_CPXSAFE_ABS_C);
    _COMPSAFE_C(X, ABS, 4, 3);
    END;
    HANDLER(OPC_CPXSAFE_ABS_N);
    _COMPSAFE_N(X, ABS, 4, 3);
    END;
    HANDLER(OPC_CPXSAFE_ABS_Z);
    _COMPSAFE_Z(X, ABS, 4, 3);
    END;
#endif
    HANDLER(OPC_CPXSAFE_ABS_ALL);
    _COMPSAFE_ALL(X, ABS, 4, 3);
    END;


  /* CPY #immediate */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_CPY_IMM);
    _COMP(Y, IMM, 2, 2);
    END;
    HANDLER(OPC_CPY_IMM_C);
    _COMP_C(Y, IMM, 2, 2);
    END;
    HANDLER(OPC_CPY_IMM_N);
    _COMP_N(Y, IMM, 2, 2);
    END;
    HANDLER(OPC_CPY_IMM_Z);
    _COMP_Z(Y, IMM, 2, 2);
    END;
#endif
    HANDLER(OPC_CPY_IMM_ALL);
    _COMP_ALL(Y, IMM, 2, 2);
    END;

  /* CPY ZP */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_CPY_ZP);
    _COMP(Y, ZP, 3, 2);
    END;
    HANDLER(OPC_CPY_ZP_C);
    _COMP_C(Y, ZP, 3, 2);
    END;
    HANDLER(OPC_CPY_ZP_N);
    _COMP_N(Y, ZP, 3, 2);
    END;
    HANDLER(OPC_CPY_ZP_Z);
    _COMP_Z(Y, ZP, 3, 2);
    END;
#endif
    HANDLER(OPC_CPY_ZP_ALL);
    _COMP_ALL(Y, ZP, 3, 2);
    END;

  /* CPY ABS */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_CPY_ABS);
    _COMP(Y, ABS, 4, 3);
    END;
    HANDLER(OPC_CPY_ABS_C);
    _COMP_C(Y, ABS, 4, 3);
    END;
    HANDLER(OPC_CPY_ABS_N);
    _COMP_N(Y, ABS, 4, 3);
    END;
    HANDLER(OPC_CPY_ABS_Z);
    _COMP_Z(Y, ABS, 4, 3);
    END;
#endif
    HANDLER(OPC_CPY_ABS_ALL);
    _COMP_ALL(Y, ABS, 4, 3);
    END;

  /* CPYSAFE ABS */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_CPYSAFE_ABS);
    _COMPSAFE(Y, ABS, 4, 3);
    END;
    HANDLER(OPC_CPYSAFE_ABS_C);
    _COMPSAFE_C(Y, ABS, 4, 3);
    END;
    HANDLER(OPC_CPYSAFE_ABS_N);
    _COMPSAFE_N(Y, ABS, 4, 3);
    END;
    HANDLER(OPC_CPYSAFE_ABS_Z);
    _COMPSAFE_Z(Y, ABS, 4, 3);
    END;
#endif
    HANDLER(OPC_CPYSAFE_ABS_ALL);
    _COMPSAFE_ALL(Y, ABS, 4, 3);
    END;


#define _ADDCONST(n,mode,cyc,size) temp_uchar = read (mode) + (n); \
      write (mode, temp_uchar); ADDCYC(cyc); PC += size
#define _ADDCONST_N(n,mode,cyc,size) temp_uchar = read (mode) + (n); \
      write (mode, temp_uchar); SET_N (temp_uchar); ADDCYC(cyc); PC += size
#define _ADDCONST_Z(n,mode,cyc,size) temp_uchar = read (mode) + (n); \
      write (mode, temp_uchar); SET_Z (temp_uchar); ADDCYC(cyc); PC += size
#define _ADDCONST_ALL(n,mode,cyc,size) temp_uchar = read (mode) + (n); \
      write (mode, temp_uchar); SET_N (temp_uchar); SET_Z (temp_uchar); \
	ADDCYC(cyc); PC += size

  /* DEC A  [65c02] */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_DEC_ACC);
    _ADDCONST(-1, AREG, 2, 1);
    END;
    HANDLER(OPC_DEC_ACC_N);
    _ADDCONST_N(-1, AREG, 2, 1);
    END;
    HANDLER(OPC_DEC_ACC_Z);
    _ADDCONST_Z(-1, AREG, 2, 1);
    END;
#endif
    HANDLER(OPC_DEC_ACC_ALL);
    _ADDCONST_ALL(-1, AREG, 2, 1);
    END;

#ifdef USE_CC_VARIANTS
    HANDLER(OPC_DEC_ZP);
    _ADDCONST(-1, ZP, 5, 2);
    END;
    HANDLER(OPC_DEC_ZP_N);
    _ADDCONST_N(-1, ZP, 5, 2);
    END;
    HANDLER(OPC_DEC_ZP_Z);
    _ADDCONST_Z(-1, ZP, 5, 2);
    END;
#endif
    HANDLER(OPC_DEC_ZP_ALL);
    _ADDCONST_ALL(-1, ZP, 5, 2);
    END;

#ifdef USE_CC_VARIANTS
    HANDLER(OPC_DEC_ZP_X);
    _ADDCONST(-1, ZPX, 6, 2);
    END;
    HANDLER(OPC_DEC_ZP_X_N);
    _ADDCONST_N(-1, ZPX, 6, 2);
    END;
    HANDLER(OPC_DEC_ZP_X_Z);
    _ADDCONST_Z(-1, ZPX, 6, 2);
    END;
#endif
    HANDLER(OPC_DEC_ZP_X_ALL);
    _ADDCONST_ALL(-1, ZPX, 6, 2);
    END;

#ifdef USE_CC_VARIANTS
    HANDLER(OPC_DEC_ABS);
    _ADDCONST(-1, ABS, 6, 3);
    END;
    HANDLER(OPC_DEC_ABS_N);
    _ADDCONST_N(-1, ABS, 6, 3);
    END;
    HANDLER(OPC_DEC_ABS_Z);
    _ADDCONST_Z(-1, ABS, 6, 3);
    END;
#endif
    HANDLER(OPC_DEC_ABS_ALL);
    _ADDCONST_ALL(-1, ABS, 6, 3);
    END;

#ifdef USE_CC_VARIANTS
    HANDLER(OPC_DEC_ABS_X);
    _ADDCONST(-1, ABSX, 7, 3);
    END;
    HANDLER(OPC_DEC_ABS_X_N);
    _ADDCONST_N(-1, ABSX, 7, 3);
    END;
    HANDLER(OPC_DEC_ABS_X_Z);
    _ADDCONST_Z(-1, ABSX, 7, 3);
    END;
#endif
    HANDLER(OPC_DEC_ABS_X_ALL);
    _ADDCONST_ALL(-1, ABSX, 7, 3);
    END;


  /* DEX */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_DEX);
    _ADDCONST(-1, XREG, 2, 1);
    END;
    HANDLER(OPC_DEX_N);
    _ADDCONST_N(-1, XREG, 2, 1);
    END;
    HANDLER(OPC_DEX_Z);
    _ADDCONST_Z(-1, XREG, 2, 1);
    END;
#endif
    HANDLER(OPC_DEX_ALL);
    _ADDCONST_ALL(-1, XREG, 2, 1);
    END;


  /* DEY */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_DEY);
    _ADDCONST(-1, YREG, 2, 1);
    END;
    HANDLER(OPC_DEY_N);
    _ADDCONST_N(-1, YREG, 2, 1);
    END;
    HANDLER(OPC_DEY_Z);
    _ADDCONST_Z(-1, YREG, 2, 1);
    END;
#endif
    HANDLER(OPC_DEY_ALL);
    _ADDCONST_ALL(-1, YREG, 2, 1);
    END;


  /* INC */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_INC_ACC);
    _ADDCONST(1, AREG, 2, 1);
    END;
    HANDLER(OPC_INC_ACC_N);
    _ADDCONST_N(1, AREG, 2, 1);
    END;
    HANDLER(OPC_INC_ACC_Z);
    _ADDCONST_Z(1, AREG, 2, 1);
    END;
#endif
    HANDLER(OPC_INC_ACC_ALL);
    _ADDCONST_ALL(1, AREG, 2, 1);
    END;

#ifdef USE_CC_VARIANTS
    HANDLER(OPC_INC_ZP);
    _ADDCONST(1, ZP, 5, 2);
    END;
    HANDLER(OPC_INC_ZP_N);
    _ADDCONST_N(1, ZP, 5, 2);
    END;
    HANDLER(OPC_INC_ZP_Z);
    _ADDCONST_Z(1, ZP, 5, 2);
    END;
#endif
    HANDLER(OPC_INC_ZP_ALL);
    _ADDCONST_ALL(1, ZP, 5, 2);
    END;

#ifdef USE_CC_VARIANTS
    HANDLER(OPC_INC_ZP_X);
    _ADDCONST(1, ZPX, 6, 2);
    END;
    HANDLER(OPC_INC_ZP_X_N);
    _ADDCONST_N(1, ZPX, 6, 2);
    END;
    HANDLER(OPC_INC_ZP_X_Z);
    _ADDCONST_Z(1, ZPX, 6, 2);
    END;
#endif
    HANDLER(OPC_INC_ZP_X_ALL);
    _ADDCONST_ALL(1, ZPX, 6, 2);
    END;

#ifdef USE_CC_VARIANTS
    HANDLER(OPC_INC_ABS);
    _ADDCONST(1, ABS, 6, 3);
    END;
    HANDLER(OPC_INC_ABS_N);
    _ADDCONST_N(1, ABS, 6, 3);
    END;
    HANDLER(OPC_INC_ABS_Z);
    _ADDCONST_Z(1, ABS, 6, 3);
    END;
#endif
    HANDLER(OPC_INC_ABS_ALL);
    _ADDCONST_ALL(1, ABS, 6, 3);
    END;

#ifdef USE_CC_VARIANTS
    HANDLER(OPC_INC_ABS_X);
    _ADDCONST(1, ABSX, 7, 3);
    END;
    HANDLER(OPC_INC_ABS_X_N);
    _ADDCONST_N(1, ABSX, 7, 3);
    END;
    HANDLER(OPC_INC_ABS_X_Z);
    _ADDCONST_Z(1, ABSX, 7, 3);
    END;
#endif
    HANDLER(OPC_INC_ABS_X_ALL);
    _ADDCONST_ALL(1, ABSX, 7, 3);
    END;


  /* INX */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_INX);
    _ADDCONST(1, XREG, 2, 1);
    END;
    HANDLER(OPC_INX_N);
    _ADDCONST_N(1, XREG, 2, 1);
    END;
    HANDLER(OPC_INX_Z);
    _ADDCONST_Z(1, XREG, 2, 1);
    END;
#endif
    HANDLER(OPC_INX_ALL);
    _ADDCONST_ALL(1, XREG, 2, 1);
    END;


  /* INY */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_INY);
    _ADDCONST(1, YREG, 2, 1);
    END;
    HANDLER(OPC_INY_N);
    _ADDCONST_N(1, YREG, 2, 1);
    END;
    HANDLER(OPC_INY_Z);
    _ADDCONST_Z(1, YREG, 2, 1);
    END;
#endif
    HANDLER(OPC_INY_ALL);
    _ADDCONST_ALL(1, YREG, 2, 1);
    END;


  /* JMP */
    HANDLER(OPC_JMP_ABS);
    ADDCYC(3);
    PC = USHORTARG;
    RELOAD_PC_BASE();
    CHECK_FOR_INTERRUPT();
    END;
    HANDLER(OPC_JMP_ABS_NEWBANK);
    PC = USHORTARG;
    ADDCYC(3);
    RELOAD_PC_BASE();
    CHECK_FOR_INTERRUPT();
    END;
    HANDLER(OPC_JMP_ABSIND);
    PC = (READADDR(USHORTARG)
	  | (READADDR((unsigned short)(USHORTARG + 1)) << 8));
    ADDCYC(5);
    RELOAD_PC_BASE();
    CHECK_FOR_INTERRUPT();
    END;
    HANDLER(OPC_JMP_ABSIND_X);	/* [65c02] */
    {
      Reg                 temp_reg = x;

      temp_reg.s.s += USHORTARG;
      pc.c.c = READADDR(temp_reg.l);
      ++temp_reg.s.s;
      pc.c.hi = READADDR(temp_reg.l);
      ADDCYC(6);
    }
    RELOAD_PC_BASE();
    CHECK_FOR_INTERRUPT();
    END;

  /* JSR */
    HANDLER(OPC_JSR);
    PC += 2;
    PUSH(PC >> 8);
    PUSH(PC);
    ADDCYC(6);
    PC = USHORTARG;
    CHECK_FOR_INTERRUPT();
    END;
    HANDLER(OPC_JSR_NEWBANK);
    PC += 2;
    PUSH(PC >> 8);
    PUSH(PC);
    ADDCYC(6);
    PC = USHORTARG;
    RELOAD_PC_BASE();
    CHECK_FOR_INTERRUPT();
    END;


#define _LOAD(dest,mode,cyc,size) dest = read (mode); ADDCYC(cyc); PC += size
#define _LOAD_N(dest,mode,cyc,size) SET_N (dest = read (mode)); \
      ADDCYC (cyc); PC += size
#define _LOAD_Z(dest,mode,cyc,size) SET_Z (dest = read (mode)); \
	ADDCYC (cyc); PC += size
#define _LOAD_ALL(dest,mode,cyc,size) SET_N (dest = read (mode)); \
	  SET_Z (dest); ADDCYC (cyc); PC += size

#define _LOADSAFE(dest,mode,cyc,size) dest = readsafe (mode); \
	    ADDCYC(cyc); PC += size
#define _LOADSAFE_N(dest,mode,cyc,size) SET_N (dest = readsafe (mode)); \
      ADDCYC (cyc); PC += size
#define _LOADSAFE_Z(dest,mode,cyc,size) SET_Z (dest = readsafe (mode)); \
	ADDCYC (cyc); PC += size
#define _LOADSAFE_ALL(dest,mode,cyc,size) SET_N (dest = readsafe (mode)); \
	  SET_Z (dest); ADDCYC (cyc); PC += size

  /* LDA #imm */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_LDA_IMM);
    _LOAD(A, IMM, 2, 2);
    END;
    HANDLER(OPC_LDA_IMM_N);
    _LOAD_N(A, IMM, 2, 2);
    END;
    HANDLER(OPC_LDA_IMM_Z);
    _LOAD_Z(A, IMM, 2, 2);
    END;
#endif
    HANDLER(OPC_LDA_IMM_ALL);
    _LOAD_ALL(A, IMM, 2, 2);
    END;

  /* LDA zp */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_LDA_ZP);
    _LOAD(A, ZP, 3, 2);
    END;
    HANDLER(OPC_LDA_ZP_N);
    _LOAD_N(A, ZP, 3, 2);
    END;
    HANDLER(OPC_LDA_ZP_Z);
    _LOAD_Z(A, ZP, 3, 2);
    END;
#endif
    HANDLER(OPC_LDA_ZP_ALL);
    _LOAD_ALL(A, ZP, 3, 2);
    END;

  /* LDA zp,x */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_LDA_ZP_X);
    _LOAD(A, ZPX, 4, 2);
    END;
    HANDLER(OPC_LDA_ZP_X_N);
    _LOAD_N(A, ZPX, 4, 2);
    END;
    HANDLER(OPC_LDA_ZP_X_Z);
    _LOAD_Z(A, ZPX, 4, 2);
    END;
#endif
    HANDLER(OPC_LDA_ZP_X_ALL);
    _LOAD_ALL(A, ZPX, 4, 2);
    END;

  /* LDA ABS */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_LDA_ABS);
    _LOAD(A, ABS, 4, 3);
    END;
    HANDLER(OPC_LDA_ABS_N);
    _LOAD_N(A, ABS, 4, 3);
    END;
    HANDLER(OPC_LDA_ABS_Z);
    _LOAD_Z(A, ABS, 4, 3);
    END;
#endif
    HANDLER(OPC_LDA_ABS_ALL);
    _LOAD_ALL(A, ABS, 4, 3);
    END;

  /* LDA ABS,x */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_LDA_ABS_X);
    _LOAD(A, ABSX, 4, 3);
    END;
    HANDLER(OPC_LDA_ABS_X_N);
    _LOAD_N(A, ABSX, 4, 3);
    END;
    HANDLER(OPC_LDA_ABS_X_Z);
    _LOAD_Z(A, ABSX, 4, 3);
    END;
#endif
    HANDLER(OPC_LDA_ABS_X_ALL);
    _LOAD_ALL(A, ABSX, 4, 3);
    END;

  /* LDA ABS,y */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_LDA_ABS_Y);
    _LOAD(A, ABSY, 4, 3);
    END;
    HANDLER(OPC_LDA_ABS_Y_N);
    _LOAD_N(A, ABSY, 4, 3);
    END;
    HANDLER(OPC_LDA_ABS_Y_Z);
    _LOAD_Z(A, ABSY, 4, 3);
    END;
#endif
    HANDLER(OPC_LDA_ABS_Y_ALL);
    _LOAD_ALL(A, ABSY, 4, 3);
    END;

  /* LDASAFE ABS */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_LDASAFE_ABS);
    _LOADSAFE(A, ABS, 4, 3);
    END;
    HANDLER(OPC_LDASAFE_ABS_N);
    _LOADSAFE_N(A, ABS, 4, 3);
    END;
    HANDLER(OPC_LDASAFE_ABS_Z);
    _LOADSAFE_Z(A, ABS, 4, 3);
    END;
#endif
    HANDLER(OPC_LDASAFE_ABS_ALL);
    _LOADSAFE_ALL(A, ABS, 4, 3);
    END;

  /* LDASAFE ABS,x */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_LDASAFE_ABS_X);
    _LOADSAFE(A, ABSX, 4, 3);
    END;
    HANDLER(OPC_LDASAFE_ABS_X_N);
    _LOADSAFE_N(A, ABSX, 4, 3);
    END;
    HANDLER(OPC_LDASAFE_ABS_X_Z);
    _LOADSAFE_Z(A, ABSX, 4, 3);
    END;
#endif
    HANDLER(OPC_LDASAFE_ABS_X_ALL);
    _LOADSAFE_ALL(A, ABSX, 4, 3);
    END;

  /* LDASAFE ABS,y */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_LDASAFE_ABS_Y);
    _LOADSAFE(A, ABSY, 4, 3);
    END;
    HANDLER(OPC_LDASAFE_ABS_Y_N);
    _LOADSAFE_N(A, ABSY, 4, 3);
    END;
    HANDLER(OPC_LDASAFE_ABS_Y_Z);
    _LOADSAFE_Z(A, ABSY, 4, 3);
    END;
#endif
    HANDLER(OPC_LDASAFE_ABS_Y_ALL);
    _LOADSAFE_ALL(A, ABSY, 4, 3);
    END;

  /* LDA (zp) */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_LDA_IND);
    _LOAD(A, IND, 5, 2);
    END;
    HANDLER(OPC_LDA_IND_N);
    _LOAD_N(A, IND, 5, 2);
    END;
    HANDLER(OPC_LDA_IND_Z);
    _LOAD_Z(A, IND, 5, 2);
    END;
#endif
    HANDLER(OPC_LDA_IND_ALL);
    _LOAD_ALL(A, IND, 5, 2);
    END;

  /* LDA (zp,x) */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_LDA_IND_X);
    _LOAD(A, INDX, 6, 2);
    END;
    HANDLER(OPC_LDA_IND_X_N);
    _LOAD_N(A, INDX, 6, 2);
    END;
    HANDLER(OPC_LDA_IND_X_Z);
    _LOAD_Z(A, INDX, 6, 2);
    END;
#endif
    HANDLER(OPC_LDA_IND_X_ALL);
    _LOAD_ALL(A, INDX, 6, 2);
    END;

  /* LDA (zp),y */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_LDA_IND_Y);
    _LOAD(A, INDY, 5, 2);
    END;
    HANDLER(OPC_LDA_IND_Y_N);
    _LOAD_N(A, INDY, 5, 2);
    END;
    HANDLER(OPC_LDA_IND_Y_Z);
    _LOAD_Z(A, INDY, 5, 2);
    END;
#endif
    HANDLER(OPC_LDA_IND_Y_ALL);
    _LOAD_ALL(A, INDY, 5, 2);
    END;


  /* LDX #imm */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_LDX_IMM);
    _LOAD(X, IMM, 2, 2);
    END;
    HANDLER(OPC_LDX_IMM_N);
    _LOAD_N(X, IMM, 2, 2);
    END;
    HANDLER(OPC_LDX_IMM_Z);
    _LOAD_Z(X, IMM, 2, 2);
    END;
#endif
    HANDLER(OPC_LDX_IMM_ALL);
    _LOAD_ALL(X, IMM, 2, 2);
    END;

  /* LDX zp */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_LDX_ZP);
    _LOAD(X, ZP, 3, 2);
    END;
    HANDLER(OPC_LDX_ZP_N);
    _LOAD_N(X, ZP, 3, 2);
    END;
    HANDLER(OPC_LDX_ZP_Z);
    _LOAD_Z(X, ZP, 3, 2);
    END;
#endif
    HANDLER(OPC_LDX_ZP_ALL);
    _LOAD_ALL(X, ZP, 3, 2);
    END;

  /* LDX zp,y */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_LDX_ZP_Y);
    _LOAD(X, ZPY, 4, 2);
    END;
    HANDLER(OPC_LDX_ZP_Y_N);
    _LOAD_N(X, ZPY, 4, 2);
    END;
    HANDLER(OPC_LDX_ZP_Y_Z);
    _LOAD_Z(X, ZPY, 4, 2);
    END;
#endif
    HANDLER(OPC_LDX_ZP_Y_ALL);
    _LOAD_ALL(X, ZPY, 4, 2);
    END;

  /* LDX ABS */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_LDX_ABS);
    _LOAD(X, ABS, 4, 3);
    END;
    HANDLER(OPC_LDX_ABS_N);
    _LOAD_N(X, ABS, 4, 3);
    END;
    HANDLER(OPC_LDX_ABS_Z);
    _LOAD_Z(X, ABS, 4, 3);
    END;
#endif
    HANDLER(OPC_LDX_ABS_ALL);
    _LOAD_ALL(X, ABS, 4, 3);
    END;

  /* LDX ABS,y */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_LDX_ABS_Y);
    _LOAD(X, ABSY, 4, 3);
    END;
    HANDLER(OPC_LDX_ABS_Y_N);
    _LOAD_N(X, ABSY, 4, 3);
    END;
    HANDLER(OPC_LDX_ABS_Y_Z);
    _LOAD_Z(X, ABSY, 4, 3);
    END;
#endif
    HANDLER(OPC_LDX_ABS_Y_ALL);
    _LOAD_ALL(X, ABSY, 4, 3);
    END;

  /* LDXSAFE ABS */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_LDXSAFE_ABS);
    _LOADSAFE(X, ABS, 4, 3);
    END;
    HANDLER(OPC_LDXSAFE_ABS_N);
    _LOADSAFE_N(X, ABS, 4, 3);
    END;
    HANDLER(OPC_LDXSAFE_ABS_Z);
    _LOADSAFE_Z(X, ABS, 4, 3);
    END;
#endif
    HANDLER(OPC_LDXSAFE_ABS_ALL);
    _LOADSAFE_ALL(X, ABS, 4, 3);
    END;

  /* LDXSAFE ABS,y */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_LDXSAFE_ABS_Y);
    _LOADSAFE(X, ABSY, 4, 3);
    END;
    HANDLER(OPC_LDXSAFE_ABS_Y_N);
    _LOADSAFE_N(X, ABSY, 4, 3);
    END;
    HANDLER(OPC_LDXSAFE_ABS_Y_Z);
    _LOADSAFE_Z(X, ABSY, 4, 3);
    END;
#endif
    HANDLER(OPC_LDXSAFE_ABS_Y_ALL);
    _LOADSAFE_ALL(X, ABSY, 4, 3);
    END;


  /* LDY #imm */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_LDY_IMM);
    _LOAD(Y, IMM, 2, 2);
    END;
    HANDLER(OPC_LDY_IMM_N);
    _LOAD_N(Y, IMM, 2, 2);
    END;
    HANDLER(OPC_LDY_IMM_Z);
    _LOAD_Z(Y, IMM, 2, 2);
    END;
#endif
    HANDLER(OPC_LDY_IMM_ALL);
    _LOAD_ALL(Y, IMM, 2, 2);
    END;

  /* LDY zp */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_LDY_ZP);
    _LOAD(Y, ZP, 3, 2);
    END;
    HANDLER(OPC_LDY_ZP_N);
    _LOAD_N(Y, ZP, 3, 2);
    END;
    HANDLER(OPC_LDY_ZP_Z);
    _LOAD_Z(Y, ZP, 3, 2);
    END;
#endif
    HANDLER(OPC_LDY_ZP_ALL);
    _LOAD_ALL(Y, ZP, 3, 2);
    END;

  /* LDY zp,x */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_LDY_ZP_X);
    _LOAD(Y, ZPX, 4, 2);
    END;
    HANDLER(OPC_LDY_ZP_X_N);
    _LOAD_N(Y, ZPX, 4, 2);
    END;
    HANDLER(OPC_LDY_ZP_X_Z);
    _LOAD_Z(Y, ZPX, 4, 2);
    END;
#endif
    HANDLER(OPC_LDY_ZP_X_ALL);
    _LOAD_ALL(Y, ZPX, 4, 2);
    END;

  /* LDY ABS */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_LDY_ABS);
    _LOAD(Y, ABS, 4, 3);
    END;
    HANDLER(OPC_LDY_ABS_N);
    _LOAD_N(Y, ABS, 4, 3);
    END;
    HANDLER(OPC_LDY_ABS_Z);
    _LOAD_Z(Y, ABS, 4, 3);
    END;
#endif
    HANDLER(OPC_LDY_ABS_ALL);
    _LOAD_ALL(Y, ABS, 4, 3);
    END;

  /* LDY ABS,x */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_LDY_ABS_X);
    _LOAD(Y, ABSX, 4, 3);
    END;
    HANDLER(OPC_LDY_ABS_X_N);
    _LOAD_N(Y, ABSX, 4, 3);
    END;
    HANDLER(OPC_LDY_ABS_X_Z);
    _LOAD_Z(Y, ABSX, 4, 3);
    END;
#endif
    HANDLER(OPC_LDY_ABS_X_ALL);
    _LOAD_ALL(Y, ABSX, 4, 3);
    END;

  /* LDYSAFE ABS */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_LDYSAFE_ABS);
    _LOADSAFE(Y, ABS, 4, 3);
    END;
    HANDLER(OPC_LDYSAFE_ABS_N);
    _LOADSAFE_N(Y, ABS, 4, 3);
    END;
    HANDLER(OPC_LDYSAFE_ABS_Z);
    _LOADSAFE_Z(Y, ABS, 4, 3);
    END;
#endif
    HANDLER(OPC_LDYSAFE_ABS_ALL);
    _LOADSAFE_ALL(Y, ABS, 4, 3);
    END;

  /* LDYSAFE ABS,x */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_LDYSAFE_ABS_X);
    _LOADSAFE(Y, ABSX, 4, 3);
    END;
    HANDLER(OPC_LDYSAFE_ABS_X_N);
    _LOADSAFE_N(Y, ABSX, 4, 3);
    END;
    HANDLER(OPC_LDYSAFE_ABS_X_Z);
    _LOADSAFE_Z(Y, ABSX, 4, 3);
    END;
#endif
    HANDLER(OPC_LDYSAFE_ABS_X_ALL);
    _LOADSAFE_ALL(Y, ABSX, 4, 3);
    END;


  /* NOP */
    HANDLER(OPC_NOP);
    ADDCYC(2);
    PC++;
    END;


  /* PHA */
    HANDLER(OPC_PHA);
    PUSH(A);
    ADDCYC(3);
    PC++;
    END;

  /* PLA */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_PLA);
    A = POP();
    ADDCYC(4);
    PC++;
    END;
    HANDLER(OPC_PLA_N);
    SET_N(A = POP());
    ADDCYC(4);
    PC++;
    END;
    HANDLER(OPC_PLA_Z);
    SET_Z(A = POP());
    ADDCYC(4);
    PC++;
    END;
#endif
    HANDLER(OPC_PLA_ALL);
    SET_N(A = POP());
    SET_Z(A);
    ADDCYC(4);
    PC++;
    END;

  /* PHX  [65c02] */
    HANDLER(OPC_PHX);
    PUSH(X);
    ADDCYC(3);
    PC++;
    END;

  /* PLX  [65c02] */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_PLX);
    X = POP();
    ADDCYC(4);
    PC++;
    END;
    HANDLER(OPC_PLX_N);
    SET_N(X = POP());
    ADDCYC(4);
    PC++;
    END;
    HANDLER(OPC_PLX_Z);
    SET_Z(X = POP());
    ADDCYC(4);
    PC++;
    END;
#endif
    HANDLER(OPC_PLX_ALL);
    SET_N(X = POP());
    SET_Z(X);
    \
    ADDCYC(4);
    PC++;
    END;

  /* PHY  [65c02] */
    HANDLER(OPC_PHY);
    PUSH(Y);
    ADDCYC(3);
    PC++;
    END;

  /* PLY  [65c02] */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_PLY);
    Y = POP();
    ADDCYC(4);
    PC++;
    END;
    HANDLER(OPC_PLY_N);
    SET_N(Y = POP());
    ADDCYC(4);
    PC++;
    END;
    HANDLER(OPC_PLY_Z);
    SET_Z(Y = POP());
    ADDCYC(4);
    PC++;
    END;
#endif
    HANDLER(OPC_PLY_ALL);
    SET_N(Y = POP());
    SET_Z(Y);
    \
    ADDCYC(4);
    PC++;
    END;

  /* PHP */
    HANDLER(OPC_PHP);		/* If this is buggy, fix BRK/interrupt code too */
    PUSH(((NFLAG != 0) << N_BIT) | ((NZFLAG == 0) << Z_BIT)
	 | (BFLAG << B_BIT) | (CFLAG << C_BIT)
	 | (IFLAG << I_BIT) | (DFLAG << D_BIT)
	 | ((VFLAG != 0) << V_BIT) | (1 << X_BIT));
    ADDCYC(3);
    PC++;
    END;


#define _PLP() temp_uchar = POP (); IFLAG = (temp_uchar >> I_BIT) & 1; \
	DFLAG = (temp_uchar >> D_BIT) & 1; BFLAG = (temp_uchar >> B_BIT) & 1; \
	PC++; ADDCYC (4);

#ifdef USE_CC_VARIANTS
    HANDLER(OPC_PLP);
    _PLP();
    END;
    HANDLER(OPC_PLP_C);
    _PLP();
    CFLAG = (temp_uchar >> C_BIT) & 1;
    END;
    HANDLER(OPC_PLP_N);
    _PLP();
    NFLAG = (temp_uchar & (1 << N_BIT));
    END;
    HANDLER(OPC_PLP_V);
    _PLP();
    VFLAG = (temp_uchar & (1 << V_BIT));
    END;
    HANDLER(OPC_PLP_Z);
    _PLP();
    NZFLAG = !(temp_uchar & (1 << Z_BIT));
    END;
#endif
    HANDLER(OPC_PLP_ALL);
    _PLP();
    NFLAG = temp_uchar & (1 << N_BIT);
    NZFLAG = !(temp_uchar & (1 << Z_BIT));
    CFLAG = (temp_uchar >> C_BIT) & 1;
    VFLAG = temp_uchar & (1 << V_BIT);
    END;


  /* RTI */
    HANDLER(OPC_RTI);
  /* pop off CC reg... */
    _PLP();
    ADDCYC(2);			/* PLP already added 4, so 2 + 4 == 6 */
    NFLAG = temp_uchar & (1 << N_BIT);
    NZFLAG = !(temp_uchar & (1 << Z_BIT));
    CFLAG = (temp_uchar >> C_BIT) & 1;
    VFLAG = temp_uchar & (1 << V_BIT);
  /* ...and return */
    pc.c.c = POP(), pc.c.hi = POP();
    RELOAD_PC_BASE();
    CHECK_FOR_INTERRUPT();
    END;


  /* RTS */
    HANDLER(OPC_RTS);
    {
      ++sp.c.c;			/* Increment twice ... */
      pc.c.c = READZPADDR(sp.l);
      ++sp.c.c;			/* ...in case it wraps around. */
      pc.c.hi = READZPADDR(sp.l);
      ADDCYC(6);
      ++PC;
    }
    RELOAD_PC_BASE();
    CHECK_FOR_INTERRUPT();
    END;




  /* SEC */
    HANDLER(OPC_SEC);
    CFLAG = 1;
    ADDCYC(2);
    PC++;
    END;
    HANDLER(OPC_SED);
    DFLAG = 1;
    ADDCYC(2);
    PC++;
    END;
    HANDLER(OPC_SEI);
    IFLAG = 1;
    ADDCYC(2);
    PC++;
    END;


#define _STORE(val,mode,cyc,size) write (mode, val); ADDCYC (cyc); PC += size

  /* STA */
    HANDLER(OPC_STA_ZP);
    _STORE(A, ZP, 3, 2);
    END;
    HANDLER(OPC_STA_ZP_X);
    _STORE(A, ZPX, 4, 2);
    END;
    HANDLER(OPC_STA_ABS);
    _STORE(A, ABS, 4, 3);
    END;
    HANDLER(OPC_STA_ABS_X);
    _STORE(A, ABSX, 5, 3);
    END;
    HANDLER(OPC_STA_ABS_Y);
    _STORE(A, ABSY, 5, 3);
    END;
    HANDLER(OPC_STA_IND);
    _STORE(A, IND, 5, 2);
    END;			/* [65c02] */
    HANDLER(OPC_STA_IND_X);
    _STORE(A, INDX, 6, 2);
    END;
    HANDLER(OPC_STA_IND_Y);
    _STORE(A, INDY, 6, 2);
    END;

  /* STX */
    HANDLER(OPC_STX_ZP);
    _STORE(X, ZP, 3, 2);
    END;
    HANDLER(OPC_STX_ZP_Y);
    _STORE(X, ZPY, 4, 2);
    END;
    HANDLER(OPC_STX_ABS);
    _STORE(X, ABS, 4, 3);
    END;

  /* STY */
    HANDLER(OPC_STY_ZP);
    _STORE(Y, ZP, 3, 2);
    END;
    HANDLER(OPC_STY_ZP_X);
    _STORE(Y, ZPX, 4, 2);
    END;
    HANDLER(OPC_STY_ABS);
    _STORE(Y, ABS, 4, 3);
    END;

  /* STZ  [65c02] */
    HANDLER(OPC_STZ_ZP);
    _STORE(0, ZP, 3, 2);
    END;
    HANDLER(OPC_STZ_ZP_X);
    _STORE(0, ZPX, 4, 2);
    END;
    HANDLER(OPC_STZ_ABS);
    _STORE(0, ABS, 4, 3);
    END;
    HANDLER(OPC_STZ_ABS_X);
    _STORE(0, ABSX, 5, 3);
    END;


  /* Tjk */
#define _XFER(from,to) to = from; ADDCYC (2); PC++
#define _XFER_N(from,to) SET_N (to = from); ADDCYC (2); PC++
#define _XFER_Z(from,to) SET_Z (to = from); ADDCYC (2); PC++
#define _XFER_ALL(from,to) SET_N (to = from); SET_Z (to); ADDCYC (2); PC++

  /* TAX */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_TAX);
    _XFER(A, X);
    END;
    HANDLER(OPC_TAX_N);
    _XFER_N(A, X);
    END;
    HANDLER(OPC_TAX_Z);
    _XFER_Z(A, X);
    END;
#endif
    HANDLER(OPC_TAX_ALL);
    _XFER_ALL(A, X);
    END;

  /* TAY */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_TAY);
    _XFER(A, Y);
    END;
    HANDLER(OPC_TAY_N);
    _XFER_N(A, Y);
    END;
    HANDLER(OPC_TAY_Z);
    _XFER_Z(A, Y);
    END;
#endif
    HANDLER(OPC_TAY_ALL);
    _XFER_ALL(A, Y);
    END;

  /* TSX */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_TSX);
    _XFER(SP, X);
    END;
    HANDLER(OPC_TSX_N);
    _XFER_N(SP, X);
    END;
    HANDLER(OPC_TSX_Z);
    _XFER_Z(SP, X);
    END;
#endif
    HANDLER(OPC_TSX_ALL);
    _XFER_ALL(SP, X);
    END;

  /* TXA */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_TXA);
    _XFER(X, A);
    END;
    HANDLER(OPC_TXA_N);
    _XFER_N(X, A);
    END;
    HANDLER(OPC_TXA_Z);
    _XFER_Z(X, A);
    END;
#endif
    HANDLER(OPC_TXA_ALL);
    _XFER_ALL(X, A);
    END;

  /* TXS */
    HANDLER(OPC_TXS);
    SP = X;
    ADDCYC(2);
    PC++;
    END;			/* Sets no cc bits! */

  /* TYA */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_TYA);
    _XFER(Y, A);
    END;
    HANDLER(OPC_TYA_N);
    _XFER_N(Y, A);
    END;
    HANDLER(OPC_TYA_Z);
    _XFER_Z(Y, A);
    END;
#endif
    HANDLER(OPC_TYA_ALL);
    _XFER_ALL(Y, A);
    END;


  /* TSB  [65c02] */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_TSB_ZP);
    write(ZP, A | read(ZP));
    ADDCYC(4);
    PC += 2;
    END;
#endif
    HANDLER(OPC_TSB_ZP_ALL);
    temp_uchar = read(ZP);
    SET_Z(A & temp_uchar);
    write(ZP, A | temp_uchar);
    ADDCYC(4);
    PC += 2;
    END;
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_TSB_ABS);
    write(ABS, A | read(ABS));
    ADDCYC(5);
    PC += 3;
    END;
#endif
    HANDLER(OPC_TSB_ABS_ALL);
    temp_uchar = read(ABS);
    SET_Z(A & temp_uchar);
    write(ABS, A | temp_uchar);
    ADDCYC(5);
    PC += 3;
    END;

  /* TRB  [65c02] */
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_TRB_ZP);
    write(ZP, (~A) & read(ZP));
    ADDCYC(4);
    PC += 2;
    END;
#endif
    HANDLER(OPC_TRB_ZP_ALL);
    temp_uchar = read(ZP);
    SET_Z(A & temp_uchar);
    write(ZP, (~A) & temp_uchar);
    ADDCYC(4);
    PC += 2;
    END;
#ifdef USE_CC_VARIANTS
    HANDLER(OPC_TRB_ABS);
    write(ABS, (~A) & read(ABS));
    ADDCYC(5);
    PC += 3;
    END;
#endif
    HANDLER(OPC_TRB_ABS_ALL);
    temp_uchar = read(ABS);
    SET_Z(A & temp_uchar);
    write(ABS, (~A) & temp_uchar);
    ADDCYC(5);
    PC += 3;
    END;

handle_interrupt:
    {
      int                 n = state->interrupt_pending;

      if (n & POWEROFF_PENDING)
      {
	SAVE_MACHINE_STATE();
	mutex_lock(state->interrupt_mutex);
	state->interrupt_pending &= ~POWEROFF_PENDING;
	mutex_unlock(state->interrupt_mutex);
	return;
      }
      else
      if (n & RESET_PENDING)
      {
      /* Set up memory correctly (like later Apple //'s seem to). */
	SAVE_RW_HANDLER_INFO(state);
	write_mem_byte(state, 0xC000, 0);	/* Turn 80store off.         */
	write_mem_byte(state, 0xC089, 0);	/* Read/write ROMs.          */
	write_mem_byte(state, 0xC008, 0);	/* Turn AltZP off.           */
	write_mem_byte(state, 0xC002, 0);	/* Read main memory.         */
	write_mem_byte(state, 0xC004, 0);	/* Write enable main memory. */
	LOAD_RW_HANDLER_INFO(state);

      /* Fetch interrupt vector from ROM. */
	pc.c.c = state->sysmem[ROM + 0x2FFC].mem.byte;
	pc.c.hi = state->sysmem[ROM + 0x2FFD].mem.byte;

      /*
       * Clean up pc base since pc changed in an unpredictable
       * way. 
       */
	RELOAD_PC_BASE();

      /* Turn of reset request. */
	mutex_lock(state->interrupt_mutex);
	state->interrupt_pending &= ~RESET_PENDING;
	mutex_unlock(state->interrupt_mutex);
      }
      else
      if (n & NMI_PENDING)
      {
	PUSH(PC);
	PUSH(PC >> 8);
	PUSH(((NFLAG != 0) << N_BIT) | ((NZFLAG == 0) << Z_BIT)
	     | (BFLAG << B_BIT) | (CFLAG << C_BIT)
	     | (IFLAG << I_BIT) | (DFLAG << D_BIT)
	     | ((VFLAG != 0) << V_BIT) | (1 << X_BIT));
	IFLAG = 1;
	pc.c.c = READADDR(0xFFFA);
	pc.c.hi = READADDR(0xFFFB);
	RELOAD_PC_BASE();
	mutex_lock(state->interrupt_mutex);
	state->interrupt_pending &= ~NMI_PENDING;
	mutex_unlock(state->interrupt_mutex);
      }
      else
      if ((n & IRQ_PENDING) && !IFLAG)
      {
	PUSH(PC);
	PUSH(PC >> 8);
	PUSH(((NFLAG != 0) << N_BIT) | ((NZFLAG == 0) << Z_BIT)
	     | (BFLAG << B_BIT) | (CFLAG << C_BIT)
	     | (IFLAG << I_BIT) | (DFLAG << D_BIT)
	     | ((VFLAG != 0) << V_BIT) | (1 << X_BIT));
	IFLAG = 1;
	BFLAG = 0;
	pc.c.c = READADDR(0xFFFE);
	pc.c.hi = READADDR(0xFFFF);
	RELOAD_PC_BASE();
	mutex_lock(state->interrupt_mutex);
	state->interrupt_pending &= ~IRQ_PENDING;
	mutex_unlock(state->interrupt_mutex);
      }
      else
      if (n & PAUSE_PENDING)
      {
      /* Turn off the paused bit. */
	mutex_lock(state->interrupt_mutex);
	state->interrupt_pending &= ~PAUSE_PENDING;
	mutex_unlock(state->interrupt_mutex);

#if NX_CURRENT_COMPILER_RELEASE > 310
	[state->pauseLock unlockWith:CONDITION_MACHINE_PAUSED];
	[state->pauseLock lockWhen:CONDITION_MACHINE_RUNNING];
#else				/* FIXME!  Well, fix gcc. */
	objc_msgSend(state->pauseLock, sel_getUid("unlockWith:"),
		     CONDITION_MACHINE_PAUSED);
	objc_msgSend(state->pauseLock, sel_getUid("lockWhen:"),
		     CONDITION_MACHINE_RUNNING);
#endif


      /*
       * Make sure that this thread can be paused then killed
       * without letting it run any more code. 
       */
	if (state->interrupt_pending)
	  goto handle_interrupt;
      }

      continue;
    }
  }
}
