/*
 * This file is part of zaniWok.
 *
 * zaniWok is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.  See the file COPYING for more information.
 *
 * zaniWok is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */


/* This should ONLY be included from execute.c! */
#include "execute.h"

/* The first four have been grouped together in one struct to help free up
 * registers for other uses.  Because of the simple ways most of them are
 * manipulated, it seems that this should be an improvement.
 */
#define CFLAG cc_bits.c_flag
#define VFLAG cc_bits.v_flag
#define NFLAG cc_bits.n_flag
#define NZFLAG cc_bits.nz_flag
#define BFLAG b_flag
#define DFLAG d_flag
#define IFLAG i_flag


#define LOAD_MACHINE_STATE() \
  { \
    pc.s.s = state->pc; \
    sp.l   = state->sp | 0x100; \
    a.c.c  = state->a; \
    x.c.c  = state->x; \
    y.c.c  = state->y; \
    BFLAG  = state->b_flag;  \
    CFLAG  = state->c_flag;   /* Make darn sure this is 0 or 1! */ \
    DFLAG  = state->d_flag;  \
    IFLAG  = state->i_flag;  \
    NFLAG  = state->n_flag;  \
    NZFLAG = !state->z_flag; \
    VFLAG  = state->v_flag;  \
    cycles = state->cycles;  \
    read_base = state->read_base_p; \
    write_base = state->write_base_p; \
  }

#define SAVE_MACHINE_STATE() \
{ \
    state->pc = pc.s.s; \
    state->sp = sp.c.c; \
    state->a  = a.c.c;  \
    state->x  = x.c.c;  \
    state->y  = y.c.c;  \
    state->b_flag = BFLAG;   \
    state->d_flag = DFLAG;   \
    state->i_flag = IFLAG;   \
    state->c_flag = CFLAG;   \
    state->n_flag = NFLAG;   \
    state->v_flag = VFLAG;   \
    state->z_flag = !NZFLAG; \
    state->cycles = cycles;  \
}

#define RELOAD_PC_BASE() \
  (pc_base = (((void *) (read_base[pc.l >> 8] - (pc.l & ~0xFFL))) \
	      + offsetof (MemoryElement, inst)))


#define SAVE_RW_HANDLER_INFO(h) ((h)->cycles = cycles)

#define LOAD_RW_HANDLER_INFO(h) \
{ \
  read_base  = (h)->read_base_p; \
  write_base = (h)->write_base_p; \
/*  cycles     = (h)->cycles;   Ignore changes to cycles. */ \
  RELOAD_PC_BASE (); \
}

#define CHECK_FOR_INTERRUPT() \
  if (state->interrupt_pending) \
    goto handle_interrupt; 
      

/* Convenience macros. */
#define PC pc.s.s
#define SP sp.c.c
#define X x.c.c
#define Y y.c.c
#define A a.c.c
#define X_AS_USHORT x.s.s
#define Y_AS_USHORT y.s.s
#define X_AS_LONG x.l
#define Y_AS_LONG y.l

#define USHORTARG i.u.s
#define SHORTARG ((short) i.u.s)
#define BYTEARG  i.u.c.c

#define ADDCYC(n) (cycles += (n))       /* Modify to turn counting off */
#define SET_N(x) (NFLAG = (x) & 0x80)
#define SET_Z(x) (NZFLAG = (x))

/* Helper macros: */

#define READELEM(a)  ({ int _a = (a); &read_base[_a >> 8][_a & 255]; })
#define WRITEELEM(a)  ({ int _a = (a); &write_base[_a >> 8][_a & 255]; })
#define WRITEZPELEM(a) (&write_base[0][a])
#define READADDR(a)  READELEM(a)->mem.byte
#define READZPADDR(a) read_base[0][a].mem.byte
#define WRITEADDR(a,v) (WRITEELEM (a)->mem.byte = (v))
#define WRITEZPADDR(a,v) (write_base[0][a].mem.byte = (v))

#define AREGindex 0
#define XREGindex 0
#define YREGindex 0
#define TEMPUCHARindex 0
#define IMMindex 0
#define ZPindex SHORTARG
#define ZPXindex ({Reg temp_reg = x; temp_reg.c.c += BYTEARG; temp_reg.l; })
#define ZPYindex ({Reg temp_reg = y; temp_reg.c.c += BYTEARG; temp_reg.l; })
#define ABSindex USHORTARG
#define ABSXindex ({Reg temp_reg = x; temp_reg.s.s += USHORTARG; temp_reg.l; })
#define ABSYindex ({Reg temp_reg = y; temp_reg.s.s += USHORTARG; temp_reg.l; })
#define INDindex ({Reg temp_reg, temp_addr; temp_reg.l = BYTEARG; \
		     temp_addr.l = READZPADDR (temp_reg.l); \
		     ++temp_reg.c.c; \
		     temp_addr.c.hi = READZPADDR (temp_reg.l); \
		     temp_addr.l; })
#define INDXindex ({Reg temp_reg = x, temp_addr; temp_addr.l = 0; \
		    temp_reg.c.c += BYTEARG; \
		    temp_addr.c.c = READZPADDR (temp_reg.l); \
		     ++temp_reg.c.c; \
		    temp_addr.c.hi = READZPADDR (temp_reg.l); \
		    temp_addr.l; })
#define INDYindex ({register Reg addr, index; addr.l = USHORTARG; index.l = 0;\
		    index.c.c = READZPADDR (addr.l); ++addr.c.c; \
		    index.s.s += Y_AS_USHORT + (READZPADDR (addr.l) << 8); \
		    index.l; })

#define readAREG(index) A
#define readXREG(index) X
#define readYREG(index) Y
#define readTEMPUCHAR(index) temp_uchar
#define readIMM(index)  SHORTARG
#define readZP(index)   READZPADDR (index)
#define readZPX(index)  READZPADDR (index)
#define readZPY(index)  READZPADDR (index)
#define readABS(index)  READADDR (index)
#define readABSX(index) READADDR (index)
#define readABSY(index) READADDR (index)
#define readIND(index)  READADDR (index)
#define readINDX(index) READADDR (index)
#define readINDY(index) READADDR (index)

#define CHECKREAD(index) \
({ \
   struct MemInfo *m = &READELEM (index)->mem; \
   unsigned char n; \
\
   if (m->read_special) \
     { \
       SAVE_RW_HANDLER_INFO (state);\
       n = state->rw_func_list[m->read_func] (READ, (MemoryElement *) m, \
					      state); \
       LOAD_RW_HANDLER_INFO (state); \
     } \
   else n = m->byte; \
   n; \
})

#define readcheckAREG(index)      A
#define readcheckXREG(index)      X
#define readcheckYREG(index)      Y
#define readcheckIMM(index)       SHORTARG
#define readcheckTEMPUCHAR(index) temp_uchar
#define readcheckZP(index)        READZPADDR (index)
#define readcheckZPX(index)       READZPADDR (index)
#define readcheckZPY(index)       READZPADDR (index)
#define readcheckABS(index)       CHECKREAD(index)
#define readcheckABSX(index)      CHECKREAD(index)
#define readcheckABSY(index)      CHECKREAD(index)
#define readcheckIND(index)       CHECKREAD(index)
#define readcheckINDX(index)      CHECKREAD(index)
#define readcheckINDY(index)      CHECKREAD(index)

#define read(mode) ({ unsigned long tmp = mode ## index; \
		      readcheck ## mode (tmp); })
#define readsafe(mode) read ## mode (mode ## index)
#define readcheck(mode) read(mode)   /* Could make this not lookup value... */

#define CHECKWRITE(index, val) \
{ \
  struct MemInfo *m = &WRITEELEM (index)->mem; \
  m->byte = (val); \
  if (m->write_special) \
    { \
       SAVE_RW_HANDLER_INFO (state); \
       state->rw_func_list[m->write_func] (WRITE, (MemoryElement *)m, state); \
       LOAD_RW_HANDLER_INFO (state); \
    } \
}

#define CHECKWRITEZP(index, val) \
{ \
  struct MemInfo *m = &WRITEZPELEM (index)->mem; \
  m->byte = (val); \
  if (m->write_special) \
    { \
       SAVE_RW_HANDLER_INFO (state); \
       state->rw_func_list[m->write_func](WRITE, (MemoryElement *)m, state); \
       LOAD_RW_HANDLER_INFO (state); \
    } \
}


#define writeAREG(val) (A = (val))
#define writeXREG(val) (X = (val))
#define writeYREG(val) (Y = (val))
#define writeIMM(val)  (abort ())
#define writeZP(val)   CHECKWRITEZP (ZPindex,  (val))
#define writeZPX(val)  CHECKWRITEZP (ZPXindex, (val))
#define writeZPY(val)  CHECKWRITEZP (ZPYindex, (val))
#define writeABS(val)  CHECKWRITE (ABSindex,  (val))
#define writeABSX(val) CHECKWRITE (ABSXindex, (val))
#define writeABSY(val) CHECKWRITE (ABSYindex, (val))
#define writeIND(val)  CHECKWRITE (INDindex,  (val))
#define writeINDX(val) CHECKWRITE (INDXindex, (val))
#define writeINDY(val) CHECKWRITE (INDYindex, (val))

#define write(mode, val) write ## mode (val)

#define	PUSH(val) WRITEZPADDR (({unsigned long tmp = sp.l; --sp.c.c; tmp; }), \
			       (val))
#define	POP()     READZPADDR ((++sp.c.c, sp.l))

#define N_BIT 7
#define V_BIT 6
#define X_BIT 5   /* Not used;  logic 1 */
#define B_BIT 4
#define D_BIT 3
#define I_BIT 2
#define Z_BIT 1
#define C_BIT 0
