/*
 * This file is part of zaniWok.
 *
 * zaniWok is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.  See the file COPYING for more information.
 *
 * zaniWok is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */


/*
 *     memory.h
 */

#include "execute.h"

#define REALMEMSIZE 0x28000

#define AUXOFFSET 0x11000

#define INTROM       (-0x4000)   /* C100-CFFF internal ROM */
#define ROM          (-0x3100)
#define ROM2         (ROM + 0x1000)
#define JUNK         (-0x100)
#define MAINZP       0x0
#define MAIN48       0x200
#define TEXT1        0x400
#define TEXT2        0x800
#define HIRES1       0x2000
#define HIRES2       0x4000
#define SOFTSWITCHES 0xC000  /* This is guaranteed to be #defined as 0xC000  */
#define DEVICE_SELECT_ADDRESS_BEGIN(slot_number)	\
	(0xC080 + (slot_number * 0x10))
#define DEVICE_SELECT_ADDRESS_END(slot_number)		\
	(0xC080 + 0x0f + (slot_number * 0x10))
#define SWITCHES_FOR_SLOT(slot_number)			\
	DEVICE_SELECT_ADDRESS_BEGIN(slot_number)
#define SWITCHES_FOR_SLOT_END(slot_number)		\
	DEVICE_SELECT_ADDRESS_END(slot_number)
#define IO_SELECT_ADDRESS_BEGIN(slot_number)		\
        (0xC000 + (slot_number * 0x0100))
#define IO_SELECT_ADDRESS_END(slot_number)		\
        (0xC000 + 0xff + (slot_number * 0x0100))
#define ROM_FOR_SLOT(slot_number) IO_SELECT_ADDRESS_BEGIN(slot_number)
#define ROM_FOR_SLOT_END(slot_number) IO_SELECT_ADDRESS_END(slot_number)
#define SLOT1ROM     0xC100  /* The indices for the slots are guaranteed     */
#define SLOT2ROM     0xC200  /*   to be contiguous, so you can just multiply */
#define SLOT3ROM     0xC300  /*   your slot # times 0x100, etc.              */
#define SLOT4ROM     0xC400 
#define SLOT5ROM     0xC500
#define SLOT6ROM     0xC600
#define SLOT7ROM     0xC700
#define EXPROM       0xC800   /* Peripheral expansion ROM */
#define LCBANK1      0xD000   /* Language card bank 1 */
#define LCBANK2      0xE000
#define LCRAM        0xF000
#define AUXZP        (AUXOFFSET + MAINZP)
#define AUX48        (AUXOFFSET + MAIN48)
#define AUXTEXT1     (AUXOFFSET + TEXT1)
#define AUXHIRES1    (AUXOFFSET + HIRES1)
#define AUXLCBANK1   (AUXOFFSET + LCBANK1)
#define AUXLCBANK2   (AUXOFFSET + LCBANK2)
#define AUXLCRAM     (AUXOFFSET + LCRAM)

extern void write_mem_byte (MachineState *s, unsigned short addr, int x);
extern unsigned char read_mem_byte (MachineState *s, unsigned short addr);
extern unsigned short read_mem_short (MachineState *s, unsigned short addr);
extern void write_mem_short (MachineState *s, unsigned short addr, int value);
extern void copy_bytes_to_apple (MachineState *s, const unsigned char *src,
				 unsigned short dst_addr, int num_bytes);
extern void copy_bytes_from_apple (MachineState *s, unsigned short src_addr,
				   unsigned char *dst, int num_bytes);
extern MemoryElement *get_memory_element (MachineState *state,
					  unsigned short addr,AccessType type);
extern unsigned char read_memory_element (MachineState *s, MemoryElement *m);
extern void write_memory_element (MachineState *s, MemoryElement *m, int x);
