/*
 * This file is part of zaniWok.
 *
 * zaniWok is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.  See the file COPYING for more information.
 *
 * zaniWok is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include <stdio.h>
#include <stdlib.h>
#include "softswitches.h"
#include "memory.h"

static inline void
offset_page_bases(MemoryElement ** base, int offset, short num_pages)
{
  if (num_pages == 2)		/* inlining will optimize this away. */
    base[0] += offset, base[1] += offset;
  else
  if (num_pages == 4)
    base[0] += offset, base[1] += offset, base[2] += offset, base[3] += offset;
  else
    while (num_pages--)
      *base++ += offset;
}


static inline void
set_page_bases(MemoryElement ** base, MemoryElement * elem, short num_pages)
{
  if (num_pages == 2)		/* inlining will optimize this away. */
    base[0] = elem, base[1] = elem + 0x100;
  else
    while (num_pages--)
      *base++ = elem, elem += 0x100;
}


int
read_softswitch_C000_C00F(AccessType rw, MemoryElement * m,
			  MachineState * state)
{
  int                 key;

 /*
  * The keyboard value is updated by another thread, so just
  * return the current value. 
  */
  mutex_lock(state->keyboard.mutex);
  key = state->keyboard.buffer[state->keyboard.head]
    | (state->keyboard.key_ready << 7);
  mutex_unlock(state->keyboard.mutex);

  return key;
}


int
write_softswitch_C000_C00F(AccessType rw, MemoryElement * m,
			   MachineState * state)
{
  MemoryElement     **p;
  int                 n, addr = m - state->sysmem;

  switch (addr)
  {
  case 0xC000:			/* Store80Off */
  case 0xC001:			/* Store80On */
    if (state->flags.store80 != (addr & 1))
    {
      n = (state->flags.page2 != state->flags.store80)
	? AUXOFFSET : -AUXOFFSET;

      if (state->flags.page2 != state->flags.aux_read)
      {
#if 0
	offset_page_bases(&state->read_base_p[0x4], n, 4);
	if (state->flags.hires)
	  offset_page_bases(&state->read_base_p[0x20], n, 0x20);
#else
	set_page_bases(&state->read_base_p[0x4],
		       state->read_base_p[0x4] + n, 4);
	if (state->flags.hires)
	  set_page_bases(&state->read_base_p[0x20],
			 state->read_base_p[0x20] + n, 0x20);
#endif
      }
      if (state->flags.page2 != state->flags.aux_write)
      {
#if 0
	offset_page_bases(&state->write_base_p[0x4], n, 4);
	if (state->flags.hires)
	  offset_page_bases(&state->write_base_p[0x20], n, 0x20);
#else
	set_page_bases(&state->write_base_p[0x4],
		       state->write_base_p[0x4] + n, 4);
	if (state->flags.hires)
	  set_page_bases(&state->write_base_p[0x20],
			 state->write_base_p[0x20] + n, 0x20);
#endif
      }

      state->flags.store80 = !state->flags.store80;
    }
    break;

  case 0xC002:
  case 0xC003:
  case 0xC004:
  case 0xC005:
    if ((addr & 1) != ((addr & 2) ? state->flags.aux_read : state->flags.aux_write))
    {
      n = (addr & 1) ? AUXOFFSET : -AUXOFFSET;
      p = (addr & 2) ? state->read_base_p : state->write_base_p;

      if (!state->flags.store80)
      {
#if 0
	offset_page_bases(&p[0x2], n, 0xBE);
#else
	set_page_bases(&p[0x2], p[0x2] + n, 0xBE);
#endif
      }
      else
      {
#if 0
	offset_page_bases(&p[0x2], n, 0x2);
#else
	set_page_bases(&p[0x2], p[0x2] + n, 0x2);
#endif

	if (state->flags.hires)
	{
#if 0
	  offset_page_bases(&p[0x08], n, 0x18);
	  offset_page_bases(&p[0x40], n, 0x80);
#else
	  set_page_bases(&p[0x08], p[0x8] + n, 0x18);
	  set_page_bases(&p[0x40], p[0x40] + n, 0x80);
#endif
	}
	else
	{
#if 0
	  offset_page_bases(&p[0x08], n, 0xB8);
#else
	  set_page_bases(&p[0x8], p[0x8] + n, 0xB8);
#endif
	}
      }
      if (addr & 2)
	state->flags.aux_read = !state->flags.aux_read;
      else
	state->flags.aux_write = !state->flags.aux_write;
    }
    break;

  case 0xC006:			/* FIXME - incomplete */
    if (state->flags.intCXrom)
    {
      state->read_base_p[0xC1] = &state->sysmem[SLOT1ROM];
      state->write_base_p[0xC1] = &state->sysmem[SLOT1ROM];
      state->read_base_p[0xC2] = &state->sysmem[SLOT2ROM];
      state->write_base_p[0xC2] = &state->sysmem[SLOT2ROM];
    /* Leave slot 3 alone */
      state->read_base_p[0xC4] = &state->sysmem[SLOT4ROM];
      state->write_base_p[0xC4] = &state->sysmem[SLOT4ROM];
      state->read_base_p[0xC5] = &state->sysmem[SLOT5ROM];
      state->write_base_p[0xC5] = &state->sysmem[SLOT5ROM];
      state->read_base_p[0xC6] = &state->sysmem[SLOT6ROM];
      state->write_base_p[0xC6] = &state->sysmem[SLOT6ROM];
      state->read_base_p[0xC7] = &state->sysmem[SLOT7ROM];
      state->write_base_p[0xC7] = &state->sysmem[SLOT7ROM];
      state->flags.intCXrom = 0;
    }
    break;

  case 0xC007:
    if (!state->flags.intCXrom)
    {
      set_page_bases(&state->read_base_p[0xC1], &state->sysmem[INTROM], 15);
      set_page_bases(&state->write_base_p[0xC1], &state->sysmem[JUNK], 7);
      state->flags.intCXrom = 1;
    }
    break;

  case 0xC008:
  case 0xC009:
    if (state->flags.alt_zp != (addr & 1))
    {
      n = state->flags.alt_zp ? -AUXOFFSET : AUXOFFSET;

#if 0
      offset_page_bases(&state->read_base_p[0x00], n, 2);
      offset_page_bases(&state->write_base_p[0x00], n, 2);
#else
      set_page_bases(&state->read_base_p[0x00], state->read_base_p[0] + n, 2);
      set_page_bases(&state->write_base_p[0x00], state->write_base_p[0] + n, 2);
#endif

      if (state->flags.BSRreadram)
	offset_page_bases(&state->read_base_p[0xD0], n, 0x30);
      if (state->flags.BSRwriteram)
	offset_page_bases(&state->write_base_p[0xD0], n, 0x30);
      state->flags.alt_zp = !state->flags.alt_zp;
    }
    break;

  case 0xC00A:			/* SlotC3RomOff */
    state->flags.slotC3rom = 0;	/* FIXME */
    break;
  case 0xC00B:			/* SlotC3RomOn */
    state->flags.slotC3rom = 1;	/* FIXME */
    break;
  case 0xC00C:
    state->flags.col80 = 0;	/* FIXME? */
    break;
  case 0xC00D:
    state->flags.col80 = 1;	/* FIXME? */
    break;
  case 0xC00E:
    state->flags.altcharset = 0;/* FIXME? */
    break;
  case 0xC00F:
    state->flags.altcharset = 1;/* FIXME? */
    break;
  }

  return 0;
}


int
softswitch_C010(AccessType rw, MemoryElement * m, MachineState * state)
{
  int                 key;

  mutex_lock(state->keyboard.mutex);
  {
    key = state->keyboard.buffer[state->keyboard.head]
      | state->keyboard.key_down << 7;
    if (state->keyboard.tail != state->keyboard.head)
    {
      state->keyboard.head++;
      state->keyboard.head %= KEY_BUFFER_SIZE;
      state->keyboard.key_ready =
	(state->keyboard.tail != state->keyboard.head);
    }
  }
  mutex_unlock(state->keyboard.mutex);

  return key;
}





int
softswitch_C019(AccessType rw, MemoryElement * m, MachineState * state)
{
  return ((state->cycles * 60 / 1023000) & 1) << 7;
}



int
softswitch_C011_C018(AccessType rw, MemoryElement * m, MachineState * state)
{
  switch (m - state->sysmem)
  {
    case 0xC011:return (state->flags.BSRbank2 << 7);
  case 0xC012:
    return (state->flags.BSRreadram << 7);
  case 0xC013:
    return (state->flags.aux_read << 7);
  case 0xC014:
    return (state->flags.aux_write << 7);
  case 0xC015:
    return (state->flags.intCXrom << 7);
  case 0xC016:
    return (state->flags.alt_zp << 7);
  case 0xC017:
    return (state->flags.slotC3rom << 7);
  case 0xC018:
    return (state->flags.store80 << 7);
  }

 /*
  * If we get this far, this handler was called when it shouldn't
  * have been 
  */
  abort();
}


int
softswitch_C080_C08F(AccessType rw, MemoryElement * m, MachineState * state)
{
  static const struct
  {
    int                 D000, E000;
  }                   read_index[] =
  {
    {
                          LCBANK2, LCRAM
    }                  ,
    {
                          ROM, ROM2
    }                  ,
    {
                          ROM, ROM2
    }                  ,
    {
                          LCBANK2, LCRAM
    }                  ,
    {
                          AUXLCBANK2, AUXLCRAM
    }                  ,
    {
                          ROM, ROM2
    }                  ,
    {
                          ROM, ROM2
    }                  ,
    {
                          AUXLCBANK2, AUXLCRAM
    }                  ,
    {
                          LCBANK1, LCRAM
    }                  ,
    {
                          ROM, ROM2
    }                  ,
    {
                          ROM, ROM2
    }                  ,
    {
                          LCBANK1, LCRAM
    }                  ,
    {
                          AUXLCBANK1, AUXLCRAM
    }                  ,
    {
                          ROM, ROM2
    }                  ,
    {
                          ROM, ROM2
    }                  ,
    {
                          AUXLCBANK1, AUXLCRAM
    }
  };
  static const struct
  {
    int                 D000, E000;
  }                   write_index[] =
  {
    {
                          JUNK, JUNK
    }                  ,
    {
                          LCBANK2, LCRAM
    }                  ,
    {
                          JUNK, JUNK
    }                  ,
    {
                          LCBANK2, LCRAM
    }                  ,
    {
                          JUNK, JUNK
    }                  ,
    {
                          AUXLCBANK2, AUXLCRAM
    }                  ,
    {
                          JUNK, JUNK
    }                  ,
    {
                          AUXLCBANK2, AUXLCRAM
    }                  ,
    {
                          JUNK, JUNK
    }                  ,
    {
                          LCBANK1, LCRAM
    }                  ,
    {
                          JUNK, JUNK
    }                  ,
    {
                          LCBANK1, LCRAM
    }                  ,
    {
                          JUNK, JUNK
    }                  ,
    {
                          AUXLCBANK1, AUXLCRAM
    }                  ,
    {
                          JUNK, JUNK
    }                  ,
    {
                          AUXLCBANK1, AUXLCRAM
    }
  };
  MemoryElement     **rmem, **wmem, *read1, *read2, *write1, *write2;
  int                 n = ((m - state->sysmem) & 0xB) | (state->flags.alt_zp << 2);
  int                 writeplus;
  short               i;

  read1 = &state->sysmem[read_index[n].D000];
  read2 = &state->sysmem[read_index[n].E000];
  write1 = &state->sysmem[write_index[n].D000];
  write2 = &state->sysmem[write_index[n].E000];
  rmem = &state->read_base_p[0xD0];
  wmem = &state->write_base_p[0xD0];
  writeplus = (n & 1) ? 256 : 0;/* If "writing" to ROM's, don't inc ptr */

  for (i = 15; i >= 0; i--)
  {
    *rmem++ = read1;
    *wmem++ = write1;
    read1 += 256;
    write1 += writeplus;
  }

  for (i = 31; i >= 0; i--)
  {
    *rmem++ = read2;
    *wmem++ = write2;
    read2 += 256;
    write2 += writeplus;
  }

 /* Set up status bits */
  state->flags.BSRbank2 = !(n & 8);
  state->flags.BSRreadram = (((n & 2) >> 1) == (n & 1));
  state->flags.BSRwriteram = (n & 1);

  return 0;
}


/* Set display flags */
int
softswitch_C050_C057(AccessType rw, MemoryElement * m, MachineState * state)
{
  int                 addr = (m - state->sysmem);
  int                 onoff = (addr & 1);
  MemoryElement      *base;

  switch (addr)
  {
  case 0xC050:
  case 0xC051:
    state->flags.text = onoff;
    break;

  case 0xC052:
  case 0xC053:
    state->flags.mixed = onoff;
    break;

  case 0xC054:			/* PAGE2OFF */
  case 0xC055:			/* PAGE2ON  */
    if (state->flags.page2 != onoff)
    {
      if (state->flags.store80)
      {
	base = &state->sysmem[state->flags.page2 ? TEXT1 : AUXTEXT1];

	set_page_bases(&state->read_base_p[0x04], base, 0x4);
	set_page_bases(&state->write_base_p[0x04], base, 0x4);
	if (state->flags.hires)
	{
	  base = &state->sysmem[state->flags.page2 ? HIRES1 : AUXHIRES1];
	  set_page_bases(&state->read_base_p[0x20], base, 0x20);
	  set_page_bases(&state->write_base_p[0x20], base, 0x20);
	}
      }

      state->flags.page2 = !state->flags.page2;
    }
    break;

  case 0xC056:			/* HIRESOFF */
  case 0xC057:			/* HIRESON  */
    if (state->flags.hires != onoff)
    {
      if (state->flags.store80)
      {
	base = ((state->flags.page2 == state->flags.hires)
		? &state->sysmem[HIRES1] : &state->sysmem[AUXHIRES1]);
	if (state->flags.page2 != state->flags.aux_read)
	{
	  set_page_bases(&state->read_base_p[0x20], base, 0x20);
	}
	if (state->flags.page2 != state->flags.aux_write)
	{
	  set_page_bases(&state->write_base_p[0x20], base, 0x20);
	}
      }
      state->flags.hires = !state->flags.hires;
    }
    break;
  }

  return 0;
}


/* Annunciators */
int
softswitch_C058_C05F(AccessType rw, MemoryElement * m, MachineState * state)
{
  int                 addr = (m - state->sysmem);
  int                 onoff = addr & 1;

  switch (addr)
  {
  case 0xC058:
  case 0xC059:
    state->flags.an0 = onoff;
    break;

  case 0xC05A:
  case 0xC05B:
    state->flags.an1 = onoff;
    break;

  case 0xC05C:
  case 0xC05D:
    state->flags.an2 = onoff;
    break;

  case 0xC05E:
  case 0xC05F:
    state->flags.an3 = onoff;
    break;

  default:
    abort();
  }

  return 0;
}


/* Read display flags */
int
softswitch_C01A_C01F(AccessType rw, MemoryElement * m, MachineState * state)
{
  switch (m - state->sysmem)
  {
    case 0xC01A:return (state->flags.text << 7);
  case 0xC01B:
    return (state->flags.mixed << 7);
  case 0xC01C:
    return (state->flags.page2 << 7);
  case 0xC01D:
    return (state->flags.hires << 7);
  case 0xC01E:
    return (state->flags.altcharset << 7);
  case 0xC01F:
    return (state->flags.col80 << 7);
  }

 /*
  * If we get this far, this handler was called when it shouldn't
  * have been 
  */
  abort();
}


/* Paddle timers. */
int
softswitch_C064_C067(AccessType rw, MemoryElement * m, MachineState * state)
{
  int                 n = 0x80;

  switch (m - state->sysmem)
  {
  case 0xC064:
    if (!state->flags.gc0 || state->cycles - state->pdl_timeout[0] < -50000UL)
    {
      state->flags.gc0 = 0;
      n = 0;
    }
    break;
  case 0xC065:
    if (!state->flags.gc1 || state->cycles - state->pdl_timeout[1] < -50000UL)
    {
      state->flags.gc1 = 0;
      n = 0;
    }
    break;
  case 0xC066:
  case 0xC067:
    break;
  }

  return n;
}


/* Reset paddle timer. */
int
softswitch_C070(AccessType rw, MemoryElement * m, MachineState * state)
{
 /* Only start a new timer if it's not running */
  if (!state->flags.gc0 || state->cycles - state->pdl_timeout[0] < -50000UL)
  {
    state->pdl_timeout[0] = state->cycles + state->pdl[0] * 11;
    state->flags.gc0 = 1;
  }

 /* Only start a new timer if it's not running */
  if (!state->flags.gc1 || state->cycles - state->pdl_timeout[1] < -50000UL)
  {
    state->pdl_timeout[1] = state->cycles + state->pdl[1] * 11;
    state->flags.gc1 = 1;
  }

  return 0;
}


int
softswitch_C061_C063(AccessType rw, MemoryElement * m, MachineState * state)
{
  switch (m - state->sysmem)
  {
    case 0xC061:
    return (state->flags.pb0 << 7);
  case 0xC062:
    return (state->flags.pb1 << 7);
  }

  return 0;
}
