/*
 * This file is part of zaniWok.
 *
 * zaniWok is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.  See the file COPYING for more information.
 *
 * zaniWok is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */


#include "translate.h"
#include "opcodes.h"
#ifndef FALSE
#define FALSE 0
#endif
#ifndef TRUE
#define TRUE 1
#endif

const TranslateInfo translate_info[256] = {
 /* 0x00 */
					   {OPC_BRK, 0, FALSE, FALSE, FALSE, FALSE, AM_IMPLICIT, "BRK"},
					   {OPC_OR_IND_X, 1, FALSE, FALSE, FALSE, TRUE, AM_IND_X, "ORA"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_TSB_ZP, 1, FALSE, FALSE, FALSE, TRUE, AM_ZP, "TSB"},
					   {OPC_OR_ZP, 1, FALSE, FALSE, FALSE, TRUE, AM_ZP, "ORA"},
					   {OPC_ASL_ZP, 1, FALSE, FALSE, FALSE, TRUE, AM_ZP, "ASL"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_PHP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "PHP"},
					   {OPC_OR_IMM, 1, FALSE, FALSE, FALSE, TRUE, AM_IMM, "ORA"},
					   {OPC_ASL_ACC, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "ASL"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_TSB_ABS, 2, FALSE, FALSE, FALSE, TRUE, AM_ABS, "TSB"},
					   {OPC_OR_ABS, 2, FALSE, FALSE, FALSE, TRUE, AM_ABS, "ORA"},
					   {OPC_ASL_ABS, 2, FALSE, FALSE, FALSE, TRUE, AM_ABS, "ASL"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},

 /* 0x10 */
					   {OPC_BPL, 1, TRUE, FALSE, FALSE, FALSE, AM_RELATIVE, "BPL"},
					   {OPC_OR_IND_Y, 1, FALSE, FALSE, FALSE, TRUE, AM_IND_Y, "ORA"},
					   {OPC_OR_IND, 1, FALSE, FALSE, FALSE, TRUE, AM_IND, "ORA"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_TRB_ZP, 1, FALSE, FALSE, FALSE, TRUE, AM_ZP, "TRB"},
					   {OPC_OR_ZP_X, 1, FALSE, FALSE, FALSE, TRUE, AM_ZP_X, "ORA"},
					   {OPC_ASL_ZP_X, 1, FALSE, FALSE, FALSE, TRUE, AM_ZP_X, "ASL"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_CLC, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "CLC"},
					   {OPC_OR_ABS_Y, 2, FALSE, FALSE, FALSE, TRUE, AM_ABS_Y, "ORA"},
					   {OPC_INC_ACC, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "INC"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_TRB_ABS, 0, FALSE, FALSE, FALSE, TRUE, AM_ABS, "TRB"},
					   {OPC_OR_ABS_X, 2, FALSE, FALSE, FALSE, TRUE, AM_ABS_X, "ORA"},
					   {OPC_ASL_ABS_X, 2, FALSE, FALSE, FALSE, TRUE, AM_ABS_X, "ASL"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},

 /* 0x20 */
					   {OPC_JSR, 2, FALSE, FALSE, FALSE, FALSE, AM_ABS, "JSR"},
					   {OPC_AND_IND_X, 1, FALSE, FALSE, FALSE, TRUE, AM_IND_X, "AND"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_BIT_ZP, 1, FALSE, FALSE, FALSE, TRUE, AM_ZP, "BIT"},
					   {OPC_AND_ZP, 1, FALSE, FALSE, FALSE, TRUE, AM_ZP, "AND"},
					   {OPC_ROL_ZP, 1, FALSE, FALSE, FALSE, TRUE, AM_ZP, "ROL"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_PLP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "PLP"},
					   {OPC_AND_IMM, 1, FALSE, FALSE, FALSE, TRUE, AM_IMM, "AND"},
					   {OPC_ROL_ACC, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "ROL"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_BIT_ABS, 2, FALSE, FALSE, FALSE, TRUE, AM_ABS, "BIT"},
					   {OPC_AND_ABS, 2, FALSE, FALSE, FALSE, TRUE, AM_ABS, "AND"},
					   {OPC_ROL_ABS, 2, FALSE, FALSE, FALSE, TRUE, AM_ABS, "ROL"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},

 /* 0x30 */
					   {OPC_BMI, 1, TRUE, FALSE, FALSE, FALSE, AM_RELATIVE, "BMI"},
					   {OPC_AND_IND_Y, 1, FALSE, FALSE, FALSE, TRUE, AM_IND_Y, "AND"},
					   {OPC_AND_IND, 1, FALSE, FALSE, FALSE, TRUE, AM_IND, "AND"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_BIT_ZP_X, 1, FALSE, FALSE, FALSE, TRUE, AM_ZP_X, "BIT"},
					   {OPC_AND_ZP_X, 1, FALSE, FALSE, FALSE, TRUE, AM_ZP_X, "AND"},
					   {OPC_ROL_ZP_X, 1, FALSE, FALSE, FALSE, TRUE, AM_ZP_X, "ROL"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_SEC, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "SEC"},
					   {OPC_AND_ABS_Y, 2, FALSE, FALSE, FALSE, TRUE, AM_ABS_Y, "AND"},
					   {OPC_DEC_ACC, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "DEC"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_BIT_ABS_X, 2, FALSE, FALSE, FALSE, TRUE, AM_ABS_X, "BIT"},
					   {OPC_AND_ABS_X, 2, FALSE, FALSE, FALSE, TRUE, AM_ABS_X, "AND"},
					   {OPC_ROL_ABS_X, 2, FALSE, FALSE, FALSE, TRUE, AM_ABS_X, "ROL"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},

 /* 0x40 */
					   {OPC_RTI, 0, FALSE, FALSE, FALSE, FALSE, AM_IMPLICIT, "RTI"},
					   {OPC_EOR_IND_X, 1, FALSE, FALSE, FALSE, TRUE, AM_IND_X, "EOR"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_EOR_ZP, 1, FALSE, FALSE, FALSE, TRUE, AM_ZP, "EOR"},
					   {OPC_LSR_ZP, 1, FALSE, FALSE, FALSE, TRUE, AM_ZP, "LSR"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_PHA, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "PHA"},
					   {OPC_EOR_IMM, 1, FALSE, FALSE, FALSE, TRUE, AM_IMM, "EOR"},
					   {OPC_LSR_ACC, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "LSR"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_JMP_ABS, 2, FALSE, FALSE, FALSE, FALSE, AM_ABS, "JMP"},
					   {OPC_EOR_ABS, 2, FALSE, FALSE, FALSE, TRUE, AM_ABS, "EOR"},
					   {OPC_LSR_ABS, 2, FALSE, FALSE, FALSE, TRUE, AM_ABS, "LSR"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},

 /* 0x50 */
					   {OPC_BVC, 1, TRUE, FALSE, FALSE, FALSE, AM_RELATIVE, "BVC"},
					   {OPC_EOR_IND_Y, 1, FALSE, FALSE, FALSE, TRUE, AM_IND_Y, "EOR"},
					   {OPC_EOR_IND, 1, FALSE, FALSE, FALSE, TRUE, AM_IND, "EOR"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_EOR_ZP_X, 1, FALSE, FALSE, FALSE, TRUE, AM_ZP_X, "EOR"},
					   {OPC_LSR_ZP_X, 1, FALSE, FALSE, FALSE, TRUE, AM_ZP_X, "LSR"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_CLI, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "CLI"},
					   {OPC_EOR_ABS_Y, 2, FALSE, FALSE, FALSE, TRUE, AM_ABS_Y, "EOR"},
					   {OPC_PHY, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "PHY"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_EOR_ABS_X, 2, FALSE, FALSE, FALSE, TRUE, AM_ABS_X, "EOR"},
					   {OPC_LSR_ABS_X, 2, FALSE, FALSE, FALSE, TRUE, AM_ABS_X, "LSR"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},

 /* 0x60 */
					   {OPC_RTS, 0, FALSE, FALSE, FALSE, FALSE, AM_IMPLICIT, "RTS"},
					   {OPC_ADC_IND_X, 1, FALSE, FALSE, FALSE, TRUE, AM_IND_X, "ADC"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_STZ_ZP, 1, FALSE, FALSE, FALSE, TRUE, AM_ZP, "STZ"},
					   {OPC_ADC_ZP, 1, FALSE, FALSE, FALSE, TRUE, AM_ZP, "ADC"},
					   {OPC_ROR_ZP, 1, FALSE, FALSE, FALSE, TRUE, AM_ZP, "ROR"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_PLA, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "PLA"},
					   {OPC_ADC_IMM, 1, FALSE, FALSE, FALSE, TRUE, AM_IMM, "ADC"},
					   {OPC_ROR_ACC, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "ROR"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_JMP_ABSIND, 2, FALSE, FALSE, FALSE, FALSE, AM_ABSIND, "JMP"},
					   {OPC_ADC_ABS, 2, FALSE, TRUE, FALSE, TRUE, AM_ABS, "ADC"},
					   {OPC_ROR_ABS, 2, FALSE, FALSE, FALSE, TRUE, AM_ABS, "ROR"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},

 /* 0x70 */
					   {OPC_BVS, 1, TRUE, FALSE, FALSE, FALSE, AM_RELATIVE, "BVS"},
					   {OPC_ADC_IND_Y, 1, FALSE, FALSE, FALSE, TRUE, AM_IND_Y, "ADC"},
					   {OPC_ADC_IND, 1, FALSE, FALSE, FALSE, TRUE, AM_IND, "ADC"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_STZ_ZP_X, 1, FALSE, FALSE, FALSE, TRUE, AM_ZP_X, "STZ"},
					   {OPC_ADC_ZP_X, 1, FALSE, FALSE, FALSE, TRUE, AM_ZP_X, "ADC"},
					   {OPC_ROR_ZP_X, 1, FALSE, FALSE, FALSE, TRUE, AM_ZP_X, "ROR"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_SEI, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "SEI"},
					   {OPC_ADC_ABS_Y, 2, FALSE, FALSE, TRUE, TRUE, AM_ABS_Y, "ADC"},
					   {OPC_PLY, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "PLY"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_JMP_ABSIND_X, 2, FALSE, FALSE, FALSE, FALSE, AM_ABSIND_X, "JMP"},
					   {OPC_ADC_ABS_X, 2, FALSE, FALSE, TRUE, TRUE, AM_ABS_X, "ADC"},
					   {OPC_ROR_ABS_X, 2, FALSE, FALSE, FALSE, TRUE, AM_ABS_X, "ROR"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},

 /* 0x80 */
					   {OPC_BRA, 1, TRUE, FALSE, FALSE, FALSE, AM_RELATIVE, "BRA"},
					   {OPC_STA_IND_X, 1, FALSE, FALSE, FALSE, TRUE, AM_IND_X, "STA"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_STY_ZP, 1, FALSE, FALSE, FALSE, TRUE, AM_ZP, "STY"},
					   {OPC_STA_ZP, 1, FALSE, FALSE, FALSE, TRUE, AM_ZP, "STA"},
					   {OPC_STX_ZP, 1, FALSE, FALSE, FALSE, TRUE, AM_ZP, "STX"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_DEY, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "DEY"},
					   {OPC_BIT_IMM, 1, FALSE, FALSE, FALSE, TRUE, AM_IMM, "BIT"},
					   {OPC_TXA, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "TXA"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_STY_ABS, 2, FALSE, FALSE, FALSE, TRUE, AM_ABS, "STY"},
					   {OPC_STA_ABS, 2, FALSE, FALSE, FALSE, TRUE, AM_ABS, "STA"},
					   {OPC_STX_ABS, 2, FALSE, FALSE, FALSE, TRUE, AM_ABS, "STX"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},

 /* 0x90 */
					   {OPC_BCC, 1, TRUE, FALSE, FALSE, FALSE, AM_RELATIVE, "BCC"},
					   {OPC_STA_IND_Y, 1, FALSE, FALSE, FALSE, TRUE, AM_IND_Y, "STA"},
					   {OPC_STA_IND, 1, FALSE, FALSE, FALSE, TRUE, AM_IND, "STA"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_STY_ZP_X, 1, FALSE, FALSE, FALSE, TRUE, AM_ZP_X, "STY"},
					   {OPC_STA_ZP_X, 1, FALSE, FALSE, FALSE, TRUE, AM_ZP_X, "STA"},
					   {OPC_STX_ZP_Y, 1, FALSE, FALSE, FALSE, TRUE, AM_ZP_Y, "STX"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_TYA, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "TYA"},
					   {OPC_STA_ABS_Y, 2, FALSE, FALSE, FALSE, TRUE, AM_ABS_Y, "STA"},
					   {OPC_TXS, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "TXS"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_STZ_ABS, 2, FALSE, FALSE, FALSE, TRUE, AM_ABS, "STZ"},
					   {OPC_STA_ABS_X, 2, FALSE, FALSE, FALSE, TRUE, AM_ABS_X, "STA"},
					   {OPC_STZ_ABS_X, 2, FALSE, FALSE, FALSE, TRUE, AM_ABS_X, "STZ"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},

 /* 0xA0 */
					   {OPC_LDY_IMM, 1, FALSE, FALSE, FALSE, TRUE, AM_IMM, "LDY"},
					   {OPC_LDA_IND_X, 1, FALSE, FALSE, FALSE, TRUE, AM_IND_X, "LDA"},
					   {OPC_LDX_IMM, 1, FALSE, FALSE, FALSE, TRUE, AM_IMM, "LDX"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_LDY_ZP, 1, FALSE, FALSE, FALSE, TRUE, AM_ZP, "LDY"},
					   {OPC_LDA_ZP, 1, FALSE, FALSE, FALSE, TRUE, AM_ZP, "LDA"},
					   {OPC_LDX_ZP, 1, FALSE, FALSE, FALSE, TRUE, AM_ZP, "LDX"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_TAY, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "TAY"},
					   {OPC_LDA_IMM, 1, FALSE, FALSE, FALSE, TRUE, AM_IMM, "LDA"},
					   {OPC_TAX, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "TAX"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_LDY_ABS, 2, FALSE, TRUE, FALSE, TRUE, AM_ABS, "LDY"},
					   {OPC_LDA_ABS, 2, FALSE, TRUE, FALSE, TRUE, AM_ABS, "LDA"},
					   {OPC_LDX_ABS, 2, FALSE, TRUE, FALSE, TRUE, AM_ABS, "LDX"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},

 /* 0xB0 */
					   {OPC_BCS, 1, TRUE, FALSE, FALSE, FALSE, AM_RELATIVE, "BCS"},
					   {OPC_LDA_IND_Y, 1, FALSE, FALSE, FALSE, TRUE, AM_IND_Y, "LDA"},
					   {OPC_LDA_IND, 1, FALSE, FALSE, FALSE, TRUE, AM_IND, "LDA"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_LDY_ZP_X, 1, FALSE, FALSE, FALSE, TRUE, AM_ZP_X, "LDY"},
					   {OPC_LDA_ZP_X, 1, FALSE, FALSE, FALSE, TRUE, AM_ZP_X, "LDA"},
					   {OPC_LDX_ZP_Y, 1, FALSE, FALSE, FALSE, TRUE, AM_ZP_Y, "LDX"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_CLV, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "CLV"},
					   {OPC_LDA_ABS_Y, 2, FALSE, FALSE, TRUE, TRUE, AM_ABS_Y, "LDA"},
					   {OPC_TSX, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "TSX"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_LDY_ABS_X, 2, FALSE, FALSE, FALSE, TRUE, AM_ABS_X, "LDY"},
					   {OPC_LDA_ABS_X, 2, FALSE, FALSE, TRUE, TRUE, AM_ABS_X, "LDA"},
					   {OPC_LDX_ABS_Y, 2, FALSE, FALSE, TRUE, TRUE, AM_ABS_Y, "LDX"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},

 /* 0xC0 */
					   {OPC_CPY_IMM, 1, FALSE, FALSE, FALSE, TRUE, AM_IMM, "CPY"},
					   {OPC_CMP_IND_X, 1, FALSE, FALSE, FALSE, TRUE, AM_IND_X, "CMP"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_CPY_ZP, 1, FALSE, FALSE, FALSE, TRUE, AM_ZP, "CPY"},
					   {OPC_CMP_ZP, 1, FALSE, FALSE, FALSE, TRUE, AM_ZP, "CMP"},
					   {OPC_DEC_ZP, 1, FALSE, FALSE, FALSE, TRUE, AM_ZP, "DEC"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_INY, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "INY"},
					   {OPC_CMP_IMM, 1, FALSE, FALSE, FALSE, TRUE, AM_IMM, "CMP"},
					   {OPC_DEX, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "DEX"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_CPY_ABS, 2, FALSE, TRUE, FALSE, TRUE, AM_ABS, "CPY"},
					   {OPC_CMP_ABS, 2, FALSE, TRUE, FALSE, TRUE, AM_ABS, "CMP"},
					   {OPC_DEC_ABS, 2, FALSE, FALSE, FALSE, TRUE, AM_ABS, "DEC"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},

 /* 0xD0 */
					   {OPC_BNE, 1, TRUE, FALSE, FALSE, FALSE, AM_RELATIVE, "BNE"},
					   {OPC_CMP_IND_Y, 1, FALSE, FALSE, FALSE, TRUE, AM_IND_Y, "CMP"},
					   {OPC_CMP_IND, 1, FALSE, FALSE, FALSE, TRUE, AM_IND, "CMP"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_CMP_ZP_X, 1, FALSE, FALSE, FALSE, TRUE, AM_ZP_X, "CMP"},
					   {OPC_DEC_ZP_X, 1, FALSE, FALSE, FALSE, TRUE, AM_ZP_X, "DEC"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_CLD, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "CLD"},
					   {OPC_CMP_ABS_Y, 2, FALSE, FALSE, TRUE, TRUE, AM_ABS_Y, "CMP"},
					   {OPC_PHX, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "PHX"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_CMP_ABS_X, 2, FALSE, FALSE, TRUE, TRUE, AM_ABS_X, "CMP"},
					   {OPC_DEC_ABS_X, 2, FALSE, FALSE, FALSE, TRUE, AM_ABS_X, "DEC"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},

 /* 0xE0 */
					   {OPC_CPX_IMM, 1, FALSE, FALSE, FALSE, TRUE, AM_IMM, "CPX"},
					   {OPC_SBC_IND_X, 1, FALSE, FALSE, FALSE, TRUE, AM_IND_X, "SBC"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_CPX_ZP, 1, FALSE, FALSE, FALSE, TRUE, AM_ZP, "CPX"},
					   {OPC_SBC_ZP, 1, FALSE, FALSE, FALSE, TRUE, AM_ZP, "SBC"},
					   {OPC_INC_ZP, 1, FALSE, FALSE, FALSE, TRUE, AM_ZP, "INC"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_INX, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "INX"},
					   {OPC_SBC_IMM, 1, FALSE, FALSE, FALSE, TRUE, AM_IMM, "SBC"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_CPX_ABS, 2, FALSE, TRUE, FALSE, TRUE, AM_ABS, "CPX"},
					   {OPC_SBC_ABS, 2, FALSE, TRUE, FALSE, TRUE, AM_ABS, "SBC"},
					   {OPC_INC_ABS, 2, FALSE, FALSE, FALSE, TRUE, AM_ABS, "INC"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},

 /* 0xF0 */
					   {OPC_BEQ, 1, TRUE, FALSE, FALSE, FALSE, AM_RELATIVE, "BEQ"},
					   {OPC_SBC_IND_Y, 1, FALSE, FALSE, FALSE, TRUE, AM_IND_Y, "SBC"},
					   {OPC_SBC_IND, 1, FALSE, FALSE, FALSE, TRUE, AM_IND, "SBC"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_SBC_ZP_X, 1, FALSE, FALSE, FALSE, TRUE, AM_ZP_X, "SBC"},
					   {OPC_INC_ZP_X, 1, FALSE, FALSE, FALSE, TRUE, AM_ZP_X, "INC"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_SED, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "SED"},
					   {OPC_SBC_ABS_Y, 2, FALSE, FALSE, TRUE, TRUE, AM_ABS_Y, "SBC"},
					   {OPC_PLX, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "PLX"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
					   {OPC_SBC_ABS_X, 2, FALSE, FALSE, TRUE, TRUE, AM_ABS_X, "SBC"},
					   {OPC_INC_ABS_X, 2, FALSE, FALSE, FALSE, TRUE, AM_ABS_X, "INC"},
					   {OPC_NOP, 0, FALSE, FALSE, FALSE, TRUE, AM_IMPLICIT, "NOP"},
};

const CompileInfo   compile_info[] = {
				      [OPC_NOT_COMPILED / BLOCKSIZE] {NONE},
				      [OPC_INVALID / BLOCKSIZE] {NONE},
				      [OPC_ADC_IMM / BLOCKSIZE] {ALLBITS},
				      [OPC_ADC_ZP / BLOCKSIZE] {ALLBITS},
				      [OPC_ADC_ZP_X / BLOCKSIZE] {ALLBITS},
				      [OPC_ADC_ABS / BLOCKSIZE] {ALLBITS},
				      [OPC_ADC_ABS_X / BLOCKSIZE] {ALLBITS},
				      [OPC_ADC_ABS_Y / BLOCKSIZE] {ALLBITS},
				      [OPC_ADC_IND / BLOCKSIZE] {ALLBITS},
				      [OPC_ADC_IND_X / BLOCKSIZE] {ALLBITS},
				      [OPC_ADC_IND_Y / BLOCKSIZE] {ALLBITS},
				      [OPC_AND_IMM / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_AND_ZP / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_AND_ZP_X / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_AND_ABS / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_AND_ABS_X / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_AND_ABS_Y / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_AND_IND / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_AND_IND_X / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_AND_IND_Y / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_ASL_ACC / BLOCKSIZE] {CBIT | NBIT | ZBIT},
				      [OPC_ASL_ZP / BLOCKSIZE] {CBIT | NBIT | ZBIT},
				      [OPC_ASL_ZP_X / BLOCKSIZE] {CBIT | NBIT | ZBIT},
				      [OPC_ASL_ABS / BLOCKSIZE] {CBIT | NBIT | ZBIT},
				      [OPC_ASL_ABS_X / BLOCKSIZE] {CBIT | NBIT | ZBIT},
				      [OPC_BCC / BLOCKSIZE] {NONE},
				      [OPC_BCS / BLOCKSIZE] {NONE},
				      [OPC_BEQ / BLOCKSIZE] {NONE},
				      [OPC_BMI / BLOCKSIZE] {NONE},
				      [OPC_BNE / BLOCKSIZE] {NONE},
				      [OPC_BPL / BLOCKSIZE] {NONE},
				      [OPC_BRA / BLOCKSIZE] {NONE},
				      [OPC_BVC / BLOCKSIZE] {NONE},
				      [OPC_BVS / BLOCKSIZE] {NONE},
				      [OPC_BIT_IMM / BLOCKSIZE] {NBIT | VBIT | ZBIT},
				      [OPC_BIT_ZP / BLOCKSIZE] {NBIT | VBIT | ZBIT},
				      [OPC_BIT_ABS / BLOCKSIZE] {NBIT | VBIT | ZBIT},
				      [OPC_BIT_ZP_X / BLOCKSIZE] {NBIT | VBIT | ZBIT},
				      [OPC_BIT_ABS_X / BLOCKSIZE] {NBIT | VBIT | ZBIT},
				      [OPC_BRK / BLOCKSIZE] {NONE},
				      [OPC_CLC / BLOCKSIZE] {CBIT},
				      [OPC_CLD / BLOCKSIZE] {NONE},
				      [OPC_CLI / BLOCKSIZE] {NONE},
				      [OPC_CLV / BLOCKSIZE] {VBIT},
				      [OPC_CMP_IMM / BLOCKSIZE] {CBIT | NBIT | ZBIT},
				      [OPC_CMP_ZP / BLOCKSIZE] {CBIT | NBIT | ZBIT},
				      [OPC_CMP_ZP_X / BLOCKSIZE] {CBIT | NBIT | ZBIT},
				      [OPC_CMP_ABS / BLOCKSIZE] {CBIT | NBIT | ZBIT},
				      [OPC_CMP_ABS_X / BLOCKSIZE] {CBIT | NBIT | ZBIT},
				      [OPC_CMP_ABS_Y / BLOCKSIZE] {CBIT | NBIT | ZBIT},
				      [OPC_CMP_IND / BLOCKSIZE] {CBIT | NBIT | ZBIT},
				      [OPC_CMP_IND_X / BLOCKSIZE] {CBIT | NBIT | ZBIT},
				      [OPC_CMP_IND_Y / BLOCKSIZE] {CBIT | NBIT | ZBIT},
				      [OPC_CPX_IMM / BLOCKSIZE] {CBIT | NBIT | ZBIT},
				      [OPC_CPX_ZP / BLOCKSIZE] {CBIT | NBIT | ZBIT},
				      [OPC_CPX_ABS / BLOCKSIZE] {CBIT | NBIT | ZBIT},
				      [OPC_CPY_IMM / BLOCKSIZE] {CBIT | NBIT | ZBIT},
				      [OPC_CPY_ZP / BLOCKSIZE] {CBIT | NBIT | ZBIT},
				      [OPC_CPY_ABS / BLOCKSIZE] {CBIT | NBIT | ZBIT},
				      [OPC_DEC_ACC / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_DEC_ZP / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_DEC_ZP_X / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_DEC_ABS / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_DEC_ABS_X / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_DEX / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_DEY / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_EOR_IMM / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_EOR_ZP / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_EOR_ZP_X / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_EOR_ABS / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_EOR_ABS_X / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_EOR_ABS_Y / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_EOR_IND / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_EOR_IND_X / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_EOR_IND_Y / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_INC_ZP / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_INC_ACC / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_INC_ZP_X / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_INC_ABS / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_INC_ABS_X / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_INX / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_INY / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_JMP_ABS / BLOCKSIZE] {NONE},
				      [OPC_JMP_ABSIND / BLOCKSIZE] {NONE},
				      [OPC_JMP_ABSIND_X / BLOCKSIZE] {NONE},
				      [OPC_JSR / BLOCKSIZE] {NONE},
				      [OPC_LDA_IMM / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_LDA_ZP / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_LDA_ZP_X / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_LDA_ABS / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_LDA_ABS_X / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_LDA_ABS_Y / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_LDA_IND / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_LDA_IND_X / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_LDA_IND_Y / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_LDX_IMM / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_LDX_ZP / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_LDX_ZP_Y / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_LDX_ABS / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_LDX_ABS_Y / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_LDY_IMM / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_LDY_ZP / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_LDY_ZP_X / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_LDY_ABS / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_LDY_ABS_X / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_LSR_ACC / BLOCKSIZE] {CBIT | NBIT | ZBIT},
				      [OPC_LSR_ZP / BLOCKSIZE] {CBIT | NBIT | ZBIT},
				      [OPC_LSR_ZP_X / BLOCKSIZE] {CBIT | NBIT | ZBIT},
				      [OPC_LSR_ABS / BLOCKSIZE] {CBIT | NBIT | ZBIT},
				      [OPC_LSR_ABS_X / BLOCKSIZE] {CBIT | NBIT | ZBIT},
				      [OPC_NOP / BLOCKSIZE] {NONE},
				      [OPC_OR_IMM / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_OR_ZP / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_OR_ZP_X / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_OR_ABS / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_OR_ABS_X / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_OR_ABS_Y / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_OR_IND / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_OR_IND_X / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_OR_IND_Y / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_PHA / BLOCKSIZE] {NONE},
				      [OPC_PHX / BLOCKSIZE] {NONE},
				      [OPC_PHY / BLOCKSIZE] {NONE},
				      [OPC_PLX / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_PLY / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_PHP / BLOCKSIZE] {NONE},
				      [OPC_PLA / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_PLP / BLOCKSIZE] {ALLBITS},
				      [OPC_ROL_ACC / BLOCKSIZE] {CBIT | NBIT | ZBIT},
				      [OPC_ROL_ZP / BLOCKSIZE] {CBIT | NBIT | ZBIT},
				      [OPC_ROL_ZP_X / BLOCKSIZE] {CBIT | NBIT | ZBIT},
				      [OPC_ROL_ABS / BLOCKSIZE] {CBIT | NBIT | ZBIT},
				      [OPC_ROL_ABS_X / BLOCKSIZE] {CBIT | NBIT | ZBIT},
				      [OPC_ROR_ACC / BLOCKSIZE] {CBIT | NBIT | ZBIT},
				      [OPC_ROR_ZP / BLOCKSIZE] {CBIT | NBIT | ZBIT},
				      [OPC_ROR_ZP_X / BLOCKSIZE] {CBIT | NBIT | ZBIT},
				      [OPC_ROR_ABS / BLOCKSIZE] {CBIT | NBIT | ZBIT},
				      [OPC_ROR_ABS_X / BLOCKSIZE] {CBIT | NBIT | ZBIT},
				      [OPC_RTI / BLOCKSIZE] {ALLBITS},
				      [OPC_RTS / BLOCKSIZE] {NONE},
				      [OPC_SBC_IMM / BLOCKSIZE] {ALLBITS},
				      [OPC_SBC_ZP / BLOCKSIZE] {ALLBITS},
				      [OPC_SBC_ZP_X / BLOCKSIZE] {ALLBITS},
				      [OPC_SBC_ABS / BLOCKSIZE] {ALLBITS},
				      [OPC_SBC_ABS_X / BLOCKSIZE] {ALLBITS},
				      [OPC_SBC_ABS_Y / BLOCKSIZE] {ALLBITS},
				      [OPC_SBC_IND / BLOCKSIZE] {ALLBITS},
				      [OPC_SBC_IND_X / BLOCKSIZE] {ALLBITS},
				      [OPC_SBC_IND_Y / BLOCKSIZE] {ALLBITS},
				      [OPC_SEC / BLOCKSIZE] {CBIT},
				      [OPC_SED / BLOCKSIZE] {NONE},
				      [OPC_SEI / BLOCKSIZE] {NONE},
				      [OPC_STA_ZP / BLOCKSIZE] {NONE},
				      [OPC_STA_ZP_X / BLOCKSIZE] {NONE},
				      [OPC_STA_ABS / BLOCKSIZE] {NONE},
				      [OPC_STA_ABS_X / BLOCKSIZE] {NONE},
				      [OPC_STA_ABS_Y / BLOCKSIZE] {NONE},
				      [OPC_STA_IND / BLOCKSIZE] {NONE},
				      [OPC_STA_IND_X / BLOCKSIZE] {NONE},
				      [OPC_STA_IND_Y / BLOCKSIZE] {NONE},
				      [OPC_STX_ZP / BLOCKSIZE] {NONE},
				      [OPC_STX_ZP_Y / BLOCKSIZE] {NONE},
				      [OPC_STX_ABS / BLOCKSIZE] {NONE},
				      [OPC_STY_ZP / BLOCKSIZE] {NONE},
				      [OPC_STY_ZP_X / BLOCKSIZE] {NONE},
				      [OPC_STY_ABS / BLOCKSIZE] {NONE},
				      [OPC_STZ_ZP / BLOCKSIZE] {NONE},
				      [OPC_STZ_ZP_X / BLOCKSIZE] {NONE},
				      [OPC_STZ_ABS / BLOCKSIZE] {NONE},
				      [OPC_STZ_ABS_X / BLOCKSIZE] {NONE},
				      [OPC_TAX / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_TAY / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_TRB_ZP / BLOCKSIZE] {ZBIT},
				      [OPC_TRB_ABS / BLOCKSIZE] {ZBIT},
				      [OPC_TSB_ZP / BLOCKSIZE] {ZBIT},
				      [OPC_TSB_ABS / BLOCKSIZE] {ZBIT},
				      [OPC_TSX / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_TXA / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_TXS / BLOCKSIZE] {NONE},
				      [OPC_TYA / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_TRAPTHIS / BLOCKSIZE] {NONE},
				      [OPC_CARDTRAP / BLOCKSIZE] {NONE},
				      [OPC_LDASAFE_ABS / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_LDASAFE_ABS_X / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_LDASAFE_ABS_Y / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_LDXSAFE_ABS / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_LDXSAFE_ABS_Y / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_LDYSAFE_ABS / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_LDYSAFE_ABS_X / BLOCKSIZE] {NBIT | ZBIT},
				      [OPC_ADCSAFE_ABS / BLOCKSIZE] {ALLBITS},
				      [OPC_ADCSAFE_ABS_X / BLOCKSIZE] {ALLBITS},
				      [OPC_ADCSAFE_ABS_Y / BLOCKSIZE] {ALLBITS},
				      [OPC_SBCSAFE_ABS / BLOCKSIZE] {ALLBITS},
				      [OPC_SBCSAFE_ABS_X / BLOCKSIZE] {ALLBITS},
				      [OPC_SBCSAFE_ABS_Y / BLOCKSIZE] {ALLBITS},
				      [OPC_CMPSAFE_ABS / BLOCKSIZE] {CBIT | NBIT | ZBIT},
				      [OPC_CMPSAFE_ABS_X / BLOCKSIZE] {CBIT | NBIT | ZBIT},
				      [OPC_CMPSAFE_ABS_Y / BLOCKSIZE] {CBIT | NBIT | ZBIT},
				      [OPC_CPXSAFE_ABS / BLOCKSIZE] {CBIT | NBIT | ZBIT},
				      [OPC_CPYSAFE_ABS / BLOCKSIZE] {CBIT | NBIT | ZBIT}};
