#include <stdio.h>

/* This program generates the color lookup tables found in hgrtable.c */


static void get_colors (int n, int c[4], int highbit);

int
main (int argc, char *argv[])
{
  int i;

  /* Output standard colors. */
  puts ("#define C0 0x000F000FL\n"
	"#define C1 0x0E0F0E0FL\n"
	"#define C2 0xE2EFE2EFL\n"
	"#define C3 0xFFFFFFFFL\n"
	"#define C4 0x000F000FL\n"
	"#define C5 0xF60FF60FL\n"
	"#define C6 0x22FF22FFL\n"
	"#define C7 0xFFFFFFFFL\n");
  
  puts ("/* Make sure this falls on a cache line. */\n"
	"const unsigned long hgr_to_color_highoff[2][1024][4]\n"
	"  __attribute__ ((aligned (16))) = { {");

  for (i = 0; i < 1024; i += 2)
    {
      int c[4];
      get_colors (i, c, 0);
      printf ("  { C%d, C%d, C%d, C%d },", c[0], c[1], c[2], c[3]);
      get_colors (i + 1, c, 0);
      printf (" { C%d, C%d, C%d, C%d },\n", c[0], c[1], c[2], c[3]);
    }

  puts ("}, {\n");

  for (i = 0; i < 1024; i += 2)
    {
      int c[4];
      get_colors (i, c, 1);
      printf ("  { C%d, C%d, C%d, C%d },", c[0], c[1], c[2], c[3]);
      get_colors (i + 1, c, 1);
      printf (" { C%d, C%d, C%d, C%d },\n", c[0], c[1], c[2], c[3]);
    }

  puts ("} };\n\n"
	"#undef C0\n"
	"#undef C1\n"
	"#undef C2\n"
	"#undef C3\n"
	"#undef C4\n"
	"#undef C5\n"
	"#undef C6\n"
	"#undef C7\n");

  return 0;
}


static void
get_colors (int n, int c[4], int highbit)
{
  int delta = highbit ? 4 : 0;
  int i;
#define BIT(z) (n & (1 << (z)))

  c[0] = c[1] = c[2] = c[3] = delta;   /* Default to black. */

  for (i = 0; i < 4; n >>= 1, i++)
    {
      /* First, is the pixel on? */
      if (BIT (3))
	{
	  /* Is a neighbor also on or are both pixels two away off? */
#if 0  /* If you enable this, lone pixels become white. */

	  if ((BIT (2) || BIT (4)) || !(BIT (1) || BIT (5)))
	    {
	      c[i] = 3 + delta;   /* This pixel is white. */
	      continue;
	    }
#else  /* With this enabled, lone pixels are colored. */
	  if (BIT (2) || BIT (4))
	    {
	      c[i] = 3 + delta;   /* This pixel is white. */
	    }
	  else if (!(BIT (1) || BIT (5)))
	    {
	      c[i] = ((i & 1) ? 1 : 2) + delta;
	    }
#endif
	  else
	    {
	      int color = ((i & 1) ? 1 : 2) + delta;
	      c[i] = color;
	      if (i > 0) c[i - 1] = color;
	      if (i < 3) c[i + 1] = color;
	    }
	}
      else
	{
	  if ((BIT (2) && !BIT (1) && (BIT (0) || BIT (4)))
	      || (BIT (4) && !BIT (5) && (BIT (2) || BIT (6))))
	    {
	      int color = ((i & 1) ? 2 : 1) + delta;
	      c[i] = color;
	    }
	}
    }
}
