/*
 * This file is part of zaniWok.
 *
 * zaniWok is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.  See the file COPYING for more information.
 *
 * zaniWok is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

/* This utility takes a 140K DOS 3.3 disk image in Prodos block order
 * and creates a .a2disk.
 */


#include <stdio.h>
#include <stdlib.h>
#include <sys/param.h>

#define BYTES_PER_TRACK (256 * 16)

int
main (int argc, char *argv[])
{
  char outfile[MAXPATHLEN + 1];
  FILE *in, *out;
  unsigned char buf[256];
  int i, s, size, ntracks;
  static const unsigned char header[] = {
    0x04, 0x13, 0x63, 0x59, 0x00, 0x00, 0x00, 0x09, 0x02 };
  static const int sector_offset[] = {
    0x000, 0xE00, 0xD00, 0xC00, 0xB00, 0xA00, 0x900, 0x800,
    0x700, 0x600, 0x500, 0x400, 0x300, 0x200, 0x100, 0xF00
  };

  if (argc < 2 || argc > 3)
    {
      fprintf (stderr, "Usage: %s infile [outfile]\n", argv[0]);
      exit (-1);
    }

  /* Open up the input file. */
  in = fopen (argv[1], "rb");
  if (in == NULL)
    {
      fprintf (stderr, "Unable to open file \"%s\".\n", argv[1]);
      exit (-2);
    }

  fseek (in, 0, SEEK_END);
  size = ftell (in);
  ntracks = size / BYTES_PER_TRACK;
  if (size % BYTES_PER_TRACK != 0
      || (ntracks != 35 && ntracks != 40))   /* Support 40 sector disks... */
    {
      fprintf (stderr, "Input file not exactly 35 or 40 tracks.\n");
      exit (-3);
    }

  /* Open up the output file. */
  if (argc != 3)
    sprintf (outfile, "%s.a2disk", argv[1]);
  else strcpy (outfile, argv[2]);
  out = fopen (outfile, "wb");
  if (out == NULL)
    {
      fprintf (stderr, "Unable to open file \"%s\" for writing.\n", outfile);
      exit (-4);
    }

  /* Output header. */
  fwrite (header, 1, sizeof header, out);

  for (i = 0; i < ntracks; i++)
    {
      for (s = 0; s < 16; s++)
	{
	  fseek (in, i * BYTES_PER_TRACK + sector_offset[s], SEEK_SET);
	  fread (buf, 1, 256, in);
	  fwrite (buf, 1, 256, out);
	}
    }

  fclose (in);
  fclose (out);

  return EXIT_SUCCESS;
}
 
